// **************************************************************************************
// *
// * © 2012 Dell, Inc.
// * ALL RIGHTS RESERVED.
// *
// * Redistribution and use in source and binary forms, with or without modification, 
// * are permitted provided that the following conditions are met:
// * 
// *  *  Redistributions of source code must retain the above copyright notice, 
// *     this list of conditions and the following disclaimer.  
// * 
// *  *  Redistributions in binary form must reproduce the above copyright notice, 
// *     this list of conditions and the following disclaimer in the documentation 
// *     and/or other materials provided with the distribution.  
// * 
// *  *  Neither the name of Dell, Inc. nor the names of its contributors 
// *     may be used to endorse or promote products derived from this software 
// *     without specific prior written permission. 
// * 
// * 
// * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ON AN "AS IS" 
// * AND UNSUPPORTED BASIS, WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES OF ANY KIND, 
// * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, NON-
// * INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL THE COPYRIGHT 
// * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
// * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
// * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
// * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
// * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
// * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// * 
// **************************************************************************************

#include <stdexcept>
#include "LocationDialog/locationDialog.h"
#include "DeleteLocationDialog/deleteLocationDialog.h"
#include "connectivityPage.h"

const int deleteColumnIndex       = 0;
const int locationNameColumnIndex = 1;
const int makeDefaultColumnIndex  = 2;


ConnectivityPage::ConnectivityPage(Farm* farm, ManageConnectionDialog* parent)
    : ManageConnectionPage(farm, parent)
{
    setupUi(this);
    m_pLocationDialog = new LocationDialog(farm, this);
    m_pDeleteLocationDialog = new DeleteLocationDialog(this);

    m_pSocket = SocketSharedPointer(new QSslSocket);

    bool connected = true;
    connected &= connect(donePushButton, SIGNAL(clicked()), parent, SLOT(accept()));
    connected &= connect(addLocationButton, SIGNAL(clicked()), SLOT(addLocation()));
    connected &= connect(editLocationButton, SIGNAL(clicked()),  SLOT(editLocation()));
    connected &= connect(locationsTableWidget, SIGNAL(cellClicked(int, int)), this, SLOT(cellClickedProcess(int, int)));
    connected &= connect(testBrokerButton, SIGNAL(clicked()), SLOT(testBroker()));
    if (!connected)
    {
        throw std::runtime_error("QT's signal and slot connection error in ConnectivityPage constructor");
    }

    loadFarmSettings();
}

void ConnectivityPage::loadFarmSettings()
{
    // Farm doesn`t exist
    if (m_pCurrentFarm == NULL)
    {
        return;
    }

    locationsTableWidget->clear();
    const int deleteColumnWidth = 20;
    const int makeDefaultColumnWidth = 60;
    const int locationNameColumnWidth = locationsTableWidget->width() - deleteColumnWidth - makeDefaultColumnWidth;
    locationsTableWidget->setColumnWidth(deleteColumnIndex, deleteColumnWidth);
    locationsTableWidget->setColumnWidth(locationNameColumnIndex, locationNameColumnWidth);
    locationsTableWidget->setColumnWidth(makeDefaultColumnIndex, makeDefaultColumnWidth);

    QStringList locationsList = m_pCurrentFarm->getLocationNamesList();
    locationsTableWidget->setRowCount(locationsList.count());
    for (int i = 0; i < locationsList.count(); ++i)
    {
        addTableWidgetRow(i, locationsList[i]);
    }

    if (!m_pCurrentFarm->isPromptLocation())
    {
        setDefaultLocation(m_pCurrentFarm->getDefaultLocation());
    }
}

void ConnectivityPage::saveFarmSettings()
{
    // Farm doesn`t exist
    if (m_pCurrentFarm == NULL)
    {
        return;
    }

    int defaultLocation = -1;
    for (int i = 0; i < locationsTableWidget->rowCount(); ++i)
    {
        if (locationsTableWidget->item(i, makeDefaultColumnIndex)->foreground().color() == QColor(Qt::blue))
        {
            defaultLocation = i;
            break;
        }
    }

    if (defaultLocation == -1)
    {
        m_pCurrentFarm->setPromptLocation(true);
    }
    else
    {
        m_pCurrentFarm->setPromptLocation(false);
        m_pCurrentFarm->setDefaultLocation(defaultLocation);
    }
}

void ConnectivityPage::cellClickedProcess(int row, int column)
{
    switch(column)
    {
    case deleteColumnIndex:
        deleteLocation(row);
        break;
    case locationNameColumnIndex:
        selectLocation(row);
        break;
    case makeDefaultColumnIndex:
        setDefaultLocation(row);
        break;
    default:
        break;
    }

}

void ConnectivityPage::deleteLocation(int index)
{
    m_pDeleteLocationDialog->setLocationName(locationsTableWidget->item(index, locationNameColumnIndex)->text());
    if (m_pDeleteLocationDialog->exec() == QDialog::Accepted)
    {
        m_pCurrentFarm->removeLocation(index);
        locationsTableWidget->removeRow(index);
    }
}

void ConnectivityPage::selectLocation(int index)
{
    for (int i = 0; i < locationsTableWidget->rowCount(); ++i)
    {
        locationsTableWidget->setCurrentCell(i, locationNameColumnIndex, QItemSelectionModel::Deselect);
    }
    locationsTableWidget->setCurrentCell(index, locationNameColumnIndex, QItemSelectionModel::Select);

    m_pCurrentFarm->setCurrentLocation(index);
}

void ConnectivityPage::setDefaultLocation(int index)
{
    bool wasDefault = locationsTableWidget->item(index, makeDefaultColumnIndex)->foreground().color() ==
                                                        QColor(Qt::blue);

    for (int i = 0; i < locationsTableWidget->rowCount(); ++i)
    {
        locationsTableWidget->item(i, makeDefaultColumnIndex)->setForeground(Qt::gray);
    }
    if (!wasDefault)
    {
        locationsTableWidget->item(index, makeDefaultColumnIndex)->setForeground(Qt::blue);
        m_pCurrentFarm->setDefaultLocation(index);
        m_pCurrentFarm->setPromptLocation(false);
    }
    else
    {
        m_pCurrentFarm->setPromptLocation(true);
    }
}

void ConnectivityPage::addLocation()
{
    LocationSharedPointer newLocation = m_pCurrentFarm->createNewLocation();
    int newIndex = m_pCurrentFarm->addLocation(newLocation);

    locationsTableWidget->insertRow(newIndex);
    addTableWidgetRow(newIndex, newLocation->getName());

    selectLocation(newIndex);

    m_pLocationDialog->loadFarmSettings();
    m_pLocationDialog->show();
}

void ConnectivityPage::editLocation()
{
    if (!locationsTableWidget->selectedItems().isEmpty())
    {
        m_pLocationDialog->loadFarmSettings();
        m_pLocationDialog->exec();
    }

    // Setting "default" column in locations list to proper state
    for (int i = 0; i < locationsTableWidget->rowCount(); ++i)
    {
        locationsTableWidget->item(i, makeDefaultColumnIndex)->setForeground(Qt::gray);
    }
    if (!m_pCurrentFarm->isPromptLocation())
    {
        locationsTableWidget->item(m_pCurrentFarm->getDefaultLocation(), makeDefaultColumnIndex)->setForeground(Qt::blue);
    }
}

void ConnectivityPage::addTableWidgetRow(int index, const QString& locationName)
{
    QTableWidgetItem* deleteIconItem = new QTableWidgetItem(QIcon(":/img/process-stop.png"), "");
    QTableWidgetItem* locationNameItem = new QTableWidgetItem(locationName);
    QTableWidgetItem* makeDefaultLabelItem = new QTableWidgetItem(tr("ConnectivityPage.Default"));
    locationsTableWidget->setItem(index, deleteColumnIndex, deleteIconItem);
    locationsTableWidget->setItem(index, locationNameColumnIndex, locationNameItem);
    locationsTableWidget->setItem(index, makeDefaultColumnIndex, makeDefaultLabelItem);
    locationsTableWidget->item(index, makeDefaultColumnIndex)->setForeground(Qt::gray);
}

void ConnectivityPage::testBroker()
{
    // Check that something actually is selected

    if (locationsTableWidget->selectedItems().isEmpty() || m_pCurrentFarm->getCurrentLocation()->getServers().isEmpty())
    {
        testConnectionResultLabel->setText(tr("ConnectivityPage.NoBroker.SelectBroker."));
        testConnectionResultLabel->setStyleSheet("QLabel {color: black;}");
        return;
    }

    bool connected = connect(m_pSocket.data(), SIGNAL(stateChanged(QAbstractSocket::SocketState)),
                             this, SLOT(testBrokerState(QAbstractSocket::SocketState)));
    if (!connected)
    {
        throw std::runtime_error("QT's signal and slot connection error in LocationDialog testBroker function");
    }

    testConnectionResultLabel->setText(tr("ConnectivityPage.TestInProgress."));
    testConnectionResultLabel->setStyleSheet("QLabel {color: black;}");

    LocationSharedPointer currentLocation = m_pCurrentFarm->getCurrentLocation();
    if (currentLocation->getProtocol().toUpper() == httpProtocol.toUpper())
    {
        m_pSocket->connectToHost(currentLocation->getServers()[0], currentLocation->getPort().toInt());
    }
    else
    {
        // Set ignore certificates error
        m_pSocket->setPeerVerifyMode(QSslSocket::VerifyNone);
        m_pSocket->connectToHostEncrypted(currentLocation->getServers()[0], currentLocation->getPort().toInt());
    }
}

void ConnectivityPage::testBrokerState(QAbstractSocket::SocketState state)
{
    if (m_pCurrentFarm->getCurrentLocation()->getProtocol().toUpper() == httpsProtocol.toUpper() &&
        state == QAbstractSocket::ConnectedState)
    {
        // waitForEncryption calls this slot. Needs to be disconnected
        disconnect(m_pSocket.data(), 0, 0, 0);
        if (!m_pSocket->waitForEncrypted())
        {
            state = QAbstractSocket::UnconnectedState;
        }
    }

    if (state == QAbstractSocket::UnconnectedState)
    {
        testConnectionResultLabel->setText(tr("ConnectivityPage.TestConnectionError."));
        testConnectionResultLabel->setStyleSheet("QLabel {color: red;}");
        disconnect(m_pSocket.data(), 0, 0, 0);
        m_pSocket->disconnectFromHost();
    }
    else if (state == QAbstractSocket::ConnectedState)
    {
        testConnectionResultLabel->setText(tr("ConnectivityPage.TestConnectionSuccess."));
        testConnectionResultLabel->setStyleSheet("QLabel {color: green;}");
        disconnect(m_pSocket.data(), 0, 0, 0);
        m_pSocket->disconnectFromHost();
    }
}
