// **************************************************************************************
// *
// * © 2012 Dell, Inc.
// * ALL RIGHTS RESERVED.
// *
// * Redistribution and use in source and binary forms, with or without modification, 
// * are permitted provided that the following conditions are met:
// * 
// *  *  Redistributions of source code must retain the above copyright notice, 
// *     this list of conditions and the following disclaimer.  
// * 
// *  *  Redistributions in binary form must reproduce the above copyright notice, 
// *     this list of conditions and the following disclaimer in the documentation 
// *     and/or other materials provided with the distribution.  
// * 
// *  *  Neither the name of Dell, Inc. nor the names of its contributors 
// *     may be used to endorse or promote products derived from this software 
// *     without specific prior written permission. 
// * 
// * 
// * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ON AN "AS IS" 
// * AND UNSUPPORTED BASIS, WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES OF ANY KIND, 
// * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, NON-
// * INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL THE COPYRIGHT 
// * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
// * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
// * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
// * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
// * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
// * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// * 
// **************************************************************************************

#include <stdexcept>
#include "common/common.h"
#include "farm.h"
#include "farmSettingsNames.h"
#include "locationDialog.h"

const QString httpPrefix    = "http://";
const QString httpsPrefix   = "https://";
const QString httpPort      = "8080";
const QString httpsPort     = "443";


LocationDialog::LocationDialog(Farm* farm, QWidget* parent)
    : CommonDialog(parent), m_pCurrentFarm(farm)
{
    Ui::LocationDialog::setupUi(this);
    setWindowMask("", tabWidget->width(), tabWidget->height());

    m_pProxyButtonGroup = new QButtonGroup;
    m_pProxyButtonGroup->addButton(defualtProxyRadioButton);
    m_pProxyButtonGroup->addButton(dontUseProxyRadioButton);
    m_pProxyButtonGroup->addButton(manuallyProxyRadioButton);
    m_pProxyButtonGroup->setExclusive(true);

    m_pAddBrokerDialog = new AddBrokerDialog(this);

    m_pSocket = SocketSharedPointer(new QSslSocket);

    portLineEdit->setValidator(new QIntValidator(1, 65535, this));

    bool connected = true;
    connected &= connect(doneButton1, SIGNAL(clicked()), SLOT(acceptChanges()));
    connected &= connect(doneButton2, SIGNAL(clicked()), SLOT(acceptChanges()));
    connected &= connect(m_pProxyButtonGroup, SIGNAL(buttonClicked(QAbstractButton*)),
                         SLOT(currentProxyOptionChanged(QAbstractButton*)));
    connected &= connect(sslEnabledCheckBox, SIGNAL(toggled(bool)), SLOT(enableSslGateway(bool)));
    connected &= connect(upBrokerButton, SIGNAL(clicked()), SLOT(upBroker()));
    connected &= connect(downBrokerButton, SIGNAL(clicked()), SLOT(downBroker()));
    connected &= connect(removeBrokerButton, SIGNAL(clicked()), SLOT(removeBroker()));
    connected &= connect(addBrokerButton, SIGNAL(clicked()), SLOT(addBroker()));
    connected &= connect(testBrokerButton, SIGNAL(clicked()), SLOT(testBroker()));
    connected &= connect(brokersList, SIGNAL(itemSelectionChanged()), SLOT(checkSelectedBroker()));
    connected &= connect(protocolComboBox, SIGNAL(currentIndexChanged(const QString&)),
                         SLOT(currentProtocolChanged(const QString&)));
    if (!connected)
    {
        throw std::runtime_error("QT's signal and slot connection error in LocationDialog constructor");
    }

    checkSelectedBroker();

    m_appSettings = SharedApplicationSettings(new ApplicationSettings(APPLICATION_CONFIG_FILE_NAME));
    m_appSettings->beginGroup(FARM_OVERRIDE_VALUES);
}

void LocationDialog::loadFarmSettings()
{
    // Farm or current location don`t exist
    if (m_pCurrentFarm == NULL || m_pCurrentFarm->getCurrentLocation() == NULL)
    {
        return;
    }

    LocationSharedPointer location = m_pCurrentFarm->getCurrentLocation();

    // Location Settings
    locationNameEdit->setText(location->getName());

    bool promptForLocation = m_pCurrentFarm->isPromptLocation();
    alwaysPromptLocationCheckBox->setChecked(promptForLocation);
    bool isDefault = (m_pCurrentFarm->getLocation(m_pCurrentFarm->getDefaultLocation()) == m_pCurrentFarm->getCurrentLocation());
    useLocationDefaultCheckBox->setChecked(isDefault);

    portLineEdit->setText(location->getPort());
    brokersList->clear();
    brokersList->addItems(location->getServers());

    QString protocol = location->getProtocol();
    int protocolIndex = protocolComboBox->findText(protocol, Qt::MatchFixedString);
    if (protocolIndex < 0)
    {
        protocolIndex = 0;
    }
    protocolComboBox->setCurrentIndex(protocolIndex);

    bool tmpBool;
    QString tmpStr;
    int tmpInt;
    QStringList tmpStrList;
    if (m_appSettings->getValue(PROMPT_FOR_LOCATION, &tmpBool))
    {
        if (tmpBool)
        {
            useLocationDefaultCheckBox->setEnabled(false);
        }
        else
        {
            alwaysPromptLocationCheckBox->setEnabled(false);
        }
    }
    if (m_appSettings->getValue(DEFAULT_LOCATION, &tmpInt))
    {
        useLocationDefaultCheckBox->setEnabled(false);
    }
    protocolComboBox->setEnabled(!m_appSettings->getValue(LOCATION_PROTOCOL, &tmpStr));
    portLineEdit->setEnabled(!m_appSettings->getValue(LOCATION_TCP_PORT, &tmpInt));
    if (m_appSettings->getValueList(LOCATION_SERVERS, &tmpStrList))
    {
        addBrokerButton->setEnabled(false);
        brokersList->setEnabled(false);
    }

    // Firewall & Proxy Settings
    natEnabledCheckBox->setChecked(location->isNat());

    sslGatewayLineEdit->setText(location->getSslGateway());
    sslEnabledCheckBox->setChecked(location->isSsl());

    QString proxyServer = location->getProxyServer();
    QString proxyBypassList = location->getProxyBypassList();

    // Since we have the port as a seperate field now
    // we will need to break up the IP and port.
    QStringList sList = proxyServer.split( ":", QString::SkipEmptyParts );

    if ( sList.count() == 2 )
    {
        proxyLineEdit->setText( sList.at( 0 ));
        proxyPortLineEdit->setText( sList.at( 1 ));
    }

    dontUseProxyLineEdit->setText(proxyBypassList);
    if (proxyServer.isEmpty() && proxyBypassList.isEmpty())
    {
        dontUseProxyRadioButton->setChecked(true);
        currentProxyOptionChanged(dontUseProxyRadioButton);
    }
    if (proxyServer.isEmpty() && !proxyBypassList.isEmpty())
    {
        defualtProxyRadioButton->setChecked(true);
        currentProxyOptionChanged(defualtProxyRadioButton);
    }
    if (!proxyServer.isEmpty())
    {
        manuallyProxyRadioButton->setChecked(true);
        currentProxyOptionChanged(manuallyProxyRadioButton);
    }

    natEnabledCheckBox->setEnabled(!m_appSettings->getValue(LOCATION_NAT, &tmpBool));
    sslEnabledCheckBox->setEnabled(!m_appSettings->getValue(LOCATION_RDP_ON_SSL, &tmpBool));
    sslGatewayLineEdit->setEnabled(!m_appSettings->getValue(LOCATION_SSL_GATEWAY, &tmpStr) && location->isSsl());
    if (m_appSettings->getValue(LOCATION_PROXY_SERVER, &tmpStr))
    {
        if (proxyServer.isEmpty())
        {
            manuallyProxyRadioButton->setEnabled(false);
        }
        else
        {
            dontUseProxyRadioButton->setEnabled(false);
            defualtProxyRadioButton->setEnabled(false);
        }
    }
    if (m_appSettings->getValue(LOCATION_PROXY_BYPASS_LIST, &tmpStr) && !proxyBypassList.isEmpty())
    {
        dontUseProxyRadioButton->setEnabled(false);
    }
}

void LocationDialog::saveFarmSettings()
{
    // Farm or current location don`t exist
    if (m_pCurrentFarm == NULL || m_pCurrentFarm->getCurrentLocation() == NULL)
    {
        return;
    }

    LocationSharedPointer location = m_pCurrentFarm->getCurrentLocation();

    // Location Settings
    location->setName(locationNameEdit->text());

    m_pCurrentFarm->setPromptLocation(alwaysPromptLocationCheckBox->isChecked());
    if (useLocationDefaultCheckBox->isChecked())
    {
        int locationsCount = m_pCurrentFarm->getLocationNamesList().count();
        for (int i = 0; i < locationsCount; ++i)
        {
            if (location == m_pCurrentFarm->getLocation(i))
            {
                m_pCurrentFarm->setDefaultLocation(i);
                break;
            }
        }
    }

    location->setProtocol(protocolComboBox->currentText().toLower());
    location->setPort(portLineEdit->text());
    QStringList locationServers;
    for (int i = 0; i < brokersList->count(); ++i)
    {
        locationServers.append(brokersList->item(i)->text());
    }
    location->setServers(locationServers);

    // Firewall & Proxy Settings
    location->setNat(natEnabledCheckBox->isChecked());

    location->setSslGateway(sslGatewayLineEdit->text());
    location->setSsl(sslEnabledCheckBox->isChecked());

    if ( !(proxyPortLineEdit->text().isEmpty()) )
        location->setProxyServer(proxyLineEdit->text() + QString(":") + proxyPortLineEdit->text());
    else
        location->setProxyServer(proxyLineEdit->text());
    location->setProxyBypassList(dontUseProxyLineEdit->text());
}

void LocationDialog::acceptChanges()
{
    saveFarmSettings();
    testConnectionResultLabel->setText("");
    accept();
}

void LocationDialog::currentProxyOptionChanged(QAbstractButton* button)
{
    // Farm or current location don`t exist
    if (m_pCurrentFarm == NULL || m_pCurrentFarm->getCurrentLocation() == NULL)
    {
        return;
    }

    LocationSharedPointer location = m_pCurrentFarm->getCurrentLocation();


    QString tmpStr;
    if (button == defualtProxyRadioButton)
    {
        proxyLineEdit->setEnabled(false);
        proxyPortLineEdit->setEnabled(false);
        if (!m_appSettings->getValue(LOCATION_PROXY_SERVER, &tmpStr))
        {
            proxyLineEdit->setText("");
            proxyPortLineEdit->setText("");
            location->setProxyServer(proxyLineEdit->text());
        }
        if (!m_appSettings->getValue(LOCATION_PROXY_BYPASS_LIST, &tmpStr))
        {
            dontUseProxyLabel->setEnabled(true);
            descProxyLabel->setEnabled(true);
            dontUseProxyLineEdit->setEnabled(true);
        }
        else
        {
            dontUseProxyLabel->setEnabled(false);
            descProxyLabel->setEnabled(false);
            dontUseProxyLineEdit->setEnabled(false);
        }
    }
    if (button == dontUseProxyRadioButton)
    {
        proxyLineEdit->setEnabled(false);
        proxyPortLineEdit->setEnabled(false);
        if (!m_appSettings->getValue(LOCATION_PROXY_SERVER, &tmpStr))
        {
            proxyLineEdit->setText("");
            proxyPortLineEdit->setText("");
            location->setProxyServer(proxyLineEdit->text());
        }
        dontUseProxyLabel->setEnabled(false);
        descProxyLabel->setEnabled(false);
        dontUseProxyLineEdit->setEnabled(false);
        if (!m_appSettings->getValue(LOCATION_PROXY_BYPASS_LIST, &tmpStr))
        {
            dontUseProxyLineEdit->setText("");
            location->setProxyBypassList(dontUseProxyLineEdit->text());
        }
    }
    if (button == manuallyProxyRadioButton)
    {
        if (!m_appSettings->getValue(LOCATION_PROXY_SERVER, &tmpStr))
        {
            proxyLineEdit->setEnabled(true);
            proxyPortLineEdit->setEnabled(true);
        }
        else
        {
            proxyLineEdit->setEnabled(false);
            proxyPortLineEdit->setEnabled(false);
        }
        if (!m_appSettings->getValue(LOCATION_PROXY_BYPASS_LIST, &tmpStr))
        {
            dontUseProxyLabel->setEnabled(true);
            descProxyLabel->setEnabled(true);
            dontUseProxyLineEdit->setEnabled(true);
        }
        else
        {
            dontUseProxyLabel->setEnabled(false);
            descProxyLabel->setEnabled(false);
            dontUseProxyLineEdit->setEnabled(false);
        }
    }
}

void LocationDialog::enableSslGateway(bool enableSslGateway)
{
    bool tmpBool;
    if (m_appSettings->getValue(LOCATION_SSL_GATEWAY, &tmpBool))
    {
        enableSslGateway = false;
    }
    sslGatewayLineEdit->setEnabled(enableSslGateway);

    if (!enableSslGateway)
    {
        sslGatewayLineEdit->setText("");
    }
}

void LocationDialog::upBroker()
{
    int currentIndex = brokersList->currentRow();
    if (currentIndex > 0)
    {
        QListWidgetItem* tempItem = brokersList->takeItem(currentIndex);
        brokersList->insertItem(currentIndex - 1, tempItem->text());
        brokersList->setCurrentRow(currentIndex - 1);
    }
    brokersList->setFocus();
}

void LocationDialog::downBroker()
{
    int currentIndex = brokersList->currentRow();
    if (currentIndex < brokersList->count() - 1)
    {
        QListWidgetItem* tempItem = brokersList->takeItem(currentIndex);
        brokersList->insertItem(currentIndex + 1, tempItem->text());
        brokersList->setCurrentRow(currentIndex + 1);
    }
    brokersList->setFocus();
}

void LocationDialog::removeBroker()
{
    if (QMessageBox::question(this, this->windowTitle(), tr("LocationDialog.SureWantToDeleteBroker[%1]?").
                              arg(brokersList->currentItem()->text()), QMessageBox::Yes | QMessageBox::No)
                      == QMessageBox::No)
    {
        return;
    }

    if (brokersList->currentItem() != NULL)
    {
        brokersList->takeItem(brokersList->currentRow());
    }
    else
    {
        QMessageBox::critical(this, tr("LocationDialog.RemoveBroker"), tr("LocationDialog.NoSelectedBroker"));
    }
}

void LocationDialog::addBroker()
{
    bool ok;
    QString text;

    do
    {
        ok = (m_pAddBrokerDialog->exec() == QDialog::Accepted);
        text = m_pAddBrokerDialog->getBroker();
        if (ok)
        {
            text.remove(httpPrefix);
            text.remove(httpsPrefix);

            if (!text.isEmpty() && !text.trimmed().isEmpty())
            {
                if (brokersList->findItems(text, Qt::MatchExactly).isEmpty())
                {
                    brokersList->addItem(text);
                    break;
                }
                else
                {
                    QMessageBox::information(this, tr("LocationDialog.BrokersList"),
                                                   tr("LocationDialog.ListAlreadyContainsItem"));
                    ok = false;
                }
            }
            else
            {
                QMessageBox::information(this, tr("LocationDialog.BrokersList"),
                                               tr("LocationDialog.EnterValueOrClickCancel"));
                ok = false;
            }
        }
        else
        {
            break;
        }
    }
    while(!ok);
}

void LocationDialog::testBroker()
{
    // Check that something actually is selected

    if (brokersList->currentIndex().row() < 0)
    {
        testConnectionResultLabel->setText(tr("LocationDialog.NoBrokerSelected.SelectBroker."));
        testConnectionResultLabel->setStyleSheet("QLabel {color: black;}");
        return;
    }

    bool connected = connect(m_pSocket.data(), SIGNAL(stateChanged(QAbstractSocket::SocketState)),
                             this, SLOT(testBrokerState(QAbstractSocket::SocketState)));
    if (!connected)
    {
        throw std::runtime_error("QT's signal and slot connection error in LocationDialog testBroker function");
    }

    testConnectionResultLabel->setText(tr("LocationDialog.TestInProgress."));
    testConnectionResultLabel->setStyleSheet("QLabel {color: black;}");

    if (protocolComboBox->currentText().toUpper() == httpProtocol.toUpper())
    {
        m_pSocket->connectToHost(brokersList->currentItem()->text(), portLineEdit->text().toInt());
    }
    else
    {
        // Set ignore certificates error
        m_pSocket->setPeerVerifyMode(QSslSocket::VerifyNone);
        m_pSocket->connectToHostEncrypted(brokersList->currentItem()->text(), portLineEdit->text().toInt());
    }
}

void LocationDialog::testBrokerState(QAbstractSocket::SocketState state)
{
    if (protocolComboBox->currentText().toUpper() == httpsProtocol.toUpper() &&
        state == QAbstractSocket::ConnectedState)
    {
        // waitForEncryption calls this slot. Needs to be disconnected
        disconnect(m_pSocket.data(), 0, 0, 0);
        if (!m_pSocket->waitForEncrypted())
        {
            state = QAbstractSocket::UnconnectedState;
        }
    }

    if (state == QAbstractSocket::UnconnectedState)
    {
        testConnectionResultLabel->setText(tr("LocationDialog.TestConnectionError."));
        testConnectionResultLabel->setStyleSheet("QLabel {color: red;}");
        disconnect(m_pSocket.data(), 0, 0, 0);
        m_pSocket->disconnectFromHost();
    }
    else if (state == QAbstractSocket::ConnectedState)
    {
        testConnectionResultLabel->setText(tr("LocationDialog.TestConnectionSuccess."));
        testConnectionResultLabel->setStyleSheet("QLabel {color: green;}");
        disconnect(m_pSocket.data(), 0, 0, 0);
        m_pSocket->disconnectFromHost();
    }
}

void LocationDialog::checkSelectedBroker()
{
    QList<QListWidgetItem*> itemsList = brokersList->selectedItems();
    if (itemsList.count() == 0)
    {
        setEnabledBrokerButtons(false);
    }
    else
    {
        setEnabledBrokerButtons(true);
    }
}

void LocationDialog::setEnabledBrokerButtons(bool enabled)
{
    removeBrokerButton->setEnabled(enabled);
    upBrokerButton->setEnabled(enabled);
    downBrokerButton->setEnabled(enabled);
    testBrokerButton->setEnabled(enabled);
}

void LocationDialog::currentProtocolChanged(const QString& protocol)
{
    if (portLineEdit->isEnabled())
    {
        if (protocol.toUpper() == httpProtocol.toUpper())
        {
            portLineEdit->setText(httpPort);
        }
        else
        {
            portLineEdit->setText(httpsPort);
        }
    }
}
