// **************************************************************************************
// *
// * © 2012 Dell, Inc.
// * ALL RIGHTS RESERVED.
// *
// * Redistribution and use in source and binary forms, with or without modification, 
// * are permitted provided that the following conditions are met:
// * 
// *  *  Redistributions of source code must retain the above copyright notice, 
// *     this list of conditions and the following disclaimer.  
// * 
// *  *  Redistributions in binary form must reproduce the above copyright notice, 
// *     this list of conditions and the following disclaimer in the documentation 
// *     and/or other materials provided with the distribution.  
// * 
// *  *  Neither the name of Dell, Inc. nor the names of its contributors 
// *     may be used to endorse or promote products derived from this software 
// *     without specific prior written permission. 
// * 
// * 
// * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ON AN "AS IS" 
// * AND UNSUPPORTED BASIS, WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES OF ANY KIND, 
// * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, NON-
// * INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL THE COPYRIGHT 
// * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
// * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
// * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
// * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
// * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
// * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// * 
// **************************************************************************************

#include <stdexcept>
#include <syslog.h>
#include "farmSettingsNames.h"
#include "userExperiencePage.h"
#include "common/FlashPluginUtils.h"

const QString daModeLocal = "local";
const QString daQualityCd = "cd";
const QString uaModeSpeex = "speex";

UserExperiencePage::UserExperiencePage(Farm* farm, ManageConnectionDialog* parent)
    : ManageConnectionPage(farm, parent)
{
    setupUi(this);
    bool connected = true;
    connected += connect(donePushButton, SIGNAL(clicked()), parent, SLOT(accept()));
    connected += connect(audioEnableCheckBox, SIGNAL(toggled(bool)), parent, SLOT(eopAudioToggled(bool)));

    if (!connected)
    {
        throw std::runtime_error("QT's signal and slot connection error in UserExperiencePage constructor");
    }

    loadFarmSettings();
    parent->eopAudioToggled(audioEnableCheckBox->isChecked());
}

// Set the performance options when the combo box changes value
// and overide previous settings.  This is the same behaviour as
// windows and we will duplicate it here.
void UserExperiencePage::loadFarmSettings()
{
    if (m_pCurrentFarm == NULL)
    {
        return;
    }

    gaEnableCheckBox->setChecked(m_pCurrentFarm->isGaEnabled());
    fmmrEnableCheckBox->setChecked(m_pCurrentFarm->isFmmrEnabled());
    mmrEnableCheckBox->setChecked(m_pCurrentFarm->isMmrEnabled());
    pndnaEnableCheckBox->setChecked(m_pCurrentFarm->isDnaEnabled());
    reconnectCheckBox->setChecked(m_pCurrentFarm->isReconnection());

    // Fill in the speed settings
    connectionSpeedComboBox->setCurrentIndex( m_pCurrentFarm->getExperienceLevel());
    desktopBackgroundCheckBox->setChecked( m_pCurrentFarm->isWallpaperEnabled());
    showDraggingCheckBox->setChecked( m_pCurrentFarm->isDraggingEnabled());
    menuAnimationCheckBox->setChecked( m_pCurrentFarm->isAnimationEnabled());
    visualStylesCheckBox->setChecked( m_pCurrentFarm->isThemesEnabled());
    bitmapCachingCheckBox->setChecked(m_pCurrentFarm->isBitmapCachingEnabled());

    bool tmpBool;
    gaEnableCheckBox->setEnabled(!m_appSettings->getValue(GRAPHICS_ACCELERATION, &tmpBool));
    fmmrEnableCheckBox->setEnabled(!m_appSettings->getValue(FMMR, &tmpBool));
    mmrEnableCheckBox->setEnabled(!m_appSettings->getValue(MMR, &tmpBool));
    pndnaEnableCheckBox->setEnabled(!m_appSettings->getValue(ENABLE_DNA, &tmpBool));
    reconnectCheckBox->setEnabled(!m_appSettings->getValue(RECONNECT, &tmpBool));

    // Fill in the speed settings
    connectionSpeedComboBox->setEnabled( !m_appSettings->getValue(EXPERIENCE_LEVEL, &tmpBool));
    desktopBackgroundCheckBox->setEnabled( !m_appSettings->getValue(ENABLE_WALLPAPER, &tmpBool));
    showDraggingCheckBox->setEnabled( !m_appSettings->getValue(ENABLE_FULL_WINDOW_DRAG, &tmpBool));
    menuAnimationCheckBox->setEnabled( !m_appSettings->getValue(ENABLE_ANIMATION, &tmpBool));
    visualStylesCheckBox->setEnabled( !m_appSettings->getValue(ENABLE_THEMES, &tmpBool));
    bitmapCachingCheckBox->setEnabled(!m_appSettings->getValue(ENABLE_BITMAP_CACHING, &tmpBool));

    if (isFlashPluginInstalled())
    {
        flashIsAbsentLabel->hide();
    }
    else
    {
        fmmrEnableCheckBox->setEnabled(false);
    }
}

void UserExperiencePage::saveFarmSettings()
{
    if (m_pCurrentFarm == NULL)
    {
        return;
    }
    if (audioEnableCheckBox->isChecked())
    {
        m_pCurrentFarm->setDownstreamAudioEnabled(true);
        m_pCurrentFarm->setDAMode(daModeLocal);
        m_pCurrentFarm->setDAQuality(daQualityCd);
        m_pCurrentFarm->setDABuffering(0);
        m_pCurrentFarm->setDACustomQuality(0);

        m_pCurrentFarm->setUpstreamAudioEnabled(true);
        m_pCurrentFarm->setUAMode(uaModeSpeex);
        m_pCurrentFarm->setUAQuality(0);
        m_pCurrentFarm->setUAComplexity(0);
        m_pCurrentFarm->setUAEnhancedCompression(false);
    }

    m_pCurrentFarm->setGaEnable(gaEnableCheckBox->isChecked());
    m_pCurrentFarm->setFmmrEnable(fmmrEnableCheckBox->isChecked());
    m_pCurrentFarm->setMmrEnable(mmrEnableCheckBox->isChecked());
    m_pCurrentFarm->setDnaEnabled( pndnaEnableCheckBox->isChecked());
    m_pCurrentFarm->setReconnection(reconnectCheckBox->isChecked());

    // Save the performance settings
    m_pCurrentFarm->setExperienceLevel( connectionSpeedComboBox->currentIndex());
    m_pCurrentFarm->setWallpaperEnable( desktopBackgroundCheckBox->isChecked() );
    m_pCurrentFarm->setDraggingEnable ( showDraggingCheckBox->isChecked()      );
    m_pCurrentFarm->setAnimationEnable( menuAnimationCheckBox->isChecked()     );
    m_pCurrentFarm->setThemesEnable   ( visualStylesCheckBox->isChecked()      );
    m_pCurrentFarm->setBitmapCachingEnable(bitmapCachingCheckBox->isChecked()  );
}

void UserExperiencePage::on_connectionSpeedComboBox_currentIndexChanged(int index)
{
    // Set the values according to the index #
    syslog( LOG_INFO, "USer change the performance options to %i", index);

    // If modem or better (defaults)
    desktopBackgroundCheckBox->setChecked( false );
    showDraggingCheckBox->     setChecked( false );
    menuAnimationCheckBox->    setChecked( false );
    visualStylesCheckBox->     setChecked( false );
    bitmapCachingCheckBox->    setChecked( true  );

    // If Lo Speed Broadband
    if ( index >= 1 )
    {
        visualStylesCheckBox->setChecked( true );
    }

    // If Hi Speed Broadband
    // Currently no different from lo speed broadband

    // If LAN equivalent
    if ( index >= 3 )
    {
        desktopBackgroundCheckBox->setChecked( true );
        showDraggingCheckBox->     setChecked( true );
        menuAnimationCheckBox->    setChecked( true );
    }

    // Done.

}
