// **************************************************************************************
// *
// * © 2012 Dell, Inc.
// * ALL RIGHTS RESERVED.
// *
// * Redistribution and use in source and binary forms, with or without modification, 
// * are permitted provided that the following conditions are met:
// * 
// *  *  Redistributions of source code must retain the above copyright notice, 
// *     this list of conditions and the following disclaimer.  
// * 
// *  *  Redistributions in binary form must reproduce the above copyright notice, 
// *     this list of conditions and the following disclaimer in the documentation 
// *     and/or other materials provided with the distribution.  
// * 
// *  *  Neither the name of Dell, Inc. nor the names of its contributors 
// *     may be used to endorse or promote products derived from this software 
// *     without specific prior written permission. 
// * 
// * 
// * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ON AN "AS IS" 
// * AND UNSUPPORTED BASIS, WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES OF ANY KIND, 
// * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, NON-
// * INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL THE COPYRIGHT 
// * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
// * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
// * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
// * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
// * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
// * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// * 
// **************************************************************************************

#include <stdexcept>
#include <QtDebug>
#include <QMouseEvent>

#include "manageConnectionPage.h"
#include "TitlePage/titlePage.h"
#include "ConnectivityPage/connectivityPage.h"
#include "CredentialsPage/credentialsPage.h"
#include "DisplayPage/displayPage.h"
#include "LocalResourcesPage/localResourcesPage.h"
#include "UserExperiencePage/userExperiencePage.h"
#include "PasswordManagementPage/passwordManagementPage.h"
#include "AutoLaunchPage/autoLaunchPage.h"

#include "HelpWindow/helpWindow.h"
#include "DeleteFarmDialog/deleteFarmDialog.h"

#include "manageConnectionDialog.h"
#include "farmSettingsNames.h"

#include "SimpleAddFarm/simpleaddfarmdialog.h"

#include "mainWindow.h"

// Name of boolean property in Farm Settings which determines will HelpSteps window shown or not


ManageConnectionDialog::ManageConnectionDialog(FarmSettings* settings, bool isAddFarmMode,
                        const QString& actuallyFarmName, QWidget* parent)
    : QDialog(parent)
{
    setupUi(this);
    setWindowFlags(Qt::FramelessWindowHint | Qt::Dialog);
    setAttribute(Qt::WA_DeleteOnClose, true);

    m_deleteCurrentFarmOnClose = false;
    m_pFarmSettings = settings;
    m_pHelpWindow = NULL;

    createHelpWindow();

    m_pDeleteFarmDialog = new DeleteFarmDialog(this);

    // Copying settings
    QString farmName = actuallyFarmName;
    if (farmName.isEmpty() && !m_pFarmSettings->getFarmNamesList().isEmpty())
    {
        farmName = m_pFarmSettings->getFarm(0)->getName();
    }

    bool connected = true;
    connected &= connect(closeButton, SIGNAL(clicked()), SLOT(cancel()));
    connected &= connect(backButton, SIGNAL(clicked()), SLOT(back()));
    connected &= connect(nextButton, SIGNAL(clicked()), SLOT(next()));
    connected &= connect(deleteFarmButton, SIGNAL(clicked()), SLOT(deleteFarm()));
    connected &= connect(basicViewButton, SIGNAL(clicked()), SLOT(switchToBasicView()));
    connected &= connect(pagesStackedWidget, SIGNAL(currentChanged(int)), SLOT(currentPageChanged(int)));
    if (!connected)
    {
        throw std::runtime_error("QT's signal and slot connection error in ManageConnectionDialog constructor");
    }

    // Setting add or edit mode of dialog
    setDialogMode(isAddFarmMode);

    // Set editing farm
    if (isAddFarmMode)
    {
        setFarmForAddDialog(farmName);
    }
    else
    {
        setFarmForEditDialog(farmName);
    }
    createConnectionDialog();
    const int titlePageIndex = 0;
    currentPageChanged(titlePageIndex);
    nextButton->setEnabled(true);

    // Fix the Password management label
    QIcon newIcon = QIcon( ":/img2/Password_Mgmt.png");

    pagesStackedWidget->setTabIcon( 6 , newIcon);


    // Loading farm settings
    loadFarmSettings();

    updateNavButtons();
}

ManageConnectionDialog::~ManageConnectionDialog()
{
    if (m_deleteCurrentFarmOnClose)
    {
        delete m_pCurrentFarm;
    }
}

// === GUI FUNCTIONS

void ManageConnectionDialog::createHelpWindow()
{
    bool showSteps = m_pFarmSettings->value(SHOW_STEPS, true).toBool();
    if (showSteps)
    {
        m_pHelpWindow = new HelpWindow(this);
    }

    if (m_pHelpWindow)
    {
        const int initialPosX = 300;
        const int initialPosY = 50;
        move(initialPosX, initialPosY);

        // This is for normal simultaneous minimizing and maximizing of Help window
        //Needs to be fixed
        m_pHelpWindow->show();
        m_pHelpWindow->close();
        m_pHelpWindow->show();

        m_helpString.append(tr("ManageConnectionDialog.HelpStep1."));
        m_helpString.append(tr("ManageConnectionDialog.HelpStep2."));
        m_helpString.append(tr("ManageConnectionDialog.HelpStep3."));
        m_helpString.append(tr("ManageConnectionDialog.HelpStep4."));
        m_helpString.append(tr("ManageConnectionDialog.HelpStep5."));
        m_helpString.append(tr("ManageConnectionDialog.HelpStep6."));
        m_helpString.append(tr("ManageConnectionDialog.HelpStep7."));
        m_helpString.append(tr("ManageConnectionDialog.HelpStep8."));
    }
}

void ManageConnectionDialog::setDialogMode(bool isAddFarmMode)
{
    m_isAddFarmMode = isAddFarmMode;

    nextButton->setEnabled(isAddFarmMode);
    backButton->setEnabled(isAddFarmMode);

    deleteFarmButton->setVisible(!isAddFarmMode);
    basicViewButton->setVisible(isAddFarmMode);
}

void ManageConnectionDialog::addPage(QWidget* page, const QString& icon)
{
    QString styleSheet = "";

    styleSheet.append( "QWidget#widget { border: 0px solid transparent;"
                                      "background: qlineargradient(spread:pad, x1:0, y1:0, x2:0, y2:1, "
                                      "stop:0 rgba(180, 180, 180, 255), stop:0.02 rgba(255, 255, 255, 255), "
                                      "stop:0.98 rgba(255, 255, 255, 255), stop:1 rgba(180, 180, 180, 255));"
                                      "}");


    page->setStyleSheet( styleSheet );
    pagesStackedWidget->addTab(page, QIcon(icon), "");
}

void ManageConnectionDialog::createConnectionDialog()
{
    pagesStackedWidget->clear();

    // Create title page in dialog
    TitlePage *titlePage = new TitlePage(m_pCurrentFarm, this);
    if (m_isAddFarmMode)
    {
        titlePage->setTitle(tr("ManageConnectionDialog.AddFarm"));
    }
    else
    {
        titlePage->setTitle(tr("ManageConnectionDialog.EditFarm"));
    }
    addPage(titlePage, ":/img2/Edit_Farm.png");
    addPage(new ConnectivityPage(m_pCurrentFarm, this), ":/img2/Locations.png");
    addPage(new CredentialsPage(m_pCurrentFarm, this), ":/img2/Credentials.png");
    addPage(new DisplayPage(m_pCurrentFarm, this), ":/img2/Display.png");
    m_pLocalResourcesPage = new LocalResourcesPage(m_pCurrentFarm, this);
    addPage(m_pLocalResourcesPage, ":/img2/Local_resources.png");
    addPage(new UserExperiencePage(m_pCurrentFarm, this), ":/img2/Experience.png");
    addPage(new PasswordManagementPage(m_pCurrentFarm, this), ":/img2/Password_management.png");
    addPage(new AutoLaunchPage(m_pCurrentFarm, this), ":/img2/Auto-Launch.png");

    pagesStackedWidget->setCurrentIndex(0);
}

void ManageConnectionDialog::mousePressEvent(QMouseEvent* event)
{
    if (event->button() == Qt::LeftButton)
    {
        m_dragPosition = event->globalPos() - frameGeometry().topLeft();
        event->accept();
    }
}

void ManageConnectionDialog::mouseMoveEvent(QMouseEvent* event)
{
    if (event->buttons() & Qt::LeftButton)
    {
        move(event->globalPos() - m_dragPosition);
        moveHelpWindow();
        event->accept();
    }
}

void ManageConnectionDialog::next()
{
    // Additive functions, when next button was pushed
    ManageConnectionPage* currentPage = static_cast<ManageConnectionPage*>(pagesStackedWidget->currentWidget());
    bool bResult = currentPage->nextEvents();
    if (!bResult)
    {
        return;
    }
    int currentIndex = pagesStackedWidget->currentIndex();
    pagesStackedWidget->setCurrentIndex(++currentIndex);

    updateNavButtons();
}

void ManageConnectionDialog::back()
{
    // Additive functions, when back button was pushed
    ManageConnectionPage* currentPage = static_cast<ManageConnectionPage*>(pagesStackedWidget->currentWidget());
    currentPage->backEvents();

    int currentIndex = pagesStackedWidget->currentIndex();
    pagesStackedWidget->setCurrentIndex(--currentIndex);

    updateNavButtons();
}

void ManageConnectionDialog::eopAudioToggled(bool checked)
{
    m_pLocalResourcesPage->eopAudioToggled(checked);
}

void ManageConnectionDialog::closeEvent(QCloseEvent* event)
{
    if (m_pHelpWindow)
    {
        m_pHelpWindow->close();
    }
}

void ManageConnectionDialog::accept()
{
    saveFarmSettings();

    if (!isFarmLocationsCorrect(m_pCurrentFarm))
    {
        return;
    }

    // Save farm options in settings file
    if (m_isAddFarmMode)
    {
        m_pFarmSettings->addFarm(m_pCurrentFarm);
        m_deleteCurrentFarmOnClose = false;
    }
    else
    {
        *m_pFarmSettings->getFarm(m_currentFarmName) =  *m_pCurrentFarm;
    }
    m_pFarmSettings->saveFarms();

    MainWindow *mainWindowInstance = MainWindow::instance();
    mainWindowInstance->updateFarmsMenu();

    QDialog::accepted();
    hide();
}

void ManageConnectionDialog::cancel()
{
    QDialog::rejected();
    hide();
}

void ManageConnectionDialog::switchToBasicView()
{
    if ( m_pHelpWindow )
    {
        m_pHelpWindow->close();
    }

    SimpleAddFarmDialog* dialog = new SimpleAddFarmDialog( m_pFarmSettings, this->parentWidget() );

    close();

    dialog->exec();
}

void ManageConnectionDialog::updateNavButtons()
{
    // Also Enable/Disable the forward and back buttons as ncecessary
    int currentIndex = pagesStackedWidget->currentIndex();
    int count = pagesStackedWidget->count();
    if ( currentIndex < ( count ))
    {
        backButton->setEnabled(true);
        nextButton->setEnabled(true);
    }
    if ( currentIndex == ( count - 1 ))
    {
        // okButton->setEnabled(false);
        nextButton->setEnabled(false);
    }
    if (currentIndex == 0)
    {
        backButton->setEnabled(false);
        nextButton->setEnabled(true);
    }
}

void ManageConnectionDialog::currentPageChanged(int index)
{
    if (m_pHelpWindow && index >= 0)
    {
        moveHelpWindow();

        QString currentHelpString;
        if (index < m_helpString.size())
        {
            currentHelpString = m_helpString[index];
        }

        m_pHelpWindow->setStepText(tr("ManageConnectionDialog.Step ") + QString::number(index + 1));
        m_pHelpWindow->setHelpText(currentHelpString);
        m_pHelpWindow->setBackButtonVisible(index > 0);
        m_pHelpWindow->setNextButtonVisible(index < (pagesStackedWidget->count() - 1));
    }

    updateNavButtons();

}

void ManageConnectionDialog::moveHelpWindow()
{
    if (m_pHelpWindow)
    {
        const int tabHeight = 36;
        int helpWindowX = x() - m_pHelpWindow->width();
        int helpWindowY = y() + mainFrame->y() + pagesStackedWidget->y() + pagesStackedWidget->currentIndex() * tabHeight;
        m_pHelpWindow->move(helpWindowX, helpWindowY);
    }
}

// === FUNCTIONALITY

void ManageConnectionDialog::saveFarmSettings()
{
    if (m_pCurrentFarm == NULL)
    {
        return;
    }

    for (int i = 0; i < pagesStackedWidget->count(); ++i)
    {
        ManageConnectionPage* pageWidget = static_cast<ManageConnectionPage*>(pagesStackedWidget->widget(i));
        if (pageWidget != NULL)
        {
            pageWidget->saveFarmSettings();
        }
    }
}

void ManageConnectionDialog::loadFarmSettings()
{
    if (m_pCurrentFarm == NULL)
    {
        return;
    }

    for (int i = 0; i < pagesStackedWidget->count(); ++i)
    {
        ManageConnectionPage* pageWidget = static_cast<ManageConnectionPage*>(pagesStackedWidget->widget(i));
        if (pageWidget != NULL)
        {
            pageWidget->updateFarm(m_pCurrentFarm);
            pageWidget->loadFarmSettings();
        }
    }

    m_currentFarmName = m_pCurrentFarm->getName();
}

void ManageConnectionDialog::setFarmForAddDialog(const QString& name)
{
    // Initialize temp working farm

    // Create new farm with default name
    m_pCurrentFarm = m_pFarmSettings->createNewFarm();

    // Verify farm suggested name
    bool isFarmNameCorrect = m_pCurrentFarm->verifyName(name);
    if (isFarmNameCorrect)
    {
        m_pCurrentFarm->setName(name);
    }

    m_deleteCurrentFarmOnClose = true;
}

void ManageConnectionDialog::setFarmForEditDialog(const QString& name)
{
    // Initialize temp working farm

    m_pCurrentFarm = new Farm(m_pFarmSettings);
    m_deleteCurrentFarmOnClose = true;

    *m_pCurrentFarm = *m_pFarmSettings->getFarm(name);

    int locationsCount = m_pCurrentFarm->getLocationNamesList().count();
    for(int i = 0; i < locationsCount; ++i)
    {
        LocationSharedPointer loc(new Location);
        *loc = *m_pCurrentFarm->getLocation(0);
        m_pCurrentFarm->addLocation(loc);
        m_pCurrentFarm->removeLocation(0);
    }
}

void ManageConnectionDialog::setShowStepsNextTime(bool showNextTime)
{
    m_pFarmSettings->setValue(SHOW_STEPS, showNextTime);
}

void ManageConnectionDialog::setAllFarmsNotAutoConnected()
{
    QStringList farmList = m_pFarmSettings->getFarmNamesList();
    for (int i = 0; i < farmList.count(); ++i)
    {
        m_pFarmSettings->getFarm(i)->setAutoConnect(false);
    }
}

void ManageConnectionDialog::deleteFarm()
{
    m_pDeleteFarmDialog->setFarmName(m_pCurrentFarm->getName());
    if (m_pDeleteFarmDialog->exec() == QDialog::Accepted)
    {
        m_pFarmSettings->removeFarm(m_pFarmSettings->getFarm(m_pCurrentFarm->getName()));
        QDialog::accepted();
        hide();
    }
}

bool ManageConnectionDialog::isFarmLocationsCorrect(Farm* farm)
{
    QStringList locList = farm->getLocationNamesList();
    for (int i = 0; i < locList.count(); ++i)
    {
        LocationSharedPointer loc = farm->getLocation(i);
        if (loc->isSsl() && loc->getSslGateway().isEmpty())
        {
            QMessageBox::information(this, this->windowTitle(),
                                     tr("ManageConnectionDialog.SpecifySslGatewayFarm[%1]Location[%2]").
                                     arg(farm->getName()).arg(loc->getName()));
            return false;
        }
    }
    return true;
}
