// **************************************************************************************
// *
// * © 2012 Dell, Inc.
// * ALL RIGHTS RESERVED.
// *
// * Redistribution and use in source and binary forms, with or without modification, 
// * are permitted provided that the following conditions are met:
// * 
// *  *  Redistributions of source code must retain the above copyright notice, 
// *     this list of conditions and the following disclaimer.  
// * 
// *  *  Redistributions in binary form must reproduce the above copyright notice, 
// *     this list of conditions and the following disclaimer in the documentation 
// *     and/or other materials provided with the distribution.  
// * 
// *  *  Neither the name of Dell, Inc. nor the names of its contributors 
// *     may be used to endorse or promote products derived from this software 
// *     without specific prior written permission. 
// * 
// * 
// * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ON AN "AS IS" 
// * AND UNSUPPORTED BASIS, WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES OF ANY KIND, 
// * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, NON-
// * INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL THE COPYRIGHT 
// * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
// * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
// * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
// * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
// * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
// * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// * 
// **************************************************************************************

#include "simpleaddfarmdialog.h"
#include "ui_simpleaddfarmdialog.h"
#include "mainWindow.h"

SimpleAddFarmDialog::SimpleAddFarmDialog( FarmSettings* pFarmSettings, QWidget *parent) :
    CommonDialog(parent),
    ui(new Ui::SimpleAddFarmDialog)
{
    // Load the operating values
    m_pFarmSettings = pFarmSettings;

    // Build the new farm with default values
    m_pCurrentFarm = m_pFarmSettings->createNewFarm();

    // Build the GUI
    ui->setupUi(this);
    setWindowFlags(Qt::FramelessWindowHint | Qt::Dialog);
    setWindowMask( "", ui->frame->width(), ui->frame->height());

    // Load values (if any)
    // At this time, no value need to be loaded
    ui->brokerNameIPLineEdit->setStyleSheet( tr("/**/") );
    ui->passwordLineEdit->setStyleSheet( tr("/**/") );
    ui->userNameLineEdit->setStyleSheet( tr("/**/") );

}

SimpleAddFarmDialog::~SimpleAddFarmDialog()
{
    delete ui;
}


// Sanity checks

// Commit the new farm to the farm list

void SimpleAddFarmDialog::changeEvent(QEvent *e)
{
    QDialog::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}


bool SimpleAddFarmDialog::initializeFarm( void )
{
    QStringList domainList, serverList, locationList;
    LocationSharedPointer location;

    int locationIndex;

    // Validate some of the values
    if ( !validateValues())
    {
        // Complain to the user

        // Do NOT delete the current farm, as the user still has the
        // opportinity to correct the problems. Only delete the current farm
        // if the user cancels.

        return false;
    }

    // Set the values into the farm

    // Verify farm suggested name
    QString suggestedFarmName = ui->brokerNameIPLineEdit->text();
    bool isFarmNameCorrect = m_pCurrentFarm->verifyName(suggestedFarmName);
    if (isFarmNameCorrect)
    {
        m_pCurrentFarm->setName(suggestedFarmName);
    }
    m_pCurrentFarm->setUsername( ui->userNameLineEdit->text());
    m_pCurrentFarm->setPassword( ui->passwordLineEdit->text());
    m_pCurrentFarm->setCurrentDomain( ui->domainLineEdit->text());
    m_pCurrentFarm->setCredentialsSave( ui->saveCredentialsCheckBox->isChecked());

    // Also add the domain to the current domain list
    domainList = m_pCurrentFarm->getDomainList();
    domainList.append(ui->domainLineEdit->text() );
    m_pCurrentFarm->setDomainList( domainList );

    // Configure the first location as the default
    locationList = m_pCurrentFarm->getLocationNamesList();
    locationIndex = locationList.indexOf( QString("Location 1"), 0 );
    m_pCurrentFarm->setDefaultLocation( locationIndex );
    location = m_pCurrentFarm->getLocation(locationIndex );

    serverList = location->getServers();
    serverList.append( ui->brokerNameIPLineEdit->text( ));
    location->setServers( serverList );

    m_pCurrentFarm->setPromptLocation( false );

    return true;

}



bool SimpleAddFarmDialog::validateValues( void )
{
    bool rc = true;
    char errorBorderDef[1024] = "QLineEdit{ border: 1px solid rgba( 255,0,0,255); border-radius: 5; min-height: 1.5em; max-height: 1.5em }";

    ui->brokerNameIPLineEdit->setStyleSheet( tr("/**/") );
    ui->passwordLineEdit->setStyleSheet( tr("/**/") );
    ui->userNameLineEdit->setStyleSheet( tr("/**/") );

    // Make sure the name != empty()
    if ( ui->brokerNameIPLineEdit->text().isEmpty() )
    {
        ui->brokerNameIPLineEdit->setStyleSheet( QString( errorBorderDef ));
        syslog( LOG_WARNING, "SimpleAddFarmDialog: Attempted to add a farm with empty name");
        rc = false;
    }

    // Make sure the initial credentials are filled in.
    if ( ui->passwordLineEdit->text().isEmpty() )
    {
        ui->passwordLineEdit->setStyleSheet( QString( errorBorderDef ));
        syslog( LOG_WARNING, "SimpleAddFarmDialog: Attempted to add a farm with empty password");
        rc = false;
    }

    if ( ui->userNameLineEdit->text().isEmpty() )
    {
        ui->userNameLineEdit->setStyleSheet( QString( errorBorderDef ));
        syslog( LOG_WARNING, "SimpleAddFarmDialog: Attempted to add a farm with empty username");
        rc = false;
    }


    return rc;
}

/////////////////////////////////////////////////////
//
// Auto button connections
//
/////////////////////////////////////////////////////

void SimpleAddFarmDialog::on_autoConfigureButton_clicked()
{
    AutoConfigureDialog * dialog;
    QPushButton * pb = (QPushButton *) sender();
    pb->clearFocus();

    // If the current farm did not initialize currectly
    /*
    if ( initializeFarm() == false )
    {
        syslog( LOG_WARNING, "New Farm Initialization Failed. Autoconfigure cannot proceed.");

        return;
    }
    */

    dialog = new AutoConfigureDialog( m_pCurrentFarm , this );

    // LoadConfigurationDialog dialog;
    dialog->exec();

    if ( dialog->result() == QDialog::Accepted)
    {
        m_pFarmSettings->addFarm( m_pCurrentFarm );

        // Update the farm list
        MainWindow *mainWindowInstance = MainWindow::instance();
        mainWindowInstance->updateFarmsMenu();
    }
    else if ( dialog->result() == 2 )
    {
        syslog( LOG_INFO, "Autoconfig dialog 'Back' button selected.");
    }
    else hide();
}

void SimpleAddFarmDialog::on_connectButton_clicked()
{
    Farm * pFarm;

    int number = m_pFarmSettings->getFarmNamesList().count();

    // If the current farm did not initialize currectly
    if ( initializeFarm() == false )
    {
        syslog( LOG_WARNING, "New Farm Initialization Failed. Connection cannot proceed.");

        return;
    }

    // Attempt to connect and get the resulting Farm Name as reported by the Farm.
    // For the moment, default to the IP address or host name

    // Add the farm to the main list the farm
    // FIXME: Error handling please.
    m_pCurrentFarm->setCurrentLocation( 0 );

    // NOTE: CANNOT make settings changes to m_pCurrentFarm after this line
    //       because it will not be the same object
    m_pFarmSettings->addFarm( m_pCurrentFarm );

    // Load the farm
    pFarm = m_pFarmSettings->getFarm( number );
    MainWindow *mainWindowInstance = MainWindow::instance();
    hide();
    mainWindowInstance->runFarm( pFarm->getName() );

    QFile ifile( "/tmp/farminfo");
    if (!ifile.open(QIODevice::ReadOnly | QIODevice::Text))
    {
        m_pFarmSettings->removeFarm( pFarm );
        mainWindowInstance->updateFarmsMenu();
        return;
    }
    ifile.remove();

    // Then save the farms
    pFarm->saveFarm();

    // Then update the menu
    mainWindowInstance->updateFarmsMenu();

    // Close out the dialog
    hide();
}

void SimpleAddFarmDialog::on_cancelButton_clicked()
{
    this->hide();
}

void SimpleAddFarmDialog::on_advancedViewButton_clicked()
{
    // TODO: Sanity check for the farm name.

    // Close me, and then show the full settings window.

    ManageConnectionDialog* connectionDialog = new ManageConnectionDialog( m_pFarmSettings,
                                                                           true,
                                                                           ui->brokerNameIPLineEdit->text(),
                                                                           this->parentWidget() );
    close();

    connectionDialog->exec();
}
