// **************************************************************************************
// *
// * © 2012 Dell, Inc.
// * ALL RIGHTS RESERVED.
// *
// * Redistribution and use in source and binary forms, with or without modification, 
// * are permitted provided that the following conditions are met:
// * 
// *  *  Redistributions of source code must retain the above copyright notice, 
// *     this list of conditions and the following disclaimer.  
// * 
// *  *  Redistributions in binary form must reproduce the above copyright notice, 
// *     this list of conditions and the following disclaimer in the documentation 
// *     and/or other materials provided with the distribution.  
// * 
// *  *  Neither the name of Dell, Inc. nor the names of its contributors 
// *     may be used to endorse or promote products derived from this software 
// *     without specific prior written permission. 
// * 
// * 
// * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ON AN "AS IS" 
// * AND UNSUPPORTED BASIS, WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES OF ANY KIND, 
// * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, NON-
// * INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL THE COPYRIGHT 
// * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
// * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
// * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
// * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
// * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
// * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// * 
// **************************************************************************************

#include "appPortalConfigCollector.h"
#include "syslog.h"

#include <QNetworkAccessManager>
#include <QFileInfo>
#include <QFile>
#include <QUrl>
#include <QNetworkProxy>
#include <QNetworkRequest>

AppPortalConfigCollector::AppPortalConfigCollector(QObject* parent)
    : QObject(parent)
{
    // Network manager initialization
    m_pNetworkManager = new QNetworkAccessManager(this);
    connect(m_pNetworkManager, SIGNAL(finished(QNetworkReply*)), this, SLOT(replyFinished(QNetworkReply*)));
}

AppPortalConfigCollector::~AppPortalConfigCollector()
{
}

ConfigResult AppPortalConfigCollector::loadConfigFromFile(QString dir, QString fileName)
{
    // Checking parameters
    if (dir.isEmpty())
        return CONFIG_DIR_IS_EMPTY;
    if (fileName.isEmpty())
        return CONFIG_FILE_NAME_IS_EMPTY;

    // Loading the file from disk
    QString protocol("file");
    QString absName = makePathToConfig(protocol, dir, fileName);

    syslog( LOG_INFO, "Attempting to load config file: %s", absName.toAscii().data());

    // If the file doesn't exist
    QFileInfo fileInfo(absName);
    if (!fileInfo.exists())
    {
        return CONFIG_FILE_DOES_NOT_EXIST;
    }

    // If we cannot to open this file
    QFile file(absName);
    if (!file.open(QIODevice::ReadOnly))
    {
        return CONFIG_FILE_COULD_NOT_BE_READ;
    }

    SharedBuffer sharedBuffer = SharedBuffer(new QByteArray);
    *sharedBuffer = file.readAll();
    file.close();
    m_configMap[absName] = sharedBuffer;
    return CONFIG_OK;

}

ConfigResult AppPortalConfigCollector::loadConfigFromNetwork(QString protocol, QString url, QString fileName,
                                                             bool isProxy, QString proxyList)
{
    // Checking parameters
    if (url.isEmpty())
        return CONFIG_URL_IS_EMPTY;
    if (fileName.isEmpty())
        return CONFIG_FILE_NAME_IS_EMPTY;
    protocol = protocol.toLower();

    // Getting url
    QUrl fullUrl(makePathToConfig(protocol, url, fileName));

    // Getting proxy
    if (isProxy)
    {
        // Getting proxyserver and proxyport
        int index = proxyList.lastIndexOf(':');
        QString proxyServer = proxyList.mid(0, index);
        QString proxyPort = proxyList.mid(index, proxyList.length() - index - 1);

        // Converting string port to int port
        bool okPortToInt;
        int port = proxyPort.toInt(&okPortToInt);
        if (!okPortToInt)
        {
            return CONFIG_WRONG_PROXY_PORT;
        }

        // Setting proxy to connection
        QNetworkProxy proxy;
        proxy.setPort(port);
        proxy.setHostName(proxyServer);

        m_pNetworkManager->setProxy(proxy);
    }

    m_pNetworkManager->get(QNetworkRequest(fullUrl));
    return CONFIG_OK;
}

void AppPortalConfigCollector::replyFinished(QNetworkReply* reply)
{
    QString urlString = reply->request().url().toString();
    if (reply->error() == QNetworkReply::NoError)
    {
        // Reading data
        SharedBuffer sharedBuffer = SharedBuffer(new QByteArray);
        *sharedBuffer = reply->readAll();
        m_configMap[urlString] = sharedBuffer;
    }
    // emitting signal for caller of loadConfigFromNetwork(...)
    emit networkReply(reply->error(), reply->errorString(), urlString);
    reply->deleteLater();
}

SharedBuffer AppPortalConfigCollector::getConfigBuffer(QString protocol, QString url, QString fileName) const
{
    QString fullUrl = makePathToConfig(protocol, url, fileName);
    return getConfigBuffer(fullUrl);
}

SharedBuffer AppPortalConfigCollector::getConfigBuffer(QString fullUrl) const
{
    ConfigMapConstIterator mapIter = m_configMap.find(fullUrl);
    if (mapIter != m_configMap.end())
    {
        return mapIter.value();
    }
    else
    {
        return SharedBuffer();
    }
}

bool AppPortalConfigCollector::deleteConfig(QString protocol, QString url, QString fileName)
{
    QString fullUrl = makePathToConfig(protocol, url, fileName);
    return deleteConfig(fullUrl);
}

bool AppPortalConfigCollector::deleteConfig(QString fullUrl)
{
    ConfigMapIterator mapIter = m_configMap.find(fullUrl);
    if (mapIter != m_configMap.end())
    {
        m_configMap.erase(mapIter);
        return true;
    }
    return false;
}

void AppPortalConfigCollector::deleteAll()
{
    m_configMap.clear();
}

QString AppPortalConfigCollector::makePathToConfig(QString& protocol, QString& url, QString& fileName) const
{
    protocol = protocol.toLower();
    if (protocol == "file" || protocol == "")
    {
        return url + "/" + fileName;
    }
    else
    {
        return protocol + "://" + url + "/" + fileName;
    }
}
