// **************************************************************************************
// *
// * © 2012 Dell, Inc.
// * ALL RIGHTS RESERVED.
// *
// * Redistribution and use in source and binary forms, with or without modification, 
// * are permitted provided that the following conditions are met:
// * 
// *  *  Redistributions of source code must retain the above copyright notice, 
// *     this list of conditions and the following disclaimer.  
// * 
// *  *  Redistributions in binary form must reproduce the above copyright notice, 
// *     this list of conditions and the following disclaimer in the documentation 
// *     and/or other materials provided with the distribution.  
// * 
// *  *  Neither the name of Dell, Inc. nor the names of its contributors 
// *     may be used to endorse or promote products derived from this software 
// *     without specific prior written permission. 
// * 
// * 
// * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ON AN "AS IS" 
// * AND UNSUPPORTED BASIS, WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES OF ANY KIND, 
// * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, NON-
// * INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL THE COPYRIGHT 
// * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
// * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
// * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
// * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
// * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
// * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// * 
// **************************************************************************************

#ifndef APPLICATION_SETTINGS_H
#define APPLICATION_SETTINGS_H

/**************************************

This class intended for reading global application settings from specified *.ini file using QT's QSetting class.
*.ini file example:

[Group example]
name=group example
location=header file
id=57

[Array example]
1=258
2=74
3=5454
size=3

singleValueExample=100

***************************************/

#include <QStringList>
#include <QVariant>
#include <QList>

#include <QSettings>

class ApplicationSettings
{
public:
    // The constructor takes a file name in which settings are stored
    ApplicationSettings(const QString fileName);
    ~ApplicationSettings();

    // This two functions just get QVariant value and QVariant values list with specified key respectively
    //without converting it to other types.
    // getValue(...) returns true if value with specified key exists and false otherwise.
    // getValueList(...) returns elements count in array with specified key or 0 if array with this key does not exist.
    // In case of success settings from file writes in value and valueList paremeters respectively.
    bool getQvariantValue    (const QString& key, QVariant* value);
    int  getQvariantValueList(const QString& key, QList<QVariant>* valueList);

    // This two template functions do the same as previous two functions and additionaly
    //convert received values to ValueType.
    // So resulting value and valueList parameters contain only values that can be converted in needed type.
    template <typename ValueType> bool getValue    (const QString& key, ValueType* value);
    template <typename ValueType> int  getValueList(const QString& key, QList<ValueType>* valueList);

    // This function appends prefix to the current group
    void beginGroup(const QString& key);
    // This function resets group to what in was before corresponding of beginGroup(...) call
    void endGroup();

private:
    QSettings m_settings;
};


template <typename ValueType>
bool ApplicationSettings::getValue(const QString& key, ValueType* value)
{
    QVariant variantValue;
    if (getQvariantValue(key, &variantValue))
    {
        bool castResult;
        QVariant::Type type = static_cast<QVariant::Type>(qMetaTypeId<ValueType>());
        castResult = variantValue.convert(type);
        if (castResult)
        {
            *value = variantValue.value<ValueType>();
        }
        return castResult;
    }
    return false;
}

template <typename ValueType>
int ApplicationSettings::getValueList(const QString& key, QList<ValueType>* valueList)
{
    QList<QVariant> variantValueList;
    int variantValueCount;
    variantValueCount = getQvariantValueList(key, &variantValueList);

    QVariant::Type type = static_cast<QVariant::Type>(qMetaTypeId<ValueType>());
    valueList->clear();
    for (int i = 0; i < variantValueCount; ++i)
    {
        bool castResult;
        castResult = variantValueList[i].convert(type);
        if (castResult)
        {
            valueList->append(variantValueList.at(i).value<ValueType>());
        }
    }
    return valueList->size();
}

#endif
