// **************************************************************************************
// *
// * © 2012 Dell, Inc.
// * ALL RIGHTS RESERVED.
// *
// * Redistribution and use in source and binary forms, with or without modification, 
// * are permitted provided that the following conditions are met:
// * 
// *  *  Redistributions of source code must retain the above copyright notice, 
// *     this list of conditions and the following disclaimer.  
// * 
// *  *  Redistributions in binary form must reproduce the above copyright notice, 
// *     this list of conditions and the following disclaimer in the documentation 
// *     and/or other materials provided with the distribution.  
// * 
// *  *  Neither the name of Dell, Inc. nor the names of its contributors 
// *     may be used to endorse or promote products derived from this software 
// *     without specific prior written permission. 
// * 
// * 
// * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ON AN "AS IS" 
// * AND UNSUPPORTED BASIS, WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES OF ANY KIND, 
// * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, NON-
// * INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL THE COPYRIGHT 
// * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
// * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
// * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
// * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
// * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
// * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// * 
// **************************************************************************************

#include "farm.h"
#include "sublib/device.h"
#include "sublib/code.h"
#include "farmSettings.h"
#include <QDomDocument>
#include <QDomElement>
#include <QtDebug>
#include "common/common.h"
#include "farmSettingsNames.h"
#include <syslog.h>


Farm::Farm(QSettings* s, QString key, const QString& name):
    iDefault(0)
{
    this->settings = s;
    this->key = key;
    this->name = name;

    for (int i = 0; i < 1; i++)
    {
        LocationSharedPointer location(new Location);
        location->setName("Location " + QString::number(i + 1));
        m_locations << location;
    }

    m_appSettings = SharedApplicationSettings(new ApplicationSettings(APPLICATION_CONFIG_FILE_NAME));

    // Default parameters
    setAllDefaultValues();

    // Override parameters if exist
    setAllOverrideValues();
}

Farm::~Farm()
{
}


void Farm::setAllDefaultValues()
{
    m_appSettings->beginGroup(FARM_DEFAULT_VALUES);

    setSoundMode( QString("local"));

    setDefaultValue(KERBEROS_MODE, &kerberosMode, false);
    setDefaultValue(KERBEROS_ENABLED, &kerberosEnabled, false);
    setDefaultValue(SSO_ENABLED, &ssoEnabled, false);
    setDefaultValue(ALLOW_SAVE_CREDENTIALS, &saveCredentials, false);

    setDefaultValue(EXPERIENCE_LEVEL, &experienceLevel, 3);
    setDefaultValue(ENABLE_WALLPAPER, &wallpaperEnabled, true);
    setDefaultValue(ENABLE_FULL_WINDOW_DRAG, &draggingEnabled, true);
    setDefaultValue(ENABLE_ANIMATION, &animationEnabled, true);
    setDefaultValue(ENABLE_THEMES, &themesEnabled, true);
    setDefaultValue(ENABLE_BITMAP_CACHING, &bitmapCachingEnabled, true);
    setDefaultValue(ENABLE_DNA, &dnaEnabled, false);
    setDefaultValue(MMR, &mmrEnabled, false);
    setDefaultValue(FMMR, &fmmrEnabled, false);
    setDefaultValue(GRAPHICS_ACCELERATION, &gaEnabled, false);
    setDefaultValue(RECONNECT, &reconnect, false);

    setDefaultValue(DESKTOP_WIDTH, &desktopWidth, 0);
    setDefaultValue(DESKTOP_HEIGHT, &desktopHeight, 0);
    setDefaultValue(COLOR_DEPTH, &colorDepth, 16);
    setDefaultValue(DISPLAY_CONNECTION_BAR, &connectionBar, false);
    setDefaultValue(PIN_CONNECTION_BAR, &pinConnectionBar, false);
    setDefaultValue(SPAN_MONITORS, &spanMonitors, false);
    setDefaultValue(SMART_SIZING, &smartSizing, false);
    setDefaultValue(SEAMLESS_MODE, &seamlessMode, false);

    setDefaultValue(REDIRECT_CLIPBOARD, &redirectClipboard, false);
    setDefaultValue(REDIRECT_DRIVES, &redirectDrives, false);
    setDefaultValue(REDIRECT_PRINTERS, &redirectPrinters, false);
    setDefaultValue(REDIRECT_UNIVERSAL_PRINTERS, &redirectUPrinters, false);
    setDefaultValue(REDIRECT_COM_PORTS, &redirectComPorts, false);
    setDefaultValue(REDIRECT_HANDHELDS, &redirectHandhelds, false);
    setDefaultValue(REDIRECT_MICROPHONE, &redirectMicrophone, false);
    setDefaultValue(REDIRECT_SMARTCARDS, &redirectSmartCards, false);

    setDefaultValue(ENABLE_DOWNSTREAM_AUDIO, &downstreamAudioEnabled, false);
    setDefaultValue(DA_MODE, &daMode, QString("Local"));
    setDefaultValue(DA_QUALITY, &daQuality, QString("cd"));
    setDefaultValue(DA_BUFFERING, &daBuffering, 0);
    setDefaultValue(DA_CUSTOM_QUALITY, &daCustomQuality, 0);

    setDefaultValue(ENABLE_UPSTREAM_AUDIO, &upstreamAudioEnabled, false);
    setDefaultValue(UA_MODE, &uaMode, QString("wave"));
    setDefaultValue(UA_QUALITY, &uaQuality, 0);
    setDefaultValue(UA_COMPLEXITY, &uaComplexity, 0);
    setDefaultValue(UA_ENHANCED_COMPRESSION, &uaEnhancedCompression, false);

    setDefaultValue(HIDE_SETTINGS, &hideSettings, false);
    setDefaultValue(PASSWORD_MANAGEMENT_PORT, &pmPort, 443);
    setDefaultValue(ALLOW_PASSWORD_MANAGEMENT, &allowPasswordManagement, true);

    setDefaultValue(LAST_LOCATION, &iLast, 0);
    setDefaultValue(DEFAULT_LOCATION, &iDefault, 0);
    setDefaultValue(PROMPT_FOR_LOCATION, &promptLocation, true);
    m_currentLocation.clear();

    setDefaultValue(NEED_RECONNECT, &needReconnect, false);
    setDefaultValue(AUTO_CONNECT, &autoConnect, false);
    setDefaultValue(SHOW_STEPS, &showHelpSteps, false);
    setDefaultValue(SHOW_CLICK_HERE, &showClickHere, true);

    m_appSettings->endGroup();
}

void Farm::setAllOverrideValues()
{
    setLastLocation(iLast);
    setDefaultLocation(iDefault);
    setPromptLocation(promptLocation);
    setUsername(username);
    setPassword(password);
    setDomainList(domainList);
    setCurrentDomain(currentDomain);
    setCredentialsSave(saveCredentials);
    setKerberosMode(kerberosMode);
    setKerberosEnable(kerberosEnabled);
    setSsoEnable(ssoEnabled);
    setWaitTime( waitTime);
    setNumberOfRetries(numberOfRetries);
    setWaitForDesktop(waitForDesktopIsEnabled);

    setExperienceLevel(experienceLevel);
    setWallpaperEnable(wallpaperEnabled);
    setDraggingEnable(draggingEnabled);
    setAnimationEnable(animationEnabled);
    setThemesEnable(themesEnabled);
    setBitmapCachingEnable(bitmapCachingEnabled);
    setMmrEnable(mmrEnabled);
    setFmmrEnable(fmmrEnabled);
    setGaEnable(gaEnabled);
    setDnaEnabled( dnaEnabled);
    setReconnection(reconnect);

    setDesktopWidth(desktopWidth);
    setDesktopHeight(desktopHeight);
    setColorDepth(colorDepth);
    setConnectionBar(connectionBar);
    setPinConnectionBar(pinConnectionBar);
    setSpanMonitors(spanMonitors);
    setSmartSizing(smartSizing);
    setSeamlessMode(seamlessMode);
    setRedirectClipboard(redirectClipboard);
    setRedirectDrives(redirectDrives);
    setRedirectPrinters(redirectPrinters);
    setRedirectUPrinters(redirectUPrinters);
    setRedirectComPorts(redirectComPorts);
    setRedirectHandhelds(redirectHandhelds);
    setRedirectMicrophone(redirectMicrophone);
    setRedirectSmartCards(redirectSmartCards);
    setSoundMode(soundMode);
    setDownstreamAudioEnabled(downstreamAudioEnabled);
    setDAMode(daMode);
    setDAQuality(daQuality);
    setDABuffering(daBuffering);
    setDACustomQuality(daCustomQuality);
    setUpstreamAudioEnabled(upstreamAudioEnabled);
    setUAMode(uaMode);
    setUAQuality(uaQuality);
    setUAComplexity(uaComplexity);
    setUAEnhancedCompression(uaEnhancedCompression);
    setHideSettings(hideSettings);
    setPasswordManagementServer(pmServer);
    setPasswordManagementPort(pmPort);
    setPasswordManagementAllow(allowPasswordManagement);
    setSourceProtocol(sourceProtocol);
    setSourceServer(sourceServer);
    setSourceFile(sourceFile);
    setSourceProxy(sourceProxy);
    setNeedsReconnecting(needReconnect);
    setAutoConnect(autoConnect);
    setShowSteps(showHelpSteps);
    setShowClickHere( showClickHere );
}

void Farm::setName(const QString& newName)
{
    name = newName;
}

bool Farm::verifyName(const QString& newName)
{
    if ( (newName.isEmpty()) || (newName.trimmed().isEmpty()) )
        return false;

    // Get parent settings
    FarmSettings *farmSettings = (FarmSettings *) settings;
    if (farmSettings == 0)
        return true;

    QStringList allNamesList = farmSettings->getFarmNamesList();
    foreach (QString name, allNamesList)
    {
        if (name == newName)
        {
            Farm *farm = farmSettings->getFarm(name);
            if (this->getKey() != farm->getKey())
                return false;
        }
    }

    return true;
}

QString Farm::getName() const
{
    return this->name;
}

void Farm::setLastLocation(const int number)
{
    setValue(LAST_LOCATION, &iLast, number);
}

void Farm::setDefaultLocation(const int number)
{
    setValue(DEFAULT_LOCATION, &iDefault, number);
}

void Farm::setPromptLocation(const bool prompt)
{
    setValue(PROMPT_FOR_LOCATION, &promptLocation, prompt);
}

void Farm::setUsername(const QString& name)
{
    setValue(USERNAME, &username, name);
}

void Farm::setPassword(const QString& pass)
{
    setValue(PASSWORD, &password, pass);
}

void Farm::setDomainList(const QStringList& dmns)
{
    setValueList(DOMAIN_LIST, &domainList, dmns);
}

void Farm::setCurrentDomain(const QString& dmn)
{
    setValue(DOMAIN, &currentDomain, dmn);
}

void Farm::setCredentialsSave(const bool save)
{
    setValue(ALLOW_SAVE_CREDENTIALS, &saveCredentials, save);
}

void Farm::setKerberosMode(const bool km)
{
    setValue(KERBEROS_MODE, &kerberosMode, km);
}

void Farm::setKerberosEnable(const bool mode)
{
    setValue(KERBEROS_ENABLED, &kerberosEnabled, mode);
}

void Farm::setSsoEnable(const bool mode)
{
    setValue(SSO_ENABLED, &ssoEnabled, mode);
}

void Farm::setWaitTime(const int wait)
{
    setValue(WAIT_TIME, &waitTime, wait);
}

void Farm::setNumberOfRetries(const int retry)
{
    setValue(NUMBER_OF_RETRIES, &numberOfRetries, retry);
}

void Farm::setWaitForDesktop(const bool enabled)
{
    setValue(WAIT_FOR_DESKTOP_ENABLED, &waitForDesktopIsEnabled, enabled);
}

void Farm::setWallpaperEnable(const bool enabled)
{
    setValue(ENABLE_WALLPAPER, &wallpaperEnabled, enabled);
}

void Farm::setDraggingEnable(const bool enabled)
{
    setValue(ENABLE_FULL_WINDOW_DRAG, &draggingEnabled, enabled);
}

void Farm::setAnimationEnable(const bool enabled)
{
    setValue(ENABLE_ANIMATION, &animationEnabled, enabled);
}

void Farm::setThemesEnable(const bool enabled)
{
    setValue(ENABLE_THEMES, &themesEnabled, enabled);
}

void Farm::setBitmapCachingEnable(const bool enabled)
{
    setValue(ENABLE_BITMAP_CACHING, &bitmapCachingEnabled, enabled);
}

void Farm::setMmrEnable(const bool enabled)
{
    setValue(MMR, &mmrEnabled, enabled);
}

void Farm::setFmmrEnable(const bool enabled)
{
    setValue(FMMR, &fmmrEnabled, enabled);
}

void Farm::setGaEnable(const bool enabled)
{
    setValue(GRAPHICS_ACCELERATION, &gaEnabled, enabled);
}

void Farm::setDnaEnabled(const bool enabled)
{
    setValue(ENABLE_DNA, &dnaEnabled, enabled);
}

void Farm::setReconnection(const bool b)
{
    setValue(RECONNECT, &reconnect, b);
}

void Farm::setDesktopWidth(const int width)
{
    setValue(DESKTOP_WIDTH, &desktopWidth, width);
}

void Farm::setDesktopHeight(const int height)
{
    setValue(DESKTOP_HEIGHT, &desktopHeight, height);
}

void Farm::setColorDepth(const int depth)
{
    setValue(COLOR_DEPTH, &colorDepth, depth);
}

void Farm::setConnectionBar(const bool mode)
{
    setValue(DISPLAY_CONNECTION_BAR, &connectionBar, mode);
}

void Farm::setPinConnectionBar(const bool mode)
{
    setValue(PIN_CONNECTION_BAR, &pinConnectionBar, mode);
}

void Farm::setSpanMonitors(const bool mode)
{
    setValue(SPAN_MONITORS, &spanMonitors, mode);
}

void Farm::setSmartSizing(const bool mode)
{
    setValue(SMART_SIZING, &smartSizing, mode);
}

void Farm::setSeamlessMode(const bool smode)
{
    setValue(SEAMLESS_MODE, &seamlessMode, smode);
}

void Farm::setRedirectClipboard(const bool rc)
{
    setValue(REDIRECT_CLIPBOARD, &redirectClipboard, rc);
}

void Farm::setRedirectDrives(const bool rd)
{
    setValue(REDIRECT_DRIVES, &redirectDrives, rd);
}

void Farm::setRedirectPrinters(const bool rp)
{
    setValue(REDIRECT_PRINTERS, &redirectPrinters, rp);
}

void Farm::setRedirectUPrinters(const bool rp)
{
    setValue(REDIRECT_UNIVERSAL_PRINTERS, &redirectUPrinters, rp);
}

void Farm::setRedirectComPorts(const bool rc)
{
    setValue(REDIRECT_COM_PORTS, &redirectComPorts, rc);
}

void Farm::setRedirectHandhelds(const bool rh)
{
    setValue(REDIRECT_HANDHELDS, &redirectHandhelds, rh);
}

void Farm::setRedirectMicrophone(const bool rm)
{
    setValue(REDIRECT_MICROPHONE, &redirectMicrophone, rm);
}

void Farm::setRedirectSmartCards(const bool rs)
{
    setValue(REDIRECT_SMARTCARDS, &redirectSmartCards, rs);
}

void Farm::setSoundMode(const QString& mode)
{
    setValue(REDIRECT_AUDIO, &soundMode, mode);
}

void Farm::setDownstreamAudioEnabled(const bool daEnabled)
{
    setValue(ENABLE_DOWNSTREAM_AUDIO, &downstreamAudioEnabled, daEnabled);
}

void Farm::setDAMode(const QString& mode)
{
    setValue(DA_MODE, &daMode, mode);
}

void Farm::setDAQuality(const QString& quality)
{
    setValue(DA_QUALITY, &daQuality, quality);
}

void Farm::setDABuffering(const int buf)
{
    setValue(DA_BUFFERING, &daBuffering, buf);
}

void Farm::setDACustomQuality(const int cq)
{
    setValue(DA_CUSTOM_QUALITY, &daCustomQuality, cq);
}

void Farm::setUpstreamAudioEnabled(const bool uaEnabled)
{
    setValue(ENABLE_UPSTREAM_AUDIO, &upstreamAudioEnabled, uaEnabled);
}

void Farm::setUAMode(const QString& mode)
{
    setValue(UA_MODE, &uaMode, mode);
}

void Farm::setUAQuality(const int quality)
{
    setValue(UA_QUALITY, &uaQuality, quality);
}

void Farm::setUAComplexity(const int complexity)
{
    setValue(UA_COMPLEXITY, &uaComplexity, complexity);
}

void Farm::setUAEnhancedCompression(const bool ec)
{
    setValue(UA_ENHANCED_COMPRESSION, &uaEnhancedCompression, ec);
}

void Farm::setHideSettings(const bool hide)
{
    setValue(HIDE_SETTINGS, &hideSettings, hide);
}

void Farm::setPasswordManagementServer(const QString& server)
{
    setValue(PASSWORD_MANAGEMENT_SERVER, &pmServer, server);
}

void Farm::setPasswordManagementPort(const int port)
{
    setValue(PASSWORD_MANAGEMENT_PORT, &pmPort, port);
}

void Farm::setPasswordManagementAllow(const bool allow)
{
    setValue(ALLOW_PASSWORD_MANAGEMENT, &allowPasswordManagement, allow);
}

void Farm::setSourceProtocol(const QString& protocol)
{
    setValue(SOURCE_PROTOCOL, &sourceProtocol, protocol);
}

void Farm::setSourceServer(const QString& server)
{
    setValue(SOURCE_SERVER, &sourceServer, server);
}

void Farm::setSourceFile(const QString& file)
{
    setValue(SOURCE_FILE, &sourceFile, file);
}

void Farm::setSourceProxy(const QString& proxy)
{
    setValue(SOURCE_PROXY, &sourceProxy, proxy);
}

void Farm::setNeedsReconnecting(const bool need)
{
    setValue(NEED_RECONNECT, &needReconnect, need);
}

void Farm::setAutoConnect(const bool val)
{
    setValue(AUTO_CONNECT, &autoConnect, val);
}

void Farm::setShowSteps(const bool val)
{
    setValue(SHOW_STEPS, &showHelpSteps, val);
}

void Farm::setShowClickHere(const bool val)
{
    setValue(SHOW_CLICK_HERE, &showClickHere, val);
}

void Farm::setExperienceLevel(const int val)
{
    setValue(EXPERIENCE_LEVEL, &experienceLevel, val);
}


FileSharedPointer Farm::createPitFile()
{
    QSharedPointer<QTemporaryFile> pitFile(new QTemporaryFile);

    if (!pitFile->open())
    {
        return FileSharedPointer();
    }

    // Write farm settings in pit-file
    QTextStream stream(pitFile.data());
    stream << "username:s:" << getUsername() << endl
           << "password:s:" << getPassword() << endl
           << "domain:s:" << getCurrentDomain() << endl;

    // Location settings
    if (m_currentLocation != 0)
    {
        stream << "llmservers:s:" << m_currentLocation->getServers().join(",") << endl
               << "llmtcpport:s:" << m_currentLocation->getPort() << endl
               << "llmenablenat:i:" << (int) m_currentLocation->isNat() << endl
               << "enablessl:i:" << (int) m_currentLocation->isSsl() << endl
               << "sslgateway:s:" << m_currentLocation->getSslGateway() << endl;

        int protocol;
        if (m_currentLocation->getProtocol() == "http")
        {
            protocol = 0;
        }
        else
        {
            protocol = 1;
        }
        stream << "llmprotocol:i:" << protocol << endl;
    }

    if (waitForDesktop())
    {
        stream << "wait:i:" << (int) getWaitTime() << endl
               << "retry:i:" << (int) getNumberOfRetries() << endl;
    }

    // User experience options
    stream << "disable wallpaper:i:" << (int) (isWallpaperEnabled() == false) << endl
           << "disable full window drag:i:" << (int) (isDraggingEnabled() == false) << endl
           << "disable menu anims:i:" << (int) (isAnimationEnabled() == false) << endl
           << "disable themes:i:" << (int) (isThemesEnabled() == false) << endl
           << "bitmapcachepersistenable:i:" << (int) isBitmapCachingEnabled() << endl
           << "reconnect:i:" << (int) isReconnection() << endl;

    int width = getDesktopWidth();
    int height = getDesktopHeight();
    QString desktopSize;
    if (width && height)
    {
        desktopSize = QString::number(width) + "x" + QString::number(height);
    }
    else
    {
        desktopSize = "fullscreen";
    }

    // Display and sound options
    stream << "windowSize:s:" << desktopSize << endl
           << "session bpp:s:" << getColorDepth() << endl
           << "sound:s:" << getSoundMode() << endl;

    // Seamless mode
    stream << "seamless:i:" << (int) isSeamlessMode() << endl;

    // Multi-monitor feature
    stream << "spanmonitors:i:" << (int) isSpanMonitors() << endl;


    // Redirection options
    DeviceDirector deviceDirector;
    if (this->isRedirectDrives())
    {
        QStringList drives = deviceDirector.getMountDrives();
        stream << "drives:s:" << drives.join(",") << endl;
    }

    if (this->isRedirectComPorts())
    {
        QStringList comPorts = deviceDirector.getComPorts();
        stream << "comports:s:" << comPorts.join(",") << endl;
    }

    if (isRedirectPrinters())
    {
        stream << "redirectprinters:i:1" << endl;
    }

    stream << "downstreamAudioEnabled:i:" << (int) this->isDownstreamAudioEnabled() << endl;
    stream << "daMode:s:" << this->getDAMode() << endl;
    stream << "daQuality:s:" << this->getDAQuality() << endl;
    stream << "daBuffering:s:" << (int) this->getDABuffering() << endl;
    stream << "daCustomQuality:s:" << (int) this->getDACustomQuality() << endl;

    stream << "upstreamAudioEnabled:i:" << (int) this->isUpstreamAudioEnabled() << endl;
    stream << "uaMode:s:" << this->getUAMode() << endl;
    stream << "uaQuality:s:" << (int) this->getUAQuality() << endl;
    stream << "uaComplexity:s:" << (int) this->getUAComplexity() << endl;
    stream << "uaEnhancedCompression:i:" << (int) this->getUAEnhancedCompression() << endl;

    // Other redirections
    stream << "redirectclipboard:i:" << (int) this->isRedirectClipboard() << endl;

    // USB redirection handhelds
    stream << "redirecthandhelds:i:" << (int) this->isRedirectHandhelds() << endl;

    // Smartcard redirection
    stream << "redirectsmartcards:i:" << (int) this->isRedirectSmartCards() << endl;

    // MMR Redirect
    stream << "multimediaredirection:i:" << (int) this->isMmrEnabled() << endl;

    // FMMR Redirect
    stream << "flashredirection:i:" << (int) this->isFmmrEnabled() << endl;

    // GA Enable
    stream << "bitmapacceleration:i:" << (int) this->isGaEnabled() << endl;

    // DNA enabled
    stream << "dnaproxy:i:" << (int) this->isDnaEnabled() << endl;

    pitFile->close();
    return pitFile;
}

void Farm::saveFarm()
{
    settings->beginGroup(key);

    settings->setValue("Name", name);
    if (saveCredentials)
    {
        settings->setValue(USERNAME, username);
        settings->setValue(PASSWORD, encryptString256(password));
        settings->setValue(DOMAIN, currentDomain);
    }
    domainList.removeAll("");
    settings->setValue(DOMAIN_LIST, domainList);
    settings->setValue(ALLOW_SAVE_CREDENTIALS, saveCredentials);
    settings->setValue(KERBEROS_MODE, kerberosMode);
    settings->setValue(KERBEROS_ENABLED, kerberosEnabled);
    settings->setValue(SSO_ENABLED, ssoEnabled);

    settings->setValue(DEFAULT_LOCATION, iDefault);
    settings->setValue(PROMPT_FOR_LOCATION, promptLocation);
    //settings->setValue("LastLocation", iLast);

    settings->setValue(WAIT_TIME, waitTime);
    settings->setValue(NUMBER_OF_RETRIES, numberOfRetries);
    settings->setValue(WAIT_FOR_DESKTOP_ENABLED, waitForDesktopIsEnabled);

    settings->setValue(EXPERIENCE_LEVEL, experienceLevel );
    settings->setValue(ENABLE_WALLPAPER, wallpaperEnabled);
    settings->setValue(ENABLE_FULL_WINDOW_DRAG, draggingEnabled);
    settings->setValue(ENABLE_ANIMATION, animationEnabled);
    settings->setValue(ENABLE_THEMES, themesEnabled);
    settings->setValue(ENABLE_BITMAP_CACHING, bitmapCachingEnabled);
    settings->setValue(MMR, mmrEnabled);
    settings->setValue(FMMR, fmmrEnabled);
    settings->setValue(ENABLE_DNA, dnaEnabled );
    settings->setValue(GRAPHICS_ACCELERATION, gaEnabled);
    settings->setValue(RECONNECT, reconnect);

    settings->setValue(REDIRECT_CLIPBOARD, redirectClipboard);
    settings->setValue(REDIRECT_DRIVES, redirectDrives);
    settings->setValue(REDIRECT_PRINTERS, redirectPrinters);
    settings->setValue(REDIRECT_UNIVERSAL_PRINTERS, redirectUPrinters);
    settings->setValue(REDIRECT_COM_PORTS, redirectComPorts);
    settings->setValue(REDIRECT_HANDHELDS, redirectHandhelds);
    settings->setValue(REDIRECT_MICROPHONE, redirectMicrophone);
    settings->setValue(REDIRECT_SMARTCARDS, redirectSmartCards);

    settings->setValue(DISPLAY_CONNECTION_BAR, connectionBar);
    settings->setValue(PIN_CONNECTION_BAR, pinConnectionBar);
    settings->setValue(SPAN_MONITORS, spanMonitors);
    settings->setValue(SMART_SIZING, smartSizing);
    settings->setValue(SEAMLESS_MODE, seamlessMode);
    settings->setValue(REDIRECT_AUDIO, soundMode);

    settings->setValue(ENABLE_DOWNSTREAM_AUDIO, downstreamAudioEnabled);
    settings->setValue(DA_MODE, daMode);
    settings->setValue(DA_QUALITY, daQuality);
    settings->setValue(DA_BUFFERING, daBuffering);
    settings->setValue(DA_CUSTOM_QUALITY, daCustomQuality);

    settings->setValue(ENABLE_UPSTREAM_AUDIO, upstreamAudioEnabled);
    settings->setValue(UA_MODE, uaMode);
    settings->setValue(UA_QUALITY, uaQuality);
    settings->setValue(UA_COMPLEXITY, uaComplexity);
    settings->setValue(UA_ENHANCED_COMPRESSION, uaEnhancedCompression);

    settings->setValue(DESKTOP_WIDTH, desktopWidth);
    settings->setValue(DESKTOP_HEIGHT, desktopHeight);
    settings->setValue(COLOR_DEPTH, colorDepth);
    settings->setValue(HIDE_SETTINGS, hideSettings);

    settings->setValue(ALLOW_PASSWORD_MANAGEMENT, allowPasswordManagement);
    settings->setValue(PASSWORD_MANAGEMENT_PORT, pmPort);
    settings->setValue(PASSWORD_MANAGEMENT_SERVER, pmServer);

    settings->setValue(SOURCE_SERVER, sourceServer);
    settings->setValue(SOURCE_PROTOCOL, sourceProtocol);
    settings->setValue(SOURCE_FILE, sourceFile);
    settings->setValue(SOURCE_PROXY, sourceProxy);

    settings->setValue(AUTO_LAUNCH, p_autolaunch.cache());
    settings->setValue(APPLICATIONS, p_applications.cache());

    for (int i = 0; i < m_locations.count(); i++)
    {
        settings->beginGroup("Location" + QString::number(i + 1));

        settings->setValue("Name", m_locations.at(i)->getName());
        settings->setValue(LOCATION_NAT, m_locations.at(i)->isNat());
        settings->setValue(LOCATION_RDP_ON_SSL, m_locations.at(i)->isSsl());
        settings->setValue(LOCATION_SSL_GATEWAY, m_locations.at(i)->getSslGateway());
        settings->setValue(LOCATION_SERVERS, m_locations.at(i)->getServers());
        settings->setValue(LOCATION_TCP_PORT, m_locations.at(i)->getPort());
        settings->setValue(LOCATION_PROTOCOL, m_locations.at(i)->getProtocol());
        settings->setValue(LOCATION_PROXY_BYPASS_LIST, m_locations.at(i)->getProxyBypassList());
        settings->setValue(LOCATION_PROXY_SERVER, m_locations.at(i)->getProxyServer());

        settings->endGroup();
    }

    settings->setValue(AUTO_CONNECT, autoConnect);
    settings->setValue(SHOW_STEPS, showHelpSteps);

    settings->endGroup();
}

bool Farm::loadFarm()
{
    settings->beginGroup(key);

    syslog( LOG_INFO, "Loading Farm: %s", settings->value("Name").toString().toAscii().data());

    name = settings->value("Name").toString();
    setValue(USERNAME, &username, settings->value(USERNAME).toString());
    setValue(PASSWORD, &password, decryptString256(settings->value(PASSWORD).toString()));
    setValueList(DOMAIN_LIST, &domainList, settings->value(DOMAIN_LIST).toStringList());
    domainList = QStringList::fromSet(domainList.toSet());
    setValue(DOMAIN, &currentDomain, settings->value(DOMAIN).toString());
    setValue(ALLOW_SAVE_CREDENTIALS, &saveCredentials, settings->value(ALLOW_SAVE_CREDENTIALS).toBool());
    setValue(KERBEROS_MODE, &kerberosMode, settings->value(KERBEROS_MODE).toBool());
    setValue(KERBEROS_ENABLED, &kerberosEnabled, settings->value(KERBEROS_ENABLED).toBool());
    setValue(SSO_ENABLED, &ssoEnabled, settings->value(SSO_ENABLED).toBool());

    setValue(DEFAULT_LOCATION, &iDefault, settings->value(DEFAULT_LOCATION).toInt());
    setValue(PROMPT_FOR_LOCATION, &promptLocation, settings->value(PROMPT_FOR_LOCATION, true).toBool());
    //iLast = settings->value("LastLocation").toInt();

    setValue(WAIT_FOR_DESKTOP_ENABLED, &waitForDesktopIsEnabled, settings->value(WAIT_FOR_DESKTOP_ENABLED, true). toBool());
    setValue(WAIT_TIME, &waitTime, settings->value(WAIT_TIME, 6).toInt());
    setValue(NUMBER_OF_RETRIES, &numberOfRetries, settings->value(NUMBER_OF_RETRIES, 20).toInt());

    setValue(EXPERIENCE_LEVEL, &experienceLevel, settings->value(EXPERIENCE_LEVEL, 3).toInt());
    setValue(ENABLE_WALLPAPER, &wallpaperEnabled, settings->value(ENABLE_WALLPAPER).toBool());
    setValue(ENABLE_FULL_WINDOW_DRAG, &draggingEnabled, settings->value(ENABLE_FULL_WINDOW_DRAG).toBool());
    setValue(ENABLE_ANIMATION, &animationEnabled, settings->value(ENABLE_ANIMATION).toBool());
    setValue(ENABLE_THEMES, &themesEnabled, settings->value(ENABLE_THEMES).toBool());
    setValue(ENABLE_BITMAP_CACHING, &bitmapCachingEnabled, settings->value(ENABLE_BITMAP_CACHING).toBool());
    setValue(MMR, &mmrEnabled, settings->value(MMR).toBool());
    setValue(FMMR, &fmmrEnabled, settings->value(FMMR).toBool());
    setValue(ENABLE_DNA, &dnaEnabled, settings->value(ENABLE_DNA).toBool());
    setValue(GRAPHICS_ACCELERATION, &gaEnabled, settings->value(GRAPHICS_ACCELERATION).toBool());
    setValue(RECONNECT, &reconnect, settings->value(RECONNECT).toBool());

    setValue(DISPLAY_CONNECTION_BAR, &connectionBar, settings->value(DISPLAY_CONNECTION_BAR).toBool());
    setValue(PIN_CONNECTION_BAR, &pinConnectionBar, settings->value(PIN_CONNECTION_BAR).toBool());
    setValue(SPAN_MONITORS, &spanMonitors, settings->value(SPAN_MONITORS).toBool());
    setValue(SMART_SIZING, &smartSizing, settings->value(SMART_SIZING).toBool());
    setValue(SEAMLESS_MODE, &seamlessMode, settings->value(SEAMLESS_MODE).toBool());
    setValue(REDIRECT_AUDIO, &soundMode, settings->value(REDIRECT_AUDIO).toString());

    setValue(REDIRECT_CLIPBOARD, &redirectClipboard, settings->value(REDIRECT_CLIPBOARD).toBool());
    setValue(REDIRECT_DRIVES, &redirectDrives, settings->value(REDIRECT_DRIVES).toBool());
    setValue(REDIRECT_PRINTERS, &redirectPrinters, settings->value(REDIRECT_PRINTERS).toBool());
    setValue(REDIRECT_UNIVERSAL_PRINTERS, &redirectUPrinters, settings->value(REDIRECT_UNIVERSAL_PRINTERS).toBool());
    setValue(REDIRECT_COM_PORTS, &redirectComPorts, settings->value(REDIRECT_COM_PORTS).toBool());
    setValue(REDIRECT_HANDHELDS, &redirectHandhelds, settings->value(REDIRECT_HANDHELDS).toBool());
    setValue(REDIRECT_MICROPHONE, &redirectMicrophone, settings->value(REDIRECT_MICROPHONE).toBool());
    setValue(REDIRECT_SMARTCARDS, &redirectSmartCards, settings->value(REDIRECT_SMARTCARDS).toBool());

    setValue(ENABLE_DOWNSTREAM_AUDIO, &downstreamAudioEnabled, settings->value(ENABLE_DOWNSTREAM_AUDIO).toBool());
    setValue(DA_MODE, &daMode, settings->value(DA_MODE).toString());
    setValue(DA_QUALITY, &daQuality, settings->value(DA_QUALITY).toString());
    setValue(DA_BUFFERING, &daBuffering, settings->value(DA_BUFFERING).toInt());
    setValue(DA_CUSTOM_QUALITY, &daCustomQuality, settings->value(DA_CUSTOM_QUALITY).toInt());

    setValue(ENABLE_UPSTREAM_AUDIO, &upstreamAudioEnabled, settings->value(ENABLE_UPSTREAM_AUDIO).toBool());
    setValue(UA_MODE, &uaMode, settings->value(UA_MODE).toString());
    setValue(UA_QUALITY, &uaQuality, settings->value(UA_QUALITY).toInt());
    setValue(UA_COMPLEXITY, &uaComplexity, settings->value(UA_COMPLEXITY).toInt());
    setValue(UA_ENHANCED_COMPRESSION, &uaEnhancedCompression, settings->value(UA_ENHANCED_COMPRESSION).toBool());

    setValue(DESKTOP_WIDTH, &desktopWidth, settings->value(DESKTOP_WIDTH).toInt());
    setValue(DESKTOP_HEIGHT, &desktopHeight, settings->value(DESKTOP_HEIGHT).toInt());
    setValue(COLOR_DEPTH, &colorDepth, settings->value(COLOR_DEPTH).toInt());
    setValue(HIDE_SETTINGS, &hideSettings, settings->value(HIDE_SETTINGS).toBool());

    setValue(ALLOW_PASSWORD_MANAGEMENT, &allowPasswordManagement, settings->value(ALLOW_PASSWORD_MANAGEMENT, true).toBool());
    setValue(PASSWORD_MANAGEMENT_PORT, &pmPort, settings->value(PASSWORD_MANAGEMENT_PORT, "444").toInt());
    setValue(PASSWORD_MANAGEMENT_SERVER, &pmServer, settings->value(PASSWORD_MANAGEMENT_SERVER).toString());

    setValue(SOURCE_SERVER, &sourceServer, settings->value(SOURCE_SERVER).toString());
    setValue(SOURCE_PROTOCOL, &sourceProtocol, settings->value(SOURCE_PROTOCOL).toString());
    setValue(SOURCE_FILE, &sourceFile, settings->value(SOURCE_FILE).toString());
    setValue(SOURCE_PROXY, &sourceProxy, settings->value(SOURCE_PROXY).toString());

    //autoLaunch = settings->value("AutoLaunch").toStringList();
    QStringList autoLaunchCache;
    setValueList(AUTO_LAUNCH, &autoLaunchCache, settings->value(AUTO_LAUNCH).toStringList());
    p_autolaunch = ApplicationList(autoLaunchCache);
    QStringList applicationsCache;
    setValueList(APPLICATIONS, &applicationsCache, settings->value(APPLICATIONS).toStringList());
    p_applications = ApplicationList(applicationsCache);

    // Loading existing locations
    QStringList locationsList = settings->childGroups();
    m_locations.clear();

    for (int i = 0; i < locationsList.count(); i++)
    {
        settings->beginGroup(locationsList.at(i));

        LocationSharedPointer location(new Location);

        location->setName(settings->value("Name").toString());
        bool tmpBool;
        QString tmpStr;
        QStringList tmpStrList;
        setValue(LOCATION_NAT, &tmpBool, settings->value(LOCATION_NAT).toBool());
        location->setNat(tmpBool);
        setValue(LOCATION_RDP_ON_SSL, &tmpBool, settings->value(LOCATION_RDP_ON_SSL).toBool());
        location->setSsl(tmpBool);
        setValue(LOCATION_SSL_GATEWAY, &tmpStr, settings->value(LOCATION_SSL_GATEWAY).toString());
        location->setSslGateway(tmpStr);
        setValueList(LOCATION_SERVERS, &tmpStrList, settings->value(LOCATION_SERVERS).toStringList());
        location->setServers(tmpStrList);
        setValue(LOCATION_TCP_PORT, &tmpStr, settings->value(LOCATION_TCP_PORT, "8080").toString());
        location->setPort(tmpStr);
        setValue(LOCATION_PROTOCOL, &tmpStr, settings->value(LOCATION_PROTOCOL, "http").toString());
        location->setProtocol(tmpStr);
        setValue(LOCATION_PROXY_SERVER, &tmpStr, settings->value(LOCATION_PROXY_SERVER).toString());
        location->setProxyServer(tmpStr);
        setValue(LOCATION_PROXY_BYPASS_LIST, &tmpStr, settings->value(LOCATION_PROXY_BYPASS_LIST).toString());
        location->setProxyBypassList(tmpStr);

        m_locations << location;

        settings->endGroup();
    }

    setValue(NEED_RECONNECT, &needReconnect, settings->value(NEED_RECONNECT).toBool());
    setValue(AUTO_CONNECT, &autoConnect, settings->value(AUTO_CONNECT).toBool());
    setValue(SHOW_STEPS, &showHelpSteps, settings->value(SHOW_STEPS).toBool());

    this->m_currentLocation.clear();
    this->setLastLocation(0);

    settings->endGroup();

    return true;
}

QStringList Farm::getLocationNamesList()
{
    QStringList names;
    for (int i = 0; i < m_locations.count(); i++)
    {
        names << m_locations.at(i)->getName();
    }
    return names;
}

LocationSharedPointer Farm::createNewLocation()
{
    // Create new location name
    int newNumber = m_locations.count() + 1;
    QString newName = "Location " + QString::number(newNumber);

    // Initialize new location
    LocationSharedPointer newLocation(new Location);
    newLocation->setName(newName);

    return newLocation;
}

int Farm::addLocation(LocationSharedPointer location)
{
    // Appending location-parameter
    m_locations.append(location);

    return m_locations.indexOf(location);
}

void Farm::removeLocation(int number)
{
    m_locations.removeAt(number);
}

void Farm::removeAllLocations()
{
    m_locations.clear();
}

ApplicationList Farm::applications() const
{
    return p_applications;
}

void Farm::setApplications(const ApplicationList apps)
{
    // TODO: clear cache when refreshing app list
    p_applications.clearWithCache();

    p_applications = apps;
}

ApplicationList Farm::autolaunch() const
{
    return p_autolaunch;
}

void Farm::setAutolaunch(const ApplicationList apps)
{
    p_autolaunch.clear();

    p_autolaunch = apps;
}


template <typename Type>
void Farm::setDefaultValue(const QString& key, Type* variable, const Type value)
{
    if (!m_appSettings->getValue(key, variable))
    {
        *variable = value;
    }
}

template <typename Type>
void Farm::setValue(const QString& key, Type* variable, const Type value)
{
    m_appSettings->beginGroup(FARM_OVERRIDE_VALUES);
    if (!m_appSettings->getValue(key, variable))
    {
        *variable = value;
    }
    m_appSettings->endGroup();
}

template <typename Type>
void Farm::setValueList(const QString& key, QList<Type>* list, const QList<Type> valueList)
{
    m_appSettings->beginGroup(FARM_OVERRIDE_VALUES);
    if (!m_appSettings->getValueList(key, list))
    {
        *list = valueList;
    }
    m_appSettings->endGroup();
}
