// **************************************************************************************
// *
// * © 2012 Dell, Inc.
// * ALL RIGHTS RESERVED.
// *
// * Redistribution and use in source and binary forms, with or without modification, 
// * are permitted provided that the following conditions are met:
// * 
// *  *  Redistributions of source code must retain the above copyright notice, 
// *     this list of conditions and the following disclaimer.  
// * 
// *  *  Redistributions in binary form must reproduce the above copyright notice, 
// *     this list of conditions and the following disclaimer in the documentation 
// *     and/or other materials provided with the distribution.  
// * 
// *  *  Neither the name of Dell, Inc. nor the names of its contributors 
// *     may be used to endorse or promote products derived from this software 
// *     without specific prior written permission. 
// * 
// * 
// * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ON AN "AS IS" 
// * AND UNSUPPORTED BASIS, WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES OF ANY KIND, 
// * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, NON-
// * INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL THE COPYRIGHT 
// * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
// * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
// * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
// * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
// * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
// * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// * 
// **************************************************************************************

#include "updateserver.h"
#include "mainWindow.h"
#include <QX11Info>
#include <X11/Xlib.h>
#include <unistd.h>

#define  INCOMING_MESSAGE_BUFFER_SIZE       (4096)

UpdateServer::UpdateServer(QObject *parent) :
    QObject(parent)
{

    tcpServer = new QTcpServer(this);
    //tcpServerConnection = new QTcpSocket(this);

    if (!tcpServer->listen(QHostAddress::Any, AU_NETWORK_PORT)) {
        /*
             QMessageBox::critical(NULL, tr("vWorkspace Thumbnail Server"),
                                   tr("Unable to start the server: %1.")
                                   .arg(tcpServer->errorString()));

                                   */
             BasicMessageDialog bmd;

             bmd.initialize( "vWorkspace Thumbnail Server", tr("Unable to start the server: %1.")
                             .arg(tcpServer->errorString()) );
             return;
         }

    connect(tcpServer, SIGNAL(newConnection()), this, SLOT(receiveUpdate()));

    syslog( LOG_INFO, "Thumbnail Update server is ready" );
}

void UpdateServer::receiveUpdate()
{
    // I should be able to read from the incoming slot.
    // syslog( LOG_INFO, "Incoming Connection Detected");

    // Initialize the incoming message
    memset( incomingMessage, 0x0, UPDATE_MESSAGE_SIZE);
    incomingByteCount = 0;

    tcpServerConnection = tcpServer->nextPendingConnection();
    // syslog( LOG_INFO, "Got socket( %i )", (int) tcpServerConnection);

    connect( tcpServerConnection, SIGNAL(readyRead()), this, SLOT(getIncomingData()));
    connect( tcpServerConnection, SIGNAL(disconnected()), this, SLOT(handleIncomingUpdate()));
}


void UpdateServer::handleIncomingUpdate()
{
    char * marker;

    // Sanity check for message lengths, etc

    // Treat the first 4 bytes as an int.
    unsigned int * msgLength;
    struct pixel_update_header * hdr;

    marker = incomingMessage;
    msgLength = (unsigned int *) marker;

    // Advance the message pointer
    marker += sizeof( int );

    //syslog( LOG_INFO, "The Update says it has a %i byte payload", *msgLength );

    if (( incomingByteCount - sizeof(int)) != *msgLength )
    {
        syslog( LOG_ERR, "Detected invalid message length info. Discarding update");

        // reset the incoming message buffer for cleanliness
        memset( incomingMessage, 0x0, UPDATE_MESSAGE_SIZE );
        incomingByteCount = 0;
        return;
    }

    ///////////////////////////////////////////////
    //
    // Sanity checks done, message seems good.
    //
    ///////////////////////////////////////////////
    // syslog( LOG_INFO, "Update message length seems good, performing thumbnail update");

    // At this point the byte stream checks out at the correct length,
    // so we can handle it in order to load the pixmap.
    hdr = (struct pixel_update_header *) marker;
    marker += sizeof ( struct pixel_update_header );

    // syslog( LOG_INFO, "Advancing pointer %i bytes for header", sizeof( struct pixel_update_header ));


    // Handle the update - Pass the message to the hander routine
    //
    handleThumbnailUpdate( hdr, marker );


    // Done
}




void UpdateServer::getIncomingData()
{
    unsigned int msgSize;
    unsigned int bytesRead, bytesAvailable;
    char buffer[INCOMING_MESSAGE_BUFFER_SIZE];

    // Set the stream
    QDataStream in(tcpServerConnection);
    in.setVersion(QDataStream::Qt_4_0);

    // Wait for the bytes to fill in
    usleep(10000);

    bytesRead = 0;
    bytesAvailable = tcpServerConnection->bytesAvailable();

    // syslog( LOG_INFO, "Socket says it has %i bytes available to read", bytesAvailable);

    if (bytesAvailable < sizeof(msgSize))
    {
        syslog( LOG_INFO, "Not enough bytes available (%i<%i), trying again",
                (int)tcpServerConnection->bytesAvailable(),
                sizeof(msgSize));
        return;
    }

    while (tcpServerConnection->bytesAvailable() > 0)
    {
        // syslog( LOG_INFO, "Socket says it has %i bytes available to read", tcpServerConnection->bytesAvailable());
        bytesRead = in.readRawData( buffer, INCOMING_MESSAGE_BUFFER_SIZE );
        // syslog( LOG_INFO, "Read %i bytes from the socket", bytesRead);
        memcpy( &(incomingMessage[incomingByteCount]), buffer, bytesRead );

        // Move the incoming count
        incomingByteCount += bytesRead;
    }

    // syslog( LOG_INFO, "According to the message counter, %i total bytes have been read in,", incomingByteCount);
}



void UpdateServer::displayError(QAbstractSocket::SocketError socketError)
 {
     switch (socketError) {
     case QAbstractSocket::RemoteHostClosedError:
         break;
     case QAbstractSocket::HostNotFoundError:
         QMessageBox::information(NULL, tr("UpdateServer"),
                                  tr("The host was not found. Please check the "
                                     "host name and port settings."));
         break;
     case QAbstractSocket::ConnectionRefusedError:
         QMessageBox::information(NULL, tr("UpdateServer"),
                                  tr("The connection was refused by the peer. "
                                     "Make sure the server is running, "
                                     "and check that the host name and port "
                                     "settings are correct."));
         break;
     default:
         QMessageBox::information(NULL, tr("UpdateServer"),
                                  tr("The following error occurred: %1.")
                                  .arg(tcpServer->errorString()));
     }
 }


void UpdateServer::handleThumbnailUpdate( struct pixel_update_header * hdr,  char * data )
{

    int width, height, seamless;
    long * marker;
    QImage thumbnail;

    width    = hdr->width;
    height   = hdr-> height;
    seamless = hdr->seamless;
    marker   = (long *) data;

    QPixmap pix( width, height);
    QImage image( QSize( hdr->width, hdr->height), QImage::Format_RGB32 );

    for ( int y = 0; y < height; y++ )
    {
        for ( int x = 0; x < width; x ++)
        {
            image.setPixel( x, y, marker[ (y * width ) + x]);
        }
    }

    // syslog( LOG_INFO, "Detected an update from %s of %i by %i pixels",hdr->session_name, hdr->width, hdr->height );
    if ( hdr->action == AU_ACTION_REMOVE_SESSION )
    {
        syslog( LOG_WARNING, "Update server received a session close request for session %s", hdr->session_name );
    }

    else
    {
        // Use a Qpainter to draw the marker on to the icon.
        QImage close_icon( tr(":/img2/redX.png"));

        // Resize incoming image to the gridsize
        // syslog( LOG_INFO, "Resizing thumbnail to width = %i", thumbnailWidth );
        thumbnail = image.scaledToWidth( thumbnailWidth, Qt::SmoothTransformation );
        QPainter p( &thumbnail );

        p.drawImage( QPoint( 0, 0 ), close_icon);
    }

    emit updateThumbnail( pix.fromImage( thumbnail ), QString( hdr->session_name), hdr);
}

void UpdateServer::raiseWindow( int wid )
{
    int rc;

    char command[ 1024];
    sprintf( command, "/usr/bin/wmctrl -i -a 0x%08X", wid );
    rc = system( command );
}

void UpdateServer::closeWindow( int wid )
{
    int rc;

    char command[ 1024];
    sprintf( command, "/usr/bin/wmctrl -i -c 0x%08X", wid );
    rc = system( command );
}

void UpdateServer::setThumbnailWidth( int width )
{
    thumbnailWidth = width;
}

