#include "audioInputChannel.h"

#define WIDM_START       57
#define WIDM_STOP        58


AudioInputChannel::AudioInputChannel()
    : DynamicChannel("AUDIO_INPUT"), m_pRecorder(NULL)
{
}

AudioInputChannel::~AudioInputChannel()
{
    destroyRecorder();
}

void AudioInputChannel::onClose()
{
    destroyRecorder();
}

void* AudioInputChannel::audioProcessing(void* pThis)
{
    AudioInputChannel* pChannel = static_cast<AudioInputChannel*>(pThis);

    pChannel->m_needStop = false;

    configure_recorder(pChannel->m_pRecorder, WIDM_START);

    int recordedCount;
    while (!pChannel->m_needStop)
    {
        do
        {
            recordedCount = record_wave(pChannel->m_pRecorder, pChannel->m_buffer, audioBufferSize);
            if (recordedCount > 0)
            {
                pChannel->sendDataPdu(pChannel->m_buffer, audioBufferSize);
            }
        }
        while (recordedCount > 0);
        usleep(100);
    }

    configure_recorder(pChannel->m_pRecorder, WIDM_STOP);

    return NULL;
}

void AudioInputChannel::createRecorder(WAVEFORMATEX* pFormat)
{
    m_pRecorder = create_recorder(pFormat);

    if (m_pRecorder != NULL)
    {
        pthread_create(&m_audioProcessingThread, NULL, audioProcessing, this);
    }
}

void AudioInputChannel::destroyRecorder()
{
    m_needStop = true;

    if (m_pRecorder != NULL)
    {
        pthread_join(m_audioProcessingThread, NULL);
        destroy_recorder(m_pRecorder);
        m_pRecorder = NULL;
    }
}

void AudioInputChannel::sendDataPdu(uint8* data, uint32 size)
{
    uint8 header = MSG_SNDIN_DATA;
    // 1 b header + data size
    int dataLength = sizeof(header) + size;

    STREAM outStream = m_pDynamicChannelManager->prepareDataPdu(getId(), dataLength);
    if (NULL == outStream)
    {
        return;
    }
    out_uint8(outStream, header);
    out_uint8p(outStream, data, size);
    m_pDynamicChannelManager->sendDataPdu(outStream);
}

void AudioInputChannel::sendVersionPdu()
{
    uint8 header = MSG_SNDIN_VERSION;
    uint32 version = 1;
    // 1 b header + 4 b version
    int dataLength = sizeof(header) + sizeof(version);

    STREAM outStream = m_pDynamicChannelManager->prepareDataPdu(getId(), dataLength);
    if (NULL == outStream)
    {
        return;
    }
    out_uint8(outStream, header);
    out_uint32_le(outStream, version);
    m_pDynamicChannelManager->sendDataPdu(outStream);
}

void AudioInputChannel::sendFormatsPdu(int pcmIndex)
{
    uint8 header = MSG_SNDIN_FORMATS;
    uint32 numFormats = 1; // always send only 1 format - WAVE_FORMAT_PCM
    uint32 cbSizeFormatsPacket;
    WAVEFORMATEX audioFormat = audioFormats[pcmIndex];
    audioFormat.cbSize = 0;
    // 1 b header + 4 b numFormats + 4 b cbSizeFormatsPacket + 18 b SoundFormats
    int dataLength = sizeof(header) + sizeof(numFormats) + sizeof(cbSizeFormatsPacket) + sizeof(audioFormat.wFormatTag)
                   + sizeof(audioFormat.nChannels) + sizeof(audioFormat.nSamplesPerSec)
                   + sizeof(audioFormat.nAvgBytesPerSec) + sizeof(audioFormat.nBlockAlign)
                   + sizeof(audioFormat.wBitsPerSample) + sizeof(audioFormat.cbSize);
    cbSizeFormatsPacket = dataLength;

    STREAM outStream = m_pDynamicChannelManager->prepareDataPdu(getId(), dataLength);
    if (NULL == outStream)
    {
        return;
    }
    out_uint8(outStream, header);
    out_uint32_le(outStream, numFormats);
    out_uint32_le(outStream, cbSizeFormatsPacket);
    out_uint16_le(outStream, audioFormat.wFormatTag);
    out_uint16_le(outStream, audioFormat.nChannels);
    out_uint32_le(outStream, audioFormat.nSamplesPerSec);
    out_uint32_le(outStream, audioFormat.nAvgBytesPerSec);
    out_uint16_le(outStream, audioFormat.nBlockAlign);
    out_uint16_le(outStream, audioFormat.wBitsPerSample);
    out_uint16_le(outStream, audioFormat.cbSize);
    m_pDynamicChannelManager->sendDataPdu(outStream);
}

void AudioInputChannel::sendOpenReplyPdu(uint32 status)
{
    uint8 header = MSG_SNDIN_OPEN_REPLY;
    uint32 result = status;
    // 1 b header + 4 b result
    int dataLength = sizeof(header) + sizeof(result);

    STREAM outStream = m_pDynamicChannelManager->prepareDataPdu(getId(), dataLength);
    if (NULL == outStream)
    {
        return;
    }
    out_uint8(outStream, header);
    out_uint32_le(outStream, result);
    m_pDynamicChannelManager->sendDataPdu(outStream);
}

void AudioInputChannel::sendIncomingDataPdu()
{
    uint8 header = MSG_SNDIN_DATA_INCOMING;
    // 1 b header
    int dataLength = sizeof(header);

    STREAM outStream = m_pDynamicChannelManager->prepareDataPdu(getId(), dataLength);
    if (NULL == outStream)
    {
        return;
    }
    out_uint8(outStream, header);
    m_pDynamicChannelManager->sendDataPdu(outStream);
}

void AudioInputChannel::processVersionPdu(STREAM s)
{
    uint32 version;
    in_uint32_le(s, version);
    if (1 == version)
    {
        sendVersionPdu();
    }
}

void AudioInputChannel::processFormatsPdu(STREAM s)
{
    uint32 numFormats;
    uint32 cbSizeFormatsPacket;
    in_uint32_le(s, numFormats);
    in_uint32_le(s, cbSizeFormatsPacket); // Ignor it
    if (numFormats > maxNumberOfAudioFormats)
    {
        numFormats = maxNumberOfAudioFormats;
    }

    int pcmIndex = -1;
    for (uint32 i = 0; i < numFormats; ++i)
    {
        in_uint16_le(s, audioFormats[i].wFormatTag);
        in_uint16_le(s, audioFormats[i].nChannels);
        in_uint32_le(s, audioFormats[i].nSamplesPerSec);
        in_uint32_le(s, audioFormats[i].nAvgBytesPerSec);
        in_uint16_le(s, audioFormats[i].nBlockAlign);
        in_uint16_le(s, audioFormats[i].wBitsPerSample);
        in_uint16_le(s, audioFormats[i].cbSize);
        s->p += audioFormats[i].cbSize;
        if (WAVE_FORMAT_PCM == audioFormats[i].wFormatTag)
        {
            pcmIndex = i;
        }
    }

    if (pcmIndex >= 0)
    {
        sendIncomingDataPdu();
        sendFormatsPdu(pcmIndex);
    }
}

void AudioInputChannel::processOpenPdu(STREAM s)
{
    const uint32 S_OK = 0;
    const uint32 STATUS_INVALID_DEVICE_STATE = 0xC0000184;

    uint32 framesPerPacket;
    uint32 initialFormat;
    WAVEFORMATEX audioFormat;

    in_uint32_le(s, framesPerPacket);
    in_uint32_le(s, initialFormat);
    in_uint16_le(s, audioFormat.wFormatTag);
    in_uint16_le(s, audioFormat.nChannels);
    in_uint32_le(s, audioFormat.nSamplesPerSec);
    in_uint32_le(s, audioFormat.nAvgBytesPerSec);
    in_uint16_le(s, audioFormat.nBlockAlign);
    in_uint16_le(s, audioFormat.wBitsPerSample);
    in_uint16_le(s, audioFormat.cbSize);

    if (initialFormat > maxNumberOfAudioFormats)
    {
        return;
    }
    createRecorder(&audioFormats[initialFormat]);
    uint32 status;
    if (m_pRecorder != NULL)
    {
        status = S_OK;
        syslog(LOG_NOTICE, "=== Sound recording device is successfully created");
    }
    else
    {
        status = STATUS_INVALID_DEVICE_STATE;
        syslog(LOG_NOTICE, "=== Sound recording device is NOT created");
    }
    sendOpenReplyPdu(status);
}

void AudioInputChannel::processMessage(STREAM s, uint32 totalLength)
{
    uint8 header;
    in_uint8(s, header);
    switch (header)
    {
        case MSG_SNDIN_VERSION:
        {
            processVersionPdu(s);
            break;
        }
        case MSG_SNDIN_FORMATS:
        {
            processFormatsPdu(s);
            break;
        }
        case MSG_SNDIN_OPEN:
        {
            processOpenPdu(s);
            break;
        }
        case MSG_SNDIN_FORMATCHANGE:
        {
            // Not implemented !!!
            break;
        }
        default:
        {
            break;
        }
    }
}
