#include "pnsslcli.h"
#include <syslog.h>

#include <boost/bind.hpp>
#include <assert.h>

// ------------------------------------------------------------------------------
// session member functions realization
// ------------------------------------------------------------------------------
session::session (asio::io_service& io_service,
                  asio::ssl::context& context)
    : socket_ (io_service),
      ssl_socket_ (io_service, context),
      timer_ (io_service)
{
}

// ------------------------------------------------------------------------------
tcp::socket& session::socket ()
{
    return socket_;
}

// ------------------------------------------------------------------------------
ssl_socket& session::sslsocket ()
{
    return ssl_socket_;
}

// ------------------------------------------------------------------------------
void session::start (const char *rdphost, const char *username)
{
    char buf[128];
    memset (buf, '\0', 128);

    int sendbytes = sprintf (buf, "CONNECT: %s\r\nCLIENT: %s\r\n\r\n", rdphost, username);
    syslog (LOG_NOTICE, buf);
    ssl_socket_.write_some (asio::buffer (buf, sendbytes));

    //sleep (1);  // give VPN server time to prepare

    ssl_socket_.async_read_some (asio::buffer (data_in_),
        make_custom_alloc_handler (allocator_in_,
            boost::bind (&session::handle_server_read,
               shared_from_this (),
               asio::placeholders::error,
               asio::placeholders::bytes_transferred)));
    socket_.async_read_some (asio::buffer (data_out_),
        make_custom_alloc_handler (allocator_out_,
            boost::bind (&session::handle_read,
                   shared_from_this (),
                   asio::placeholders::error,
                   asio::placeholders::bytes_transferred)));
}

// ------------------------------------------------------------------------------
void session::on_timer (const asio::error_code& e)
{
    if (!e)
    {
        syslog (LOG_NOTICE, "write operation timeout");
        ssl_socket_.lowest_layer().close();
    }
}

// ------------------------------------------------------------------------------
void session::handle_read (const asio::error_code & error, size_t bytes_transferred)
{
    if (!error)
    {
        timer_.expires_from_now (boost::posix_time::seconds(10));
        timer_.async_wait (boost::bind (&session::on_timer,
                                        shared_from_this (),
                                        asio::placeholders::error));
        asio::async_write (ssl_socket_,
                asio::buffer (data_out_, bytes_transferred),
                make_custom_alloc_handler (allocator_out_,
                    boost::bind (&session::handle_server_write,
                        shared_from_this (),
                        asio::placeholders::error)));
    }
    else
    {
      throw error;
    }
}

// ------------------------------------------------------------------------------
void session::handle_server_read (const asio::error_code & error, size_t bytes_transferred)
{
    if (!error)
    {
        asio::async_write (socket_,
                asio::buffer (data_in_, bytes_transferred),
                make_custom_alloc_handler (allocator_in_,
                    boost::bind (&session::handle_write,
                        shared_from_this (),
                        asio::placeholders::error)));
    }
    else
    {
      syslog (LOG_NOTICE, "server_read error: %s", error.message().c_str());
      throw error;
    }
}

// ------------------------------------------------------------------------------
void session::handle_write (const asio::error_code & error)
{
    if (!error)
    {
        ssl_socket_.async_read_some (asio::buffer (data_in_),
            make_custom_alloc_handler (allocator_in_,
                boost::bind (&session::handle_server_read,
                    shared_from_this (),
                    asio::placeholders::error,
                    asio::placeholders::bytes_transferred)));
    }
    else
    {
      throw error;
    }
}

// ------------------------------------------------------------------------------
void session::handle_server_write (const asio::error_code & error)
{
    timer_.cancel();
    if (!error)
    {
        socket_.async_read_some (asio::buffer (data_out_),
                make_custom_alloc_handler (allocator_out_,
                    boost::bind (&session::handle_read,
                        shared_from_this (),
                        asio::placeholders::error,
                        asio::placeholders::bytes_transferred)));
    }
    else
    {
      throw error;
    }
}


// ------------------------------------------------------------------------------
//                      pnssl_server
// ------------------------------------------------------------------------------

pnssl_server::pnssl_server(asio::io_service& io_service,
                           const tcp::endpoint& endpoint,
                           tcp::resolver::iterator & iterator,
                           const char * remote_addr,
                           const char * username)
: io_service_(io_service),
acceptor_(io_service, endpoint),
endpoint_iterator_(iterator),
remote_addr_(remote_addr),
username_(username)
{
    asio::ssl::context ctx(io_service_, asio::ssl::context::sslv23);
    session_ptr new_session(new session(io_service_, ctx));
    ctx.set_verify_mode (asio::ssl::context::verify_peer);
    acceptor_.async_accept(new_session->socket(),
                           boost::bind(&pnssl_server::handle_accept, this, new_session,
                                       asio::placeholders::error));
}

void pnssl_server::handle_accept(session_ptr sess,
                   const asio::error_code& error)
{
    if (!error)
    {
        // connect to SSL VPN server
        sess->sslsocket().lowest_layer().connect(*endpoint_iterator_);
        syslog (LOG_NOTICE, "pnssl_server: connect to ssl gateway is done");
        sess->sslsocket().handshake(asio::ssl::stream_base::client);
        syslog (LOG_NOTICE, "pnssl_server: handshake passed");
        sess->start(remote_addr_.c_str(), username_.c_str());
        
        asio::ssl::context ctx(io_service_, asio::ssl::context::sslv23);
        session_ptr new_session(new session(io_service_, ctx));
        ctx.set_verify_mode (asio::ssl::context::verify_peer);
        acceptor_.async_accept(new_session->socket(),
                               boost::bind(&pnssl_server::handle_accept, this, new_session,
                                           asio::placeholders::error));
    }
}

