#include "rdesktop.h"
#include "qaudio_alsa_drv.h"

#define ALSA_PCM_NEW_HW_PARAMS_API

#include <alsa/asoundlib.h>

static char DEFAULTDEVICE[] = "default";

typedef struct
{
    snd_pcm_t * handle;
    WAVEFORMATEX format;    
}alsa_waveout_handle , alsa_wavein_handle;

static char * init_options = NULL;

int alsa_driver_init ( char * _options )
{
    if ( _options != NULL )
        init_options = xstrdup( _options );
    else
        init_options = xstrdup (DEFAULTDEVICE);

    return 0;
}

int alsa_driver_free ()
{
	xfree( init_options );
	return 0;
}

int alsa_waveout_open  ( LPHWAVEOUT _handle, WAVEFORMATEX * _format )
{
    snd_pcm_hw_params_t * params = NULL;
    snd_pcm_t * handle = NULL;
    int res = 0;
    unsigned int val = 0;
    int dir = 0;
        
    res = snd_pcm_open( &handle, init_options , SND_PCM_STREAM_PLAYBACK, 0 );
    if ( res < 0 )
    {
        /* log error */
        return -1;
    }
    
    snd_pcm_hw_params_alloca(&params);
    
    snd_pcm_hw_params_any (handle, params );
    
    snd_pcm_hw_params_set_access( handle, params , SND_PCM_ACCESS_RW_INTERLEAVED);
    
    snd_pcm_hw_params_set_format ( handle , params , (_format->wBitsPerSample==16)? SND_PCM_FORMAT_S16_LE: SND_PCM_FORMAT_S8 );
    
    snd_pcm_hw_params_set_channels( handle, params , _format->nChannels );
    
    val = _format->nSamplesPerSec;
    
    snd_pcm_hw_params_set_rate_near(handle, params , &val, &dir );
     
    res = snd_pcm_hw_params( handle , params );
    
    if ( res < 0 )
    {
        /* log error */
        snd_pcm_close( handle );

        return -1;
    }  
    
    {
    	alsa_waveout_handle * ahandle = (alsa_waveout_handle *)xmalloc (sizeof ( alsa_waveout_handle ) );
		*_handle = ahandle;
		ahandle->handle = handle;
		memcpy( &ahandle->format, _format, sizeof ( WAVEFORMATEX ) );

		snd_pcm_prepare( handle );
    }
        
    return 0;    
}

int alsa_waveout_reset ( HWAVEOUT _handle )
{
    alsa_waveout_handle * handle = (alsa_waveout_handle*) _handle;
    
    if ( handle == NULL )
        return -1;
    
    snd_pcm_drop( handle->handle );
    return 0;
}

int alsa_waveout_close ( HWAVEOUT _handle )
{
    alsa_waveout_handle * handle = (alsa_waveout_handle*) _handle;
    if ( handle == NULL )
        return -1;
    
    snd_pcm_close( handle->handle );
    xfree( handle );
    
    return 0;
}

static uint32 bytes_to_frames ( alsa_waveout_handle * _handle,  uint32 _bytes )
{
    return _bytes / _handle->format.nBlockAlign;
}

int alsa_waveout_write ( HWAVEOUT _handle , uint8 * _wave , uint32 _size )
{
    int res = 0, i = 0;
    alsa_waveout_handle * handle = (alsa_waveout_handle*) _handle;
    if ( handle == NULL )
		return -1;


    i = 0;
    do
    {
    	++ i;
    	res = snd_pcm_writei( handle->handle , _wave, bytes_to_frames (handle, _size ));
    	if ( res < 0 )
    	{
    		snd_pcm_prepare( handle->handle);
    	}
    }while (( res < 0 ) && ( i < 8 ) );

    if ( res < 0 )
    {
    	printf("snd_pcm %s\n", snd_strerror(res));
    }
   return res;
}

/*-------------------------------------------------------------------------------*/

int alsa_wavein_open   ( LPHWAVEIN _handle, WAVEFORMATEX * _format )
{
	snd_pcm_hw_params_t * params = NULL;
	snd_pcm_t * handle = NULL;
	int res = 0;
	unsigned int val = 0;
	int dir = 0;

	if ( (_handle == 0 ) || (_format == NULL) )
		return -1;

    // Setting work with sound stream in non-blocking mode to prevent a program hanging. It can
    //hangs the main thread in destroy_recorder() function when calling pthread_join()
    //for this thread that never ends its execution if snd_pcm_readi() works in blocking mode
    //and cann't get requested amount of frames
    res = snd_pcm_open(&handle, init_options,SND_PCM_STREAM_CAPTURE, SND_PCM_NONBLOCK );
	if (res < 0 )
		return res;

	snd_pcm_hw_params_alloca(&params);

    snd_pcm_hw_params_any( handle , params );

    snd_pcm_hw_params_set_access(handle, params, SND_PCM_ACCESS_RW_INTERLEAVED);

    snd_pcm_hw_params_set_format(handle, params ,
    		(_format->wBitsPerSample==16)?
    				SND_PCM_FORMAT_S16_LE : SND_PCM_FORMAT_S8);

    snd_pcm_hw_params_set_channels(handle,params ,_format->nChannels);

    val = _format->nSamplesPerSec;
    snd_pcm_hw_params_set_rate_near(handle, params,&val,&dir );

    res = snd_pcm_hw_params( handle , params );
    if (res < 0 )
    {
    	snd_pcm_close( handle );
    	return res;
    }

    snd_pcm_prepare( handle );

    {
    	alsa_wavein_handle * ahandle = (alsa_wavein_handle*) xmalloc(sizeof(alsa_wavein_handle) );
    	if (ahandle == NULL)
    	{
    		snd_pcm_close( handle );
    		return -1;
    	}
    	ahandle->handle = handle;
    	memcpy(&ahandle->format, _format , sizeof (WAVEFORMATEX));

    	*_handle = (HWAVEIN)ahandle;
    }

	return 0;
}

int alsa_wavein_close  ( HWAVEIN _handle )
{
	alsa_wavein_handle * handle = (alsa_wavein_handle*) _handle;
	if (handle == NULL)
		return -1;

	snd_pcm_drop( handle->handle );

	snd_pcm_close( handle->handle );

	return 0;
}

int alsa_wavein_start  ( HWAVEIN _handle )
{
	return 0;
}

int alsa_wavein_stop   ( HWAVEIN _handle )
{
	return 0;
}

static int frames_to_bytes ( alsa_wavein_handle * _handle , int _frames )
{
	return _frames * _handle->format.nBlockAlign;
}

int alsa_wavein_record ( HWAVEIN _handle , uint8 * _data , uint32 _size )
{
	alsa_wavein_handle * handle = (alsa_wavein_handle*)_handle;
	if ( handle == NULL )
		return 0;

	{
		int res = 0;
		res = snd_pcm_readi( handle->handle, _data, bytes_to_frames( handle, _size) );
		if ( res < 0 )
		{
			snd_pcm_prepare( handle->handle);
			return 0;
		}

		return frames_to_bytes( handle, res );
	}
}

