#include "rdesktop.h"
#include "qmic.h"

#include <speex/speex.h>
#include <speex/speex_stereo.h>

#include "qmic_codec.h"

const SpeexMode * GetMode ( int sampling )
{
    if ( sampling >= 25000 )
	return &speex_uwb_mode;
    if ( sampling >= 12500 )
	return &speex_wb_mode;
    return &speex_nb_mode;
}

uint32 codec_get_encoder_frame_samples ( SpeexCodec * codec )
{
    uint32 frame_samples;
    speex_encoder_ctl( codec->encoder , SPEEX_GET_FRAME_SIZE, &frame_samples );
    
    return frame_samples * codec->format.nChannels;
}

uint32 codec_get_encoder_frame_bytes ( SpeexCodec * codec )
{
    return sizeof(short) * codec_get_encoder_frame_samples( codec );
}

SpeexCodec * codec_create ( LPWAVEFORMATEX format, int complexity, int quality )
{
    SpeexCodec * codec = NULL;
    
    if ( format == NULL ) 
    {
	error ("CODEC: Invalid format\n");
	return 0;
    }
    
    if (!( 0 < format->nChannels && format->nChannels <= 2 ))
    {
	error ("CODEC: Invalid format, channels count\n");
	return 0;
    }
    
    codec = (SpeexCodec*)xmalloc( sizeof ( SpeexCodec ) );
    
    memcpy ( &codec->format , format, sizeof ( WAVEFORMATEX ) );
    
    codec->encoder = speex_encoder_init ( GetMode(format->nSamplesPerSec) );
    
    if ( codec->encoder == NULL )
    {
	error ( "CODEC: Can't create encoder\n");	 
	xfree(codec);
	return 0;
    }
    else
    {   
	DEBUG_QMIC(("SampesPerSec : %d\n", format->nSamplesPerSec));
        speex_encoder_ctl(codec->encoder,SPEEX_SET_SAMPLING_RATE,&format->nSamplesPerSec);
        
        speex_encoder_ctl(codec->encoder, SPEEX_SET_COMPLEXITY, &complexity);
	
	speex_encoder_ctl(codec->encoder,SPEEX_SET_QUALITY, &quality);
	
	codec->data_buffer = (short*)xmalloc( codec_get_encoder_frame_bytes(codec) );
	codec->data_buffer_size = 0;
	
	speex_bits_init ( &codec->bits );	
    }
    return codec;    
}

void codec_destroy ( SpeexCodec * codec )
{
    if ( codec == NULL )
	return;

    if ( codec->encoder != NULL )
        speex_encoder_destroy( codec->encoder );

    if ( codec->data_buffer != 0 )
	xfree ( codec->data_buffer );

    speex_bits_destroy ( &codec->bits );

    xfree( codec );
}


int codec_read_frame_16bps ( SpeexCodec * codec , char * buffer , uint32 size , uint32 * offset , uint32 to_read_samples )
{
    uint32 to_read_bytes = to_read_samples * sizeof(short);
    uint32 avaliable_bytes = size - *offset;
    uint32 to_copy_bytes = (avaliable_bytes < to_read_bytes) ? avaliable_bytes : to_read_bytes;

    if ( to_copy_bytes == 0 ) return 0;

    DEBUG_QMIC(("CODEC: bytes to copy %d\n" , to_copy_bytes));
    
    memcpy ( codec->data_buffer + codec->data_buffer_size/sizeof(short) , buffer + *offset , to_copy_bytes );

    codec->data_buffer_size += to_copy_bytes ;
    *offset += to_copy_bytes;

    return ( to_copy_bytes == to_read_bytes ) ? 1 : 0;    
}

int codec_read_frame_8bps  ( SpeexCodec * codec , char * buffer , uint32 size , uint32 * offset , uint32 to_read )
{
    uint32 i;
    short * db = codec->data_buffer + codec->data_buffer_size / sizeof(short);

    uint32 to_copy = ( (size - *offset) < to_read )? (size - *offset) : to_read;
    
    buffer += *offset;
    
    for ( i = 0; i < to_copy; ++ i )
    {
	short val = buffer[i];
	val = ( val << 8 ) ^ 0x8000;
	*db++ = val;
    }
    
    codec->data_buffer_size += to_copy * sizeof(short);
    *offset += to_copy;

    return ( to_copy == to_read )? 1: 0;
}

int codec_read_frame ( SpeexCodec * codec, char * buffer , uint32 size , uint32 * offset )
{
    uint32 to_read_samples = 0;
    uint32 frame_samples = codec_get_encoder_frame_samples ( codec );

    DEBUG_QMIC(("CODEC: frame_samples %d samples\n", frame_samples));
  
    if ( codec->format.wBitsPerSample == 16 )
	{
		to_read_samples = frame_samples - codec->data_buffer_size / sizeof(short);
		DEBUG_QMIC(("CODEC: samples to read %d\n", to_read_samples));
		return codec_read_frame_16bps ( codec , buffer , size , offset , to_read_samples );
	}
    else
	{
		to_read_samples = frame_samples - codec->data_buffer_size / sizeof(short);
		DEBUG_QMIC(("CODEC: bytes to read %d\n", to_read_samples));
		return codec_read_frame_8bps ( codec , buffer , size , offset , to_read_samples );
	}
    
    return 0;
}

int codec_compress_frame ( SpeexCodec * codec , short* frame , uint8 * out )
{
    uint32 bytes;
    
    speex_bits_reset ( &codec->bits );
    
    if ( codec->format.nChannels == 2 )
    {
		int frame_size;
	
		speex_encoder_ctl ( codec->encoder, SPEEX_GET_FRAME_SIZE, &frame_size );

		speex_encode_stereo_int ( frame, frame_size , &codec->bits );
    }
    
    speex_encode_int( codec->encoder , frame, &codec->bits);  
    
    speex_bits_insert_terminator(&codec->bits);
    
    bytes = speex_bits_nbytes( &codec->bits );
    DEBUG_QMIC(("speex_bits_nbytes: %d\n", bytes));

    bytes = speex_bits_write ( &codec->bits, (char*)out + 2 ,bytes );
    
    *((short*)out) = bytes & 0xFFFF;
    
    return bytes + 2;
}

uint32 codec_compress_buffer ( SpeexCodec * codec , char * buffer , uint32 size , uint8 * out )
{
    uint32 offset = 0, compressed_size = 0;
    
    if ( (codec == NULL) || (buffer == 0) || (size == 0) ) 
        return 0;

    DEBUG_QMIC(("CODEC: starting read frame\n"));
    while ( codec_read_frame ( codec , buffer , size , &offset ) )
    {
		DEBUG_QMIC(("CODEC: compressing frame..."));
		compressed_size += codec_compress_frame ( codec , codec->data_buffer , out + compressed_size );
		DEBUG_QMIC(("Ok [%d]\n", compressed_size));
		codec->data_buffer_size = 0; /* next frame */
    }
    
    /* codec->data_buffer contains size & frame_size bytes */

    return compressed_size;
}

