// **************************************************************************************
// *
// * © 2012 Dell, Inc.
// * ALL RIGHTS RESERVED.
// *
// * Redistribution and use in source and binary forms, with or without modification, 
// * are permitted provided that the following conditions are met:
// * 
// *  *  Redistributions of source code must retain the above copyright notice, 
// *     this list of conditions and the following disclaimer.  
// * 
// *  *  Redistributions in binary form must reproduce the above copyright notice, 
// *     this list of conditions and the following disclaimer in the documentation 
// *     and/or other materials provided with the distribution.  
// * 
// *  *  Neither the name of Dell, Inc. nor the names of its contributors 
// *     may be used to endorse or promote products derived from this software 
// *     without specific prior written permission. 
// * 
// * 
// * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ON AN "AS IS" 
// * AND UNSUPPORTED BASIS, WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES OF ANY KIND, 
// * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, NON-
// * INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL THE COPYRIGHT 
// * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
// * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
// * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
// * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
// * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
// * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// * 
// **************************************************************************************

#include <sys/types.h>
#include <signal.h>

#include <sys/stat.h>
#include <cstdlib>
#include <errno.h>
#include <fcntl.h>
#include <pthread.h>
#include <boost/bind.hpp>
#include <vector>
#include <time.h>
#include <popt.h>
#include <stdio.h>
#include <iconv.h>
#include <cups/cups.h>

#include "../../common/crypto.h"
#include "../../common/errors.h"
#include "../../common/common.h"
#include "../../common/FlashPluginUtils.h"
#include "../../connector-common/Utilities/U_Lexical.h"
#include "pnllmcli.h"
#include "pnsession.h"
#include "pnserver.h"
#include "tmp.h"

#define MAX_LAYOUT_NUM 41


//---------------------------------------------------------------------------
bool testLocalPndnaPort( int testport )
{

    socklen_t          sockaddr_len;
    int                test_socket;
    struct sockaddr_in test_sockaddr;


    // Create a TCP socket.
    test_socket = socket( AF_INET, SOCK_STREAM, IPPROTO_TCP );

    if ( test_socket == -1 )
    {
        syslog (LOG_INFO, "could not create TCP socket");
    }

    sockaddr_len = sizeof( test_socket );

    // Bind the TCP socket.
    memset( &test_sockaddr, 0, sizeof( test_sockaddr ));
    test_sockaddr.sin_family = AF_INET;
    test_sockaddr.sin_addr.s_addr = htonl( INADDR_ANY );
    test_sockaddr.sin_port = htons( testport );

    int isReuse = 1;
    if ( setsockopt(test_socket, SOL_SOCKET, SO_REUSEADDR, &isReuse, sizeof(isReuse)) == -1 )
    {
        syslog( LOG_ERR, "Could not set socket options. Error: %s", strerror(errno));
        return false;
    }

    if ( bind(test_socket, (const struct sockaddr *)&test_sockaddr, sizeof(test_sockaddr)) == -1 )
    {
        syslog ( LOG_WARNING, "could not bind TCP socket");
        return false;
    }

        // Listen on the TCP socket.
    if (listen(test_socket, 10) == -1 )
    {
        syslog (LOG_NOTICE, "could not listen on TCP socket");
        return false;
    }


    // If I got here then the socket is fine, so shut it down and report that it can be used.
    close( test_socket );
    return true;
}

//---------------------------------------------------------------------------
int getFreeLocalPort( int first, int last )
{
    int i;

    for ( i = first; i <= last; i ++ )
    {
       if ( testLocalPndnaPort( i ))
        {
            syslog( LOG_NOTICE, "Port %i appears to be usable by PNDNA", i);
            return i;
        }
    }

    return -1;

}

//---------------------------------------------------------------------------
pid_t startPNDNAProxy(const std::string& localHost, const std::string& remoteHost)
{     
    pid_t pid = fork();
    
    if (pid == 0)
    {        
        //example of command line for pndnaproxy
        //pndnaproxy -L localhost:8000 -R 192.168.99.15:3389 -c -t 3

        syslog(LOG_NOTICE, "Starting PNDNA proxy: 'pndnaproxy -L %s -R %s -c -t 3'", localHost.c_str(), remoteHost.c_str());
        int execRes = ::execlp("pndnaproxy", 
                               "pndnaproxy", 
                               "-L", localHost.c_str(), 
                               "-R", remoteHost.c_str(), 
                               "-c", 
                               "-t", "3",  //  hardcoded number of threads
                               NULL);
        
        syslog(LOG_NOTICE, "Starting PNDNA proxy: execRes = %d, errno = %d", execRes, int(errno));
        exit(0);
    }
    
    return pid;
} 

//---------------------------------------------------------------------------
//    pnsession
//---------------------------------------------------------------------------
pnsession::pnsession( asio::io_service& io_service, pnserver * server )
        : socket_( io_service ),
        server_ ( server )
{
    syslog ( LOG_NOTICE, " " );
    syslog ( LOG_NOTICE, "Instantiating session %p", this );
    syslog ( LOG_NOTICE, " " );

    pndnaProxyPid = 0;

    server_->inc_session_count();
}

//---------------------------------------------------------------------------
pnsession::~pnsession()
{
    std::string str;
    pid_t pid;
    syslog ( LOG_NOTICE, " " );
    syslog ( LOG_NOTICE, "Deleting session %p", this );
    syslog ( LOG_NOTICE, " " );

    str = server_->find_session_name( this );

    if ( str.size() > 0 )
    {
        pid = server_->find_pndnaProxyPid( server_->find_session_name( this ));
        if ( pid != 0 )
        {
            syslog ( LOG_NOTICE, "Sending terminate signal to pndnaproxy process pid %i", pid );
            kill( pid, SIGTERM );
        }
    }

    server_->dec_session_count();
    server_->remove_session ( this );
}

//---------------------------------------------------------------------------
asio::ip::tcp::socket& pnsession::socket()
{
    return socket_;
}

pid_t pnsession::getPndnaProxyPid( void )
{
    return pndnaProxyPid;
}

void pnsession::setPndnaProxyPid( pid_t pid )
{
    pndnaProxyPid = pid;
}

//---------------------------------------------------------------------------
char * pnsession::data()
{
    return data_;
}

//---------------------------------------------------------------------------
pnserver * pnsession::server()
{
    return server_;
}


//---------------------------------------------------------------------------
void pnsession::start ()
{
    socket_.async_read_some( asio::buffer( data_, max_length ),
                             boost::bind( &pnsession::handle_read, this,
                                          asio::placeholders::error,
                                          asio::placeholders::bytes_transferred ) );
}

//---------------------------------------------------------------------------
void pnsession::handle_read( const asio::error_code& error, size_t bytes_transferred )
{
    if ( !error )
    {
        pthread_t tid;
        data_[bytes_transferred] = 0;

        if ( !strncmp( data_, "#PNTSC#", 7 ) )
        {
            // syslog (LOG_NOTICE, "Handling a GUI request: %s", &data_[7]);
            syslog ( LOG_NOTICE, "Handling a GUI request." );
            /*
                        if (!strncmp(&data_[7], "END", 3))
                        {
                            pthread_join(&tid);
                            server_->remove_session (this);
                            delete this;
            //                throw std::logic_error ("stop request");
                     }
            */
            pthread_create( &tid, NULL, proceed, this );

        }
        else
        {
            // syslog (LOG_NOTICE, "Handling non-GUI request: %s", data_);
            syslog ( LOG_NOTICE, "Handling non-GUI request" );
            server_->register_session ( std::string ( data_ ), this );
        }

        socket_.async_read_some( asio::buffer( data_, max_length ),
                                 boost::bind( &pnsession::handle_read, this,
                                              asio::placeholders::error,
                                              asio::placeholders::bytes_transferred ) );
    }
    else
    {
        syslog ( LOG_NOTICE, "Entering read error handling (%i) %s, %i", error.value(), error.message().c_str(), bytes_transferred );
        if ( error != asio::error::eof )
        {
            syslog ( LOG_NOTICE, "Requesting remove session in handle_read(). Also deleting the session object." );
        }
        else
        {
            syslog ( LOG_NOTICE, "Delete pnsession object" );
            delete this;
        }
    }

    syslog ( LOG_NOTICE, "Leaving handle_read()" );
}

//---------------------------------------------------------------------------
std::string pnsession::get_list ( const std::string& broker, int type )
{
    return server_->get_list ( broker, type );
}

//---------------------------------------------------------------------------
void pnsession::set_list ( const std::string& broker, const std::string& list, int type )
{
    server_->set_list ( broker, list, type );
}

void pnsession::use_proxy( char *addr, char *proxy )
{
    server_->use_proxy( addr, proxy );
}

//---------------------------------------------------------------------------
pnsession * pnsession::find_session ( const std::string& session_name )
{
    return server_->find_session ( session_name );
}

//---------------------------------------------------------------------------
void pnsession::send ( const char * message )
{
    try
    {
        asio::async_write( socket_,
                           asio::buffer( message, strlen ( message ) ),
                           boost::bind( &pnsession::handle_write, this,
                                        asio::placeholders::error ) );
    }
    catch ( std::exception& e )
    {
        syslog ( LOG_NOTICE, "send error: %s", e.what() );
        server_->remove_session ( this );
        delete this;
    }

}

//---------------------------------------------------------------------------
void pnsession::handle_write( const asio::error_code& error )
{
    if ( error )
    {
        syslog ( LOG_NOTICE, "write error: %s", error.message().c_str() );
        server_->remove_session ( this );
        delete this;
    }
}



//---------------------------------------------------------------------------
//    pnsession::utils
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
int pnsession::utils::find_option( char *haystack, const char *needle,
                                   char *buffer )
{
    buffer[0] = 0;
    if ( haystack == NULL ) return 0;
    char *p = strstr( haystack, needle ), *q;
    if ( p )
    {
        p += strlen( needle );
        while ( isspace( *p ) ) p++;
        if ( *p )
        {
            q = p;
            while ( *p && !isspace( *p ) ) p++;
            strncpy( buffer, q, p-q );
            return 1;
        }
    }
    return 0;
}

//---------------------------------------------------------------------------
int pnsession::utils::get_free_local_ip( unsigned int *ip )
{
    // try to find free looopback address in 127.0.0.[1 - 254] interval
    // int err_bind = 1;
    int i;
    *ip = 0;
    int s = ::socket ( AF_INET, SOCK_STREAM, 0 );

    setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &i, sizeof(i));

    if ( s >= 0 )
    {
        struct sockaddr_in addr;
        addr.sin_family = AF_INET;
        addr.sin_port = ::htons( 3389 );
        for ( i = 1; i < 254; ++i )
        {
            unsigned long a = ( unsigned long )( ( unsigned long )127<<24L ) |
                              ( ( unsigned long )i );
            addr.sin_addr.s_addr = ::htonl( a );
            if ( ::bind( s, ( struct sockaddr* ) &addr, sizeof( addr ) ) == 0 )
            {
                errno = 0;
                *ip = i;
                break;
            }
            syslog( LOG_INFO, "127.0.0.%i cannot be used", i );
        }
        ::close ( s );
    }
    return errno;
}

//---------------------------------------------------------------------------
bool pnsession::utils::get_one_broker( char* broker_list, char* onebroker )
{
    if ( !broker_list[0] )
    {
        onebroker[0] = 0;
        return false;
    }

    char* brokers[64];
    brokers[0] = broker_list;
    int count = 1;
    char* p;
    while ( ( p = strchr( brokers[count - 1], ',' ) ) )
    {
        *p = 0;
        brokers[count] = p + 1;
        count++;
    }
    if ( count == 1 )
    {
        strcpy( onebroker, broker_list );
        broker_list[0] = 0;
        return true;
    }

//    srand( time( 0 ) );
//    int c = rand() % count;
//    strcpy( onebroker, brokers[c] );
    strcpy( onebroker, brokers[0] );

    char buffer[1024];
    buffer[0] = 0;
    for ( int i = 1; i < count; i++ )
    {
//        if ( i != c )
//        {
            if ( buffer[0] )
                strcat( buffer, "," );
            strcat( buffer, brokers[i] );
//        }
    }
    strcpy( broker_list, buffer );
    return true;
}




//---------------------------------------------------------------------------
void pnsession::utils::wrap_brokers( char *broker_list, const char *protocol, const char *port )
{
    char *p = broker_list, *q;
    char broker_line[256];
    strcpy( broker_line, protocol );
    while ( ( q = strchr( p, ',' ) ) )
    {
        *q = 0;
        if ( !strncmp( p, "http://", 7 ) ) p += 7;
        else if ( !strncmp( p, "https://", 8 ) ) p += 8;
        strcat ( broker_line, p );
        strcat ( broker_line, ":" );
        strcat ( broker_line, port );
        strcat ( broker_line, "," );
        strcat ( broker_line, protocol );
        p = q+1;
    }
    strcat ( broker_line, p );
    strcat ( broker_line, ":" );
    strcat ( broker_line, port );
    strcpy( broker_list, broker_line );
}


//---------------------------------------------------------------------------
void pnsession::utils::get_printer_list (std::string & printerList)
{
    cups_dest_t *dests = NULL;
    int num_dests = cupsGetDests(&dests);

    // Default printer should go first in printers list
    // because in rdpdr channel first printer in the list is considered as default
    cups_dest_t *default_dest = cupsGetDest(NULL, NULL, num_dests, dests);
    if ((default_dest->instance == NULL) && (default_dest->name != NULL))
    {
        printerList += default_dest->name;
    }

    for (int i = 0; i < num_dests; i++)
    {
        if ((dests[i].instance == NULL) && !dests[i].is_default)
        {
            printerList += ',';
            printerList += dests[i].name;
        }
    }

    cupsFreeDests(num_dests, dests);
}


//---------------------------------------------------------------------------
char *  utf8_to_cp1252(char* in_str)
{
    const char * INSET = "UTF-8";
    const char * OUTSET = "CP1252";

    char * out_str = NULL;
    unsigned int out_len;
    unsigned int len;
    
    size_t iconv_value;
    iconv_t conv_desc;
    conv_desc = iconv_open (OUTSET, INSET);

    if ((int) conv_desc == -1)
    {
        if (errno == EINVAL)
        {
            syslog(LOG_NOTICE, "Conversion from '%s' to '%s' is not supported.\n", INSET, OUTSET);
        } else
        {
            syslog(LOG_NOTICE, "Initialization failure: %s\n", strerror (errno));
        }
        return NULL;
    }


    len = strlen(in_str);
    if (!len)
    {
        syslog(LOG_NOTICE, "Input string is empty.\n");
        iconv_close(conv_desc);
        return (NULL);
    }
    
    // Assign enough space for the out_str 
    out_len =  len * 4;
    char* out_start = out_str = (char*)malloc(out_len);
    memset(out_str, 0, out_len);

    iconv_value = iconv (conv_desc, &in_str, &len, &out_start, &out_len);

    if (iconv_value == (size_t) -1) 
    {
        syslog(LOG_NOTICE, "iconv failed\n");
        switch (errno)
        {
        case EILSEQ:
               syslog ( LOG_NOTICE,"Invalid multibyte sequence.\n");
            break;
        case EINVAL:
               syslog ( LOG_NOTICE,"Incomplete multibyte sequence.\n");
            break;
        case E2BIG:
               syslog ( LOG_NOTICE,"No more room.\n");
            break;
        default:
               syslog ( LOG_NOTICE,"Error: %s.\n", strerror (errno));
        }
        free(out_str);
        iconv_close(conv_desc);
        return NULL;
    }

    iconv_close(conv_desc);
    return out_str;
}

//---------------------------------------------------------------------------
void * pnsession::proceed ( void * param )
{

    char port[32] = "";
    char protocol[32] = "";
    char broker_line[BUFFER_4096] = "";
    char app_full[256] = "";
    char broker[256] = "";
    char buffer[BUFFER_4096] = "";
    // char *protocol_name[] = {"http://", "https://"};
    const char *layouts[MAX_LAYOUT_NUM] = {"local","ar", "common", "convert-map", "cs",
                                     "da", "de", "de-ch", "en-dv", "en-gb", "en-us", "es", "et", "fi", "fo",
                                     "fr", "fr-be", "fr-ca", "fr-ch", "he", "hr", "hu", "is", "it", "ja", "ko",
                                     "lt", "lv", "mk", "modifiers", "nl", "nl-be", "no", "pl", "pt", "pt-br",
                                     "ru", "sl", "sv", "th", "tr"
                                    };
    const char *constStr;
    
    char* popt_user         = NULL;
    char* popt_password     = NULL;
    char* popt_domain       = NULL;
    char* popt_app          = NULL;
    char* popt_new_password = NULL;
    
    char *user              = NULL; 
    char *password          = NULL;
    char *domain            = NULL;
    char *app               = NULL;
    
    char *sslgateway        = NULL, sslgateway_s[256] = "";
    char *proxyaddr         = NULL, proxyaddr_s[256]  = "";
    char *options_d         = NULL, options[1024]     = "";
    char *new_password      = NULL;  // For password management   char *pitfilename  = NULL;
    char *tgtfilename       = NULL;
    char *pitfilename       = NULL;

    char option[256];
    std::vector<char> ticket;
    ticket.resize(8192);
    char* pTicket = &ticket.front();
    SimpleString ssCookie, ssRes, ssPitFile;
    SimpleXML ssXMLRes;
    int nat = 0, wait = -1, retry = -1;
    int list = 0, List = 0, lista = 0, Lista = 0, listd = 0, Listd = 0;
    int firstDesktop = 0;
    int ipOnly = 0;
    int list_type = 0;
    unsigned int rdp5_performance_flags = RDP5_DISABLE_NOTHING;
    int ret = 0;

    unsigned char i;
    char *tgt = 0;

    int argc;
    const char **argv;

    int res = 0;

    // PNDNA options
    char pndnaOptions[BUFFER_1024];
    int freeLocalPort;


    char serverip[256];
    char servername[256];
    char shell[BUFFER_4096] = "\"";
    char workdir[128];

    int appType = 0;
    int seamless = 0;

    std::vector<char> response;
    response.resize(8192);
    char* pResponse = &response.front();
    *pResponse = '\0';

    


    
    poptContext optCon;   // context for parsing options
    struct poptOption optionsTable[] = {
        { NULL, 'u', POPT_ARG_STRING, &popt_user, 0, "user name (If not specified, the user will be prompted for domain name)" },
        { NULL, 'p', POPT_ARG_STRING, &popt_password, 0, "password (If not specified, the user will be prompted for domain name)" },
        { NULL, 'd', POPT_ARG_STRING, &popt_domain, 0, "domain (If not specified, the user will be prompted for domain name)" },
        { NULL, 'l', POPT_ARG_NONE, &list, 0, "display cached list of published applications and desktops" },
        { NULL, 'L', POPT_ARG_NONE, &List, 0, "query broker for list of published applications and desktops" },
        { "la",  0,  POPT_ARG_NONE|POPT_ARGFLAG_ONEDASH, &lista, 0, "display cached list of published applications" },
        { "La",  0,  POPT_ARG_NONE|POPT_ARGFLAG_ONEDASH, &Lista, 0, "query broker for list of published applications" },
        { "ld",  0,  POPT_ARG_NONE|POPT_ARGFLAG_ONEDASH, &listd, 0, "display cached list of published desktops" },
        { "Ld",  0,  POPT_ARG_NONE|POPT_ARGFLAG_ONEDASH, &Listd, 0, "query broker for list of published desktops" },
        { NULL, 'a', POPT_ARG_STRING, &popt_app, 0, "published application or desktop to launch" },
        { NULL, 'w', POPT_ARG_INT|POPT_ARGFLAG_SHOW_DEFAULT, &wait, 0, "retry interval" },
        { NULL, 'n', POPT_ARG_INT|POPT_ARGFLAG_SHOW_DEFAULT, &retry, 0, "number of retries" },
        { NULL, 't', POPT_ARG_NONE, &nat, 0, "enable NAT support for firewall traversal (for use with Terminal Services servers only)" },
        { NULL, 'o', POPT_ARG_STRING|POPT_ARGFLAG_OPTIONAL, &options_d, 0, "qrdesktop options" },
        { NULL, 'f', POPT_ARG_STRING, &pitfilename, 0, "pit-file name" },
        { NULL, 'T', POPT_ARG_STRING, &tgtfilename, 0, "TGT-file name" },
        { NULL, 'e', POPT_ARG_STRING, &sslgateway, 0, "SSL gateway name" },
        { NULL, 'c', POPT_ARG_STRING, &popt_new_password, 0, "change current password to this" },
        { NULL, 'P', POPT_ARG_STRING, &proxyaddr, 0, "Proxy address (in the format 'address:port')" },
        { "FD" ,  0,  POPT_ARG_NONE|POPT_ARGFLAG_ONEDASH, &firstDesktop, 0, "Return the name of the FIRST published Desktop" },
        { "ip" ,  0,  POPT_ARG_NONE|POPT_ARGFLAG_ONEDASH, &ipOnly, 0, "Return only the target host's IP address" },
        POPT_AUTOHELP
        POPT_TABLEEND
    };

    
    syslog( LOG_NOTICE, "proceed" );
    pnsession * current_session = ( pnsession * ) param;

    {   // to avoid compiler error "crosses initialization of std::..."

        char* pResponseData = current_session->data() + 1;
        pResponseData[5] = ' ';

        // parse request
        poptParseArgvString ( pResponseData, &argc, &argv );
        optCon = poptGetContext( NULL, argc, argv, optionsTable, 0 );
        if ( ( res = poptGetNextOpt( optCon ) ) != -1 )
        {
            snprintf (pResponse, BUFFER_131072,
                "[ERR%05X],[Getting request parameters],Incorrect request was given to connection server. %s: %s",
                INTERNAL_ERROR,
                poptBadOption(optCon, POPT_BADOPTION_NOALIAS),
                poptStrerror(res));
            syslog(LOG_NOTICE, "%s", pResponse);
            goto send_response;
        }
        if ( ( constStr = poptGetArg( optCon ) ) != NULL )
        {
            strcpy( broker, constStr );
        }
        free ( argv );
     
        
        if(popt_user)
        {
           user = utf8_to_cp1252(popt_user); //malloc
        }
 
        if(popt_password)
        {
           password = utf8_to_cp1252(popt_password); //malloc
        }
 
        if(popt_domain)
        {
           domain = utf8_to_cp1252(popt_domain); //malloc
        }
        
        if(popt_app)
        {
           app = utf8_to_cp1252(popt_app); //malloc
        }
        
        if(popt_new_password)
        {
           new_password = utf8_to_cp1252(popt_new_password);//malloc
        }

        if ( !user )
        {
            user     = (char*)malloc(64); 
            *user = 0;
        }
        
        if ( !password )
        {
            password = (char*)malloc(64); 
            *password = 0;
        }
        
        if ( !domain )
        {
            domain   = (char*)malloc(64);
            *domain = 0;
        }
        
        if ( !app )
        {
            app      = (char*)malloc(256);
            *app = 0;
        }
    

        if ( !sslgateway ) sslgateway = sslgateway_s;
        if ( !proxyaddr ) proxyaddr = proxyaddr_s;
        if ( options_d != NULL ) strcpy( options, options_d );

        if ( list ) list_type = 2;
        else if ( List ) list_type = 3;
        else if ( lista ) list_type = 4;
        else if ( Lista ) list_type = 5;
        else if ( listd ) list_type = 6;
        else if ( Listd || firstDesktop ) list_type = 7;

        // preparing the pit-file data
        if ( pitfilename != NULL )
        {
            syslog( LOG_NOTICE, "Reading pit file: %s", pitfilename );
            int pitfile = open ( pitfilename, O_RDONLY );
            if ( pitfile == -1 )
            {
                snprintf ( pResponse, BUFFER_131072, "[ERR%05X],[Getting request parameters],Cannot open pit-file %s: %s",
                    SYSTEM_ERROR, pitfilename, strerror ( errno ) );
                syslog(LOG_NOTICE, "%s", pResponse);
                goto send_response;
            }
            int pitfilesize = lseek ( pitfile, 0L, SEEK_END );
            char *pitbuf = ( char * ) malloc ( pitfilesize );
            lseek ( pitfile, 0L, SEEK_SET );
            ret = read ( pitfile, pitbuf, pitfilesize );
            if ( ret < 0 )
            {
                snprintf ( pResponse, BUFFER_131072, "[ERR%05X],[Getting request parameters],Cannot read pit-file %s: %s",
                    SYSTEM_ERROR, pitfilename, strerror ( errno ) );
                syslog(LOG_NOTICE, "%s", pResponse);
                goto send_response;
            }
            else if ( ret == 0 )
            {
                ret = pitfilesize;
            }
            ssPitFile.LoadData ( pitbuf, ret );
            close ( pitfile );
        }

//#ifdef DEBUG
//        syslog( LOG_NOTICE, "PITFILE: \n%s\n", ssPitFile.GetData() );
//        syslog( LOG_NOTICE, "Before checking for alternate shell, *app=%i, list_type=%i, new_password='%s'",
//                  *app, list_type, new_password );
//#endif

        if ( *app == 0 && list_type == 0 && (new_password == NULL || *new_password == 0 || strlen( new_password ) == 0))
        {
            char szApp[2050];
            syslog( LOG_NOTICE, "Parsing alternate shell option" );
            if ( ssPitFile.FindItem( "alternate shell:s:", szApp ) == 0 )
            {
                snprintf ( pResponse, BUFFER_131072,
                    "[ERR%05X],[Getting request parameters],Could not find the alternate shell option in the request",
                    INPUT_PARAMETERS_ERROR);
                syslog( LOG_NOTICE, "%s", pResponse );
                goto send_response;
            }
            unsigned char rgbBuffer1[1026];
            HexToBin( rgbBuffer1, sizeof( rgbBuffer1 ), szApp, strlen( szApp ) );
            Decrypt( rgbBuffer1, sizeof ( rgbBuffer1 ) );
            Decode1024( rgbBuffer1, szApp );
            if (szApp[0] == '\"')
            {
                strncpy(app, szApp+1, strlen(szApp)-2);
            }
            else
            {
                strncpy(app, szApp, strlen(szApp));
            }
        }

       
        if ( ssPitFile.FindItem( "apptype:i:", option ) && *option == '4' && *app != 0 )
        {
            const char openFileCommand[] = "xdg-open";

            struct stat st;

            if (stat(app , &st) != 0)
            {
                snprintf(pResponse, BUFFER_131072, "[ERR%05X],[Starting the session],Cannot find a local path: %s. Error code: %d", SYSTEM_ERROR, app, errno);
                goto send_response;
            }

            snprintf(pResponse, BUFFER_131072, "Success");
            current_session->send ( pResponse );

            syslog ( LOG_NOTICE, "Shell to start: %s ", app);
            execlp(openFileCommand, openFileCommand, app, NULL);

            snprintf(pResponse, BUFFER_131072, "[ERR%05X],[Starting the session],Unknown error launching: %s. Error code: %d", SYSTEM_ERROR, app, errno);
            goto send_response;
        }

        // If the ticket is in the pit file, then no need to
        // check for the credentials.
        if ( ssPitFile.FindItem( "ticket:s:", pTicket ) == 0 )
        {
            syslog( LOG_NOTICE, "No ticket info found." );
            if ( *broker == 0 )
            {
                syslog( LOG_NOTICE, "Checking broker information" );

                // search for a broker name in the pit-file
                char szTmp[10];
                if ( ssPitFile.FindItem( "llmprotocol:i:", szTmp ) == 0 ||
                        ssPitFile.FindItem( "llmservers:s:", broker ) == 0 ||
                        ssPitFile.FindItem( "llmtcpport:s:", port ) == 0 )
                {
                    snprintf ( pResponse, BUFFER_131072,
                        "[ERR%05X],[Getting request parameters],Could not find broker information in the request",
                        INPUT_PARAMETERS_ERROR);
                    syslog( LOG_NOTICE, "%s", pResponse );
                    goto send_response;
                }
                if ( strcmp( szTmp, "0" ) == 0 )
                {
                    strcpy( protocol, "https://" );
                }
                else
                {
                    strcpy( protocol, "http://" );
                }
                pResponseData = broker;
            }
            else
            {
                if ( !strncmp( broker, "http://", 7 ) )
                {
                    strcpy( protocol, "http://" );
                    pResponseData = broker + sizeof ( "http://" ) - 1;
                }
                else
                {
                    strcpy( protocol, "https://" );
                    pResponseData = broker + sizeof ( "https://" ) - 1;
                }
                char* pPort;
                if ( ( pPort = strchr( pResponseData, ':' ) ) == NULL )
                {
                    strcpy( port, "443" );
                }
                else
                {
                    strcpy( port, pPort + 1 );
                    *pPort = 0;
                }
            }
        }
        utils::wrap_brokers( pResponseData, protocol, port );
        strncpy(broker_line, pResponseData, 256);  //btw UNREAL exploit for reverser :)

        if ( tgtfilename == NULL )
        {
            syslog( LOG_NOTICE, "Checking credentials" );
            // If a ticket was specified, then don't bother looking for other credentials
            if ( ssPitFile.FindItem( "ticket:s:", pTicket ) == 0 )
            {
                syslog( LOG_NOTICE, "Checking username, password" );
                if ( ( ( *user == 0   && ssPitFile.FindItem( "username:s:", user ) == 0 ) ||
                        ( *domain == 0 && ssPitFile.FindItem( "domain:s:", domain ) == 0 ) ) )
                {
                    snprintf ( pResponse, BUFFER_131072,
                        "[ERR%05X],[Getting request parameters],Wrong credentials - username or domain is absent",
                        INPUT_PARAMETERS_ERROR);
                    syslog( LOG_NOTICE, "%s", pResponse );
                    goto send_response;
                }
                if ( *password == 0 )
                {
                    char szPassword[514];
                    if ( ssPitFile.FindItem( "password:s:", szPassword ) == 0 )
                    {
                        snprintf ( pResponse, BUFFER_131072,
                            "[ERR%05X],[Getting request parameters],Wrong credentials - password is absent",
                            INPUT_PARAMETERS_ERROR);
                        syslog( LOG_NOTICE, "%s", pResponse );
                        goto send_response;
                    }
                    // Decrypt the password.
                    unsigned char rgbBuffer[257];
                    HexToBin( rgbBuffer, sizeof( rgbBuffer ), szPassword, strlen( szPassword ) );
                    Decrypt( rgbBuffer, sizeof ( rgbBuffer ) );
                    Decode256( rgbBuffer, password );
                }
            }
        }

#ifdef DEBUG
        syslog (LOG_NOTICE, "brokers:%s  user:%s  password:%s  domain:%s  app:%s", broker_line, user, password, domain, app);
#endif // DEBUG


        //
        //-------  Start of move from lower code
        //

        // Initialize ssRes to keep it clean.
        ssRes = "";


        // TODO: Verify that I can read the broker info here. But only for application requests
        if (!list_type && NULL == new_password)
        {
            // If I have a ticket in the pitfile, then I can skip
            // this whole section about looking for a broker
            if ( ssPitFile.FindItem( "ticket:s:", pTicket ) == 0  )
            {
                // search for working broker
                char onebroker[256];
                while ( utils::get_one_broker( broker_line, onebroker ) )
                {
                    if ( *proxyaddr != 0 )
                    {
                        // add a routing entry into the dante.conf
                        if ( strncmp( onebroker, "http://", 7 ) == 0 )
                        {
                            char *p, tmp[256];
                            strncpy( tmp, onebroker+7, strlen( onebroker )-7 );
                            if ( ( p = strchr( tmp, ':' ) ) ) *p = '\0';
                            else tmp[strlen( onebroker )-7] = '\0';
                            current_session->use_proxy( tmp, proxyaddr );
                        }
                    }
                    syslog( LOG_NOTICE, "Connect to broker: %s", onebroker );
                    int num_attempts = retry;
                    ret = Authenticate( onebroker, proxyaddr, user, password, domain, tgt, nat, ssCookie );
                    while ( ( ret == PN_ERROR_VM_POWERING_UP ||
                              ret == PN_ERROR_VM_OS_STARTING ||
                              ret == PN_ERROR_BROKER_BUSY ) &&
                            --num_attempts > 0 )
                    {
                        if ( num_attempts == retry-1 )
                        {
                            snprintf ( pResponse, BUFFER_131072, "PENDING%s", ( char * )ssCookie );
                            syslog( LOG_NOTICE, "%s", pResponse );
                            current_session->send ( pResponse );
                        }
                        syslog( LOG_NOTICE, " sleeping.." );
                        sleep ( wait );
                        ret = Authenticate( onebroker, proxyaddr, user, password, domain, tgt, nat, ssCookie );
                    }
                    if ( ret != 0 )
                    {
                        snprintf ( pResponse, BUFFER_131072, "[ERR%05X],[Starting the session],%s",
                            ret, ( char * )ssCookie );
                        syslog( LOG_NOTICE, "%s", ( char * )pResponse );
                        if ( ret == 1326 /* Logon failure: unknown user name or password incorrect */ ||
                                ret == 1327 /* Logon failure: Restriction of user account. Possible reasons are blank passwords not allowed, restrictions on hours Logon or restriction st
    rategy was applied */ ||
                                ret == 1328 /* Logon failure: violation of restriction of hours of access */ ||
                                ret == 1329 /* Logon failure: user not authorized to log on to this computer */ ||
                                ret == 1330 /* Logon failure: the password for the specified account has expired */ ||
                                ret == 1793 /* The user account has expired */ )
                        {
                            goto send_response;
                        }
                        continue;
                    }
                    break;
                }
                if ( ret != 0 )
                {
                    snprintf ( pResponse, BUFFER_131072, "[ERR%05X],[Starting the session],%s",
                        ret, (char *) ssCookie );
                    goto send_response;
                }

                syslog( LOG_NOTICE, "Authentication passed" );
                // working with UI, the application name is always exactly the same from server
                strcpy( app_full, app );

                // here try to access to the broker
                ssRes = "";
                syslog( LOG_NOTICE, "Getting launch info, broker: %s", onebroker );
                int num_attempts = retry;
                ret = GetLaunchInfo( onebroker, proxyaddr, ssCookie, app_full, ssRes );
                while ( ( ret == PN_ERROR_VM_POWERING_UP ||
                          ret == PN_ERROR_VM_OS_STARTING ||
                          ret == PN_ERROR_BROKER_BUSY ) && --num_attempts > 0 )
                {
                    snprintf ( pResponse, BUFFER_131072, "PENDING\nPOWERING_UP\n%d\n%d\n%s", wait, num_attempts, ( char * )ssRes );
                    syslog( LOG_NOTICE, "%s", pResponse );
                    current_session->send ( pResponse );

                    syslog( LOG_NOTICE, " sleeping.." );
                    sleep ( wait );
                    ret = GetLaunchInfo( onebroker, proxyaddr, ssCookie, app_full, ssRes );
                }
                if ( ret != 0 )
                {
                    snprintf ( pResponse, BUFFER_131072, "[ERR%05X],[Getting launch information from broker],%s",
                        ret, ( char * )ssRes );
                    goto send_response;
                }
                // creating start application info string
                ssRes.FindItem( "servername:s:", servername );

#ifdef DEBUG
                syslog( LOG_NOTICE, "Internal routine found servername:%s", servername );
#endif

                ssRes.FindItem( "serveraddr:s:", serverip );
                ssRes.FindItem( "alternate shell:s:", shell+1 );
                ssRes.FindItem( "shell working directory:s:", workdir );
                ssRes.FindItem( "ticket:s:", pTicket );
                if ( ssRes.FindItem( "apptype:i:", option ) )
                {
                    appType = strtol( option, NULL, 10 );
                }

            }

        }
              
        if( ssRes.FindItem ( "connecttoconsole:i:", option ) && *option != 0 && *option != '0')
        {
            syslog( LOG_INFO, "enabling console mode" );
            strcat( options, " -0" );
        }
        
        //
        //-------  END of move from lower code
        //

        if ( !nat )
        {
            char sznat[32];
            if ( ssPitFile.FindItem( "llmenablenat:i:", sznat ) )
            {
                nat = atoi( sznat );
            }
        }
        if ( *proxyaddr == 0 )
        {
            ssPitFile.FindItem( "proxyaddr:s:", proxyaddr );
        }
        
        
        if ( ( list_type > 0 ) || ( firstDesktop ) ) goto proceed;

        // try to find other unset options in the pit-file
        if ( ssPitFile.Length() != 0 )
        {
            if ( *sslgateway == 0 )
            {
                if (ssPitFile.FindItem( "enablessl:i:", option ) && *option != '0')
                {
                    ssPitFile.FindItem( "sslgateway:s:", sslgateway );
                }
            }

            if (wait == -1 && ssPitFile.FindItem( "wait:i:", option ) && *option != '0')
            {
                wait = strtol(option, NULL, 10);
            }

            if (retry == -1 && ssPitFile.FindItem( "retry:i:", option ) && *option != '0')
            {
                retry = strtol(option, NULL, 10);
            }
            
            // Extract the apptype and seamless parameters from the PIT file.
            if ( ssPitFile.FindItem( "apptype:i:", option ) )
            {
                appType = strtol( option, NULL, 10 );
            }
            if ( ssPitFile.FindItem( "seamless:i:", option ) )
            {
                seamless = strtol( option, NULL, 10 );
            }

            // If the options are app type 1 and seamless ON, then do it.
            if (( appType == 1 ) && ( seamless != 0 ))
            {
                strcat( options, " -A " );
            }
            else
            {
                if ( utils::find_option( options, "-g", option ) == 0 )
                {
                    if ( ssPitFile.FindItem( "windowSize:s:", option ) )
                    {
                        if ( strcmp( option, "fullscreen" ) != 0 )
                        {
                            strcat( options, " -g " );
                            strcat( options, option );
                        }
                        else
                        {
                            strcat( options, " -f " );
                            if (ssPitFile.FindItem( "spanmonitors:i:", option ) && *option != '0')
                            {
                                strcat ( options, " --span-monitors " );
                            }
                        }
                    }
                    // work with broker-generated pit files
                    char desktop_height[32] = { 0 };
                    if (ssPitFile.FindItem( "desktopwidth:i:", option ) && *option != 0 && *option != '0' &&
                        ssPitFile.FindItem( "desktopheight:i:", desktop_height ) && *desktop_height != 0 && *desktop_height != '0')
                    {
                        char desktop_size[32] = { 0 };
                        snprintf(desktop_size, 32, " -g %sx%s", option, desktop_height);
                        strcat( options, desktop_size);
                    }
                    if ( ssPitFile.FindItem( "screen mode id:i:", option ) && *option == '2' )
                    {
                        strcat( options, " -f " );
                    }
                }

                if ( utils::find_option( options, "-a", option ) == 0 )
                {
                    if ( ssPitFile.FindItem( "session bpp:s:", option ) && *option != '0' )
                    {
                        if ( strcmp( option, "auto" ) != 0 )
                        {
                            strcat( options, " -a " );
                            strcat( options, option );
                        }
                    }
                    if ( ssPitFile.FindItem( "session bpp:i:", option ) && *option != '0' )
                    {
                        strcat( options, " -a " );
                        strcat( options, option );
                    }
                }
            }

            if ( utils::find_option( options, "-n", option ) == 0 )
            {
                if ( ssPitFile.FindItem( "clientName:s:", option ) && *option != 0 )
                {
                    strcat( options, " -n " );
                    strcat( options, "\"");
                    strcat( options, option );
                    strcat( options, "\"");
                }
            }
            if ( utils::find_option( options, "-k", option ) == 0 )
            {
                if ( ssPitFile.FindItem( "keyboardLayout:i:", option ) && *option != '0' )
                {
                    int layout = atoi( option );
                    if ( layout >= 0 && layout < MAX_LAYOUT_NUM )
                    {
                        strcat( options, " -k " );
                        strcat( options, layouts[layout] );
                    }
                }
            }
            if ( ssPitFile.FindItem( "disable themes:i:", option ) && *option != '0' )
            {
                rdp5_performance_flags |= RDP5_NO_THEMING;
            }
            if ( ssPitFile.FindItem( "disable menu anims:i:", option ) && *option != '0' )
                    {
                rdp5_performance_flags |= RDP5_NO_MENUANIMATIONS;
                    }
            if ( ssPitFile.FindItem( "disable full window drag:i:", option ) && *option != '0' )
                    {
                rdp5_performance_flags |= RDP5_NO_FULLWINDOWDRAG;
                }
            if ( ssPitFile.FindItem( "disable wallpaper:i:", option ) && *option != '0' )
            {
                rdp5_performance_flags |= RDP5_NO_WALLPAPER;
            }
            if ( ssPitFile.FindItem( "bitmapcachepersistenable:i:", option ) && *option != '0' )
            {
                strcat( options, " -P " );
            }
            if ( ssPitFile.FindItem( "reconnect:i:", option ) && *option != '0' )
            {
                strcat( options, " -J " );
            }
            if (ssPitFile.FindItem( "sendingMouse:i:", option ) && *option != '0' )
            {
                strcat( options, " -m " );
            }
            if (ssPitFile.FindItem( "backingStore:i:", option ) && *option != '0' )
            {
                strcat( options, " -B " );
            }
            // either this
            if ( ssPitFile.FindItem( "sound:s:", option ) )
            {
                strcat( options, " -r sound:" );
                strcat( options, option );
            }
            // or this
            // I'm not sure what this option means, for now let it be the mode
            // of the sound redirection
            if ( ssPitFile.FindItem( "audiomode:i:", option ) && *option != '0')
            {
                strcat( options, " -r sound:off" );
            }
            if ( ssPitFile.FindItem( "drives:s:", option ) && *option != 0 )
            {
                std::vector<char *> drives;
                char * pch;
                pch = strtok (option,",");
                while (pch != NULL)
                {
                    drives.push_back(pch);
                    pch = strtok (NULL, ",");
                }

                for (unsigned int i = 0; i < drives.size(); i++ )
                {
                    strcat( options, " -r disk:" );
                    strcat( options, drives.at( i ) );
                }
            }
            if ( ssPitFile.FindItem( "lptports:s:", option ) && *option != 0 )
            {
                std::vector<char *> lptports = splitString( option );
                for (unsigned int i = 0; i < lptports.size(); i++ )
                {
                    strcat( options, " -r lptport:" );
                    strcat( options, lptports.at( i ) );
                }
            }
            if ( ssPitFile.FindItem( "comports:s:", option ) && *option != 0 )
            {
                std::vector<char *> comports = splitString( option );
                for (unsigned int i = 0; i < comports.size(); i++ )
                {
                    strcat( options, " -r comport:" );
                    strcat( options, comports.at( i ) );
                }
            }

            if ( ssPitFile.FindItem( "redirectprinters:i:", option ) && *option == '1' )
            {
                strcat( options, " -r printer" );
                syslog(LOG_NOTICE, "Turning Printer Redirection on");

                std::string printerList;
                utils::get_printer_list(printerList);
                if (!printerList.empty())
                {
                    strcat( options, ":" );
                    strcat( options, printerList.c_str());
                }
            }

            if ( ssPitFile.FindItem( "redirectsmartcards:i:", option ) && *option == '1' )
            {
                strcat( options, " -r scard");
                syslog(LOG_NOTICE, "Turning Smartcard Redirection on");
            }

            if ( ssPitFile.FindItem( "redirectdrives:i:", option ) && *option == '1' )
            {
                // get info about disks in the system
                FILE *pOut = popen ("mount", "r");
                if (pOut != NULL)
                {
                    char str[1024] = {0};
                    while ( fgets ( str, 1024, pOut ) )
                    {
                        if (str[0] == '/' && str[1] != '/') // exclude network shares
                        {
                            char *p = str;
                            while ( !isspace(*p)) ++p; // device name kinda /dev/sda1
                            while ( isspace(*p)) ++p; // space
                            while ( !isspace(*p)) ++p; // 'on'
                            while ( isspace(*p)) ++p; // space
                            if ( isspace(*(p+1)) )
                            {
                                // this is root disk
                                strcat( options, " -r disk:root=/" );
                            }
                            else
                            {
                                char *q = p;
                                while( !isspace(*q)) ++q;
                                *q = '\0';
                                q = strrchr(p, '/'); ++q;
                                strcat( options, " -r disk:" );
                                strcat( options, q );
                                strcat( options, "=" );
                                strcat( options, p );
                            }
                        }
                    }
                    pclose(pOut);
                }
            }
            if ( ssPitFile.FindItem( "redirectclipboard:i:", option ) && *option != 0 )
            {
                strcat( options, " -r clipboard:" );
                if ( *option == '0' )
                    strcat( options, "off" );
                else
                    strcat( options, "CLIPBOARD" );
            }

            if ( ssPitFile.FindItem( "redirecthandhelds:i:", option ) && *option != 0 )
            {
                // If the USBRDR option is set to '1'
                if ( *option == '1' )
                {
                    strcat( options, " -r usbrdr" );
                    syslog ( LOG_NOTICE, "Turning USB Redirection on");
                }
            }

            // Handle the graphics acceleration flag
            if ( ssPitFile.FindItem( "bitmapacceleration:i:", option ) && *option != 0 )
            {
                if ( *option == '1' )
                {
                    strcat( options, " --gaenable" );
                    syslog ( LOG_NOTICE, "Turning Graphics Acceleration on");
                }
            }

            // MMR Options are now stored in the rdp file.  read them in and set them
            if ( ssPitFile.FindItem( "multimediaredirection:i:", option ) && *option != 0 )
            {
                // Okay, so the mmr option exists.  Now handle it based on 1 or 0
                //
                if ( *option  == '1' )
                {

                    syslog ( LOG_NOTICE, "Turning Multimedia Redirection on");

                    // If the options are app type 1 and seamless ON, then do it.
                    if (( appType == 1 ) && ( seamless != 0 ))
                #ifdef HP_OEM
                    {
                        strcat( options, " -r addin:HPRCtrl:mmrproxy:HPRCtrl:1"\
                                         " -r addin:HPWCtrl:mmrproxy:HPWCtrl:1"\
                                         " -r addin:HPRStrm:mmrproxy:HPRStrm:1"\
                                         " -r addin:HPWStrm:mmrproxy:HPWStrm:1 " );

                    }
                    else
                    {
                        strcat( options, " -r addin:HPRCtrl:mmrproxy:HPRCtrl"\
                                         " -r addin:HPWCtrl:mmrproxy:HPWCtrl"\
                                         " -r addin:HPRStrm:mmrproxy:HPRStrm"\
                                         " -r addin:HPWStrm:mmrproxy:HPWStrm " );

                    }
                #else
                    {
                        strcat( options, " -r addin:PNRCtrl:mmrproxy:PNRCtrl:1"\
                                         " -r addin:PNWCtrl:mmrproxy:PNWCtrl:1"\
                                         " -r addin:PNRStrm:mmrproxy:PNRStrm:1"\
                                         " -r addin:PNWStrm:mmrproxy:PNWStrm:1 " );

                    }
                    else
                    {
                        strcat( options, " -r addin:PNRCtrl:mmrproxy:PNRCtrl"\
                                         " -r addin:PNWCtrl:mmrproxy:PNWCtrl"\
                                         " -r addin:PNRStrm:mmrproxy:PNRStrm"\
                                         " -r addin:PNWStrm:mmrproxy:PNWStrm " );
                    }
                #endif
                }
            }
            if ( ssPitFile.FindItem( "flashredirection:i:", option ) && *option != 0 )
            {
                if ( *option  == '1' && isFlashPluginInstalled() )
                {
                    syslog ( LOG_NOTICE, "Turning Flash Redirection on");

                    // If the options are app type 1 and seamless ON, then do it.
                    if (( appType == 1 ) && ( seamless != 0 ))
                    {
                        // to run FMMR client in seamless mode add any arg:
                        #ifdef HP_OEM
                        strcat( options, " -r addin:HPFlash:fmmr:1 " );
                        #else
                        strcat( options, " -r addin:PNFlash:fmmr:1 " );
                        #endif
                    }
                    else
                    {
                        #ifdef HP_OEM
                        strcat( options, " -r addin:HPFlash:fmmr " );
                        #else
                        strcat( options, " -r addin:PNFlash:fmmr " );
                        #endif
                    }
                }
            }
            if ( ssPitFile.FindItem( "downstreamAudioEnabled:i:", option ) && *option != 0 && *option != '0' )
            {
                strcat( options, " -r qds:on" );
                if ( ssPitFile.FindItem( "daMode:s:", option ) )
                {
                    *option = tolower( *option );
                    strcat( options, ":" );
                    if ( strcmp( option, "do not play" ) == 0 )
                        strcat( options, "donotplay" );
                    else
                        strcat( options, option );
                }
                if ( ssPitFile.FindItem( "daQuality:s", option ) )
                {
                    strcat( options, option );
                    if ( strcmp( option, ":custom" ) == 0 )
                    {
                        ssPitFile.FindItem( "daCustomQuality:s:", option );
                        strcat( options, "=" );
                        if (*option != 0 && *option != '0' )
                            strcat( options, option );
                        else
                            strcat( options, "128");
                    }
                    if ( ssPitFile.FindItem( "daBuffering:s:", option ) && *option != 0 && *option != '0' )
                    {
                        strcat( options, ":buffering=" );
                        strcat( options, option );
                    }
                }
            }
            else
                strcat( options, " -r qds:off" );

            if ( ssPitFile.FindItem( "upstreamAudioEnabled:i:", option ) && *option != 0 && *option != '0' )
            {
                ssPitFile.FindItem( "uaMode:s:", option );
                strcat( options, " -r qus:" );
                strcat( options, option );
                if ( ( strcmp( option, "speex" ) == 0 ) && ssPitFile.FindItem( "uaQuality:s", option ) )
                {
                    strcat( options, option );
                    if ( ssPitFile.FindItem( "uaComplexity:s", option ) )
                    {
                        strcat( options, option );
                        if ( ssPitFile.FindItem( "uaEnhancedCompression:i", option ) )
                        {
                            strcat( options, ":enh=" );
                            if ( *option == 0 )
                                strcat( options, "off" );
                            else
                                strcat( options, "on" );
                        }
                    }
                }
            }

            if ( ssPitFile.FindItem( "redirectmicrophone:i:", option ) && *option != 0 && *option != '0' )
            {
                syslog( LOG_INFO, "Turning on Upstream Audio" );
                strcat( options, " -r qus:wave" );
            }



            // Testing cursor hack
            strcat( options, " -M" );

            // Workaround for Windows Server 2008 and license hack, passing -l option to qrdesktop if pit file
            strcat(options, " -l");

            // Turn on rdp compression
            if ( ssPitFile.FindItem( "compression:i:", option ) && *option != 0 && *option != '0' )
            {
                syslog( LOG_INFO, "Received request to turn on Protocol Compression." );
                strcat( options, " -z" );
            }

        }

        if ( wait == -1 )
        {
            wait = 5;
        }
        if ( retry == -1 )
        {
            retry = 20;
        }

        char rdp5_performance_flags_str[32];
        snprintf (rdp5_performance_flags_str, 32, " -x %.8x", rdp5_performance_flags);
        strcat (options, rdp5_performance_flags_str);

proceed:

        syslog ( LOG_NOTICE, "qrdesktop options: %s", options );
        syslog ( LOG_NOTICE, "proxy: %s", proxyaddr );
        syslog ( LOG_NOTICE, "SSL gateway: %s", sslgateway );

#ifdef DEBUG
        syslog ( LOG_NOTICE, "creds: %s %s %s", user, password, domain);
#endif

        syslog ( LOG_NOTICE, "Proceeding" );

        if ( tgtfilename )
        {
            tgt = ::create_tgt_response( tgtfilename );
            if ( tgt == 0 )
            {
                snprintf ( pResponse, BUFFER_131072, "[ERR%05X],[Getting request parameters],Reading TGT file %s: %s",
                    SYSTEM_ERROR, tgtfilename, pResponse );
                goto send_response;
            }
        }

        if ( ( list_type > 0 ) || ( firstDesktop ) )
        {
            int type = list_type / 2 - 1;
            char onebroker[256];
            while ( utils::get_one_broker( broker_line, onebroker ) )
            {
                syslog( LOG_NOTICE, "Connect to broker: %s", onebroker );
                // construct a key from the broker address and the user name
                std::string broker_str ( onebroker );
                broker_str.append ( user );
                if ( ( list_type & 1 ) || ( current_session->get_list( broker_str, type ) == "" ) )
                {
                    if ( *proxyaddr != 0 )
                    {
                        // add a routing entry into the dante.conf
                        if ( strncmp( onebroker, "http://", 7 ) == 0 )
                        {
                            char *p, tmp[256];
                            strncpy( tmp, onebroker+7, strlen( onebroker )-7 );
                            if ( ( p = strchr( tmp, ':' ) ) ) *p = '\0';
                            else tmp[strlen( onebroker )-7] = '\0';
                            current_session->use_proxy( tmp, proxyaddr );
                        }
                    }
                    // getting a list of applications from given broker address
                    if ( ( ret = Authenticate( onebroker, proxyaddr, user, password, domain, tgt, nat, ssCookie ) ) != 0 )
                    {
                        snprintf ( pResponse, BUFFER_131072, "[ERR%05X],[Getting application list],%s",
                            AUTHENTICATION_ERROR, ( char * )ssCookie );
                        syslog( LOG_NOTICE, "%s", pResponse );

                        if ( ret == 1326 /* Logon failure: unknown user name or password incorrect */ ||
                                ret == 1327 /* Logon failure: Restriction of user account. Possible reasons are blank passwords not allowed, restrictions on hours Logon or restri ction strategy was applied */ ||
                                ret == 1328 /* Logon failure: violation of restriction of hours of access */ ||
                                ret == 1329 /* Logon failure: user not authorized to log on to this computer */ ||
                                ret == 1330 /*Logon failure: the password for the specified account has expired */ ||
                                ret == 1793 /* The user account has expired */ )
                        {
                            goto send_response;
                        }

                        continue;
                    }


                    if ( ( ret = GetServerList( onebroker, proxyaddr, ( char* )ssCookie, ssRes ) ) != 0 )
                    {
                        snprintf ( pResponse, BUFFER_131072, "[ERR%05X],[Getting server list],%s",
                            BROKER_ERROR, ( char * ) ssRes );
                        syslog( LOG_NOTICE, "%s", pResponse );
                        continue;
                    }
                    else
                    {
                        syslog( LOG_INFO, "GetServerList response: %s", ssRes.GetData());
                        // save the server list to a temp file for reading by pntsc.

                        FILE * ofile;
                        ofile = fopen("/tmp/farminfo", "w+" );
                        fprintf( ofile, "%s", ssRes.GetData() );
                        fclose( ofile );


                    }


                    if ( ( ret = GetApplicationList( onebroker, proxyaddr, ( char* )ssCookie, ssRes, type ) ) != 0 )
                    {
                        snprintf ( pResponse, BUFFER_131072, "[ERR%05X],[Getting application list],%s",
                            BROKER_ERROR, ( char * ) ssRes );
                        syslog(LOG_NOTICE, "Get app list error: %s", pResponse);
                        continue;
                    }

                    pResponseData = ssRes.GetData();
                    std::vector<char> appData;
                    appData.reserve(81920); // ~1000 apps, 8kb ~100 applications in response
                    char* pEndOfResponseUnit;
                    while (pEndOfResponseUnit = strchr(pResponseData, '\n'))
                    {
                        // For some reason this MUST be around this value or greater to
                        // allow the list to be parsed properly.  Do not have the time to deal
                        // with this fully right now.
                        usleep(5000);

                        if (appData.capacity() - appData.size() <= pEndOfResponseUnit - pResponseData )
                        {
                            appData.reserve(appData.size() * 2);
                        }

                        if (strncmp(pResponseData, "Name=", strlen("Name=")) == 0)
                        {
                            appData.insert(appData.end(), pResponseData, pEndOfResponseUnit);
                            memset( buffer, 0x0, 1024 );
                            strncpy( buffer, pResponseData, pEndOfResponseUnit - pResponseData );

                            strcpy( pResponse, strchr( buffer, '=' ) + 1 );

                            // If option D, then kick out now with just the one as the response
                            if ( firstDesktop )
                            {
                                goto send_response;
                            }

                            appData.insert(appData.end(),  ',');
                        }
                        if (strncmp(pResponseData, "ShortCut=", strlen("ShortCut=")) == 0)
                        {
                            appData.insert(appData.end(), pResponseData, pEndOfResponseUnit);
                            appData.insert(appData.end(),  ',');
                        }
                        if (strncmp(pResponseData, "FolderName=", strlen("FolderName=")) == 0)
                        {
                            appData.insert(appData.end(), pResponseData, pEndOfResponseUnit);
                            appData.insert(appData.end(),  ',');
                        }
                        if (strncmp(pResponseData, "LargeIcon=", strlen("LargeIcon=")) == 0)
                        {
                            appData.insert(appData.end(), pResponseData, pEndOfResponseUnit);
                            appData.insert(appData.end(),  ',');
                        }
                        if (strncmp(pResponseData, "SmallIcon=", strlen("SmallIcon=")) == 0)
                        {
                            appData.insert(appData.end(), pResponseData, pEndOfResponseUnit);
                            appData.insert(appData.end(),  '\n');
                        }
                        pResponseData = pEndOfResponseUnit + 1;
                    }

                    appData.insert(appData.end(),  '\0');

                    std::string s(appData.begin(), appData.end());
                    current_session->set_list( broker_str, s, type );

                }   // refreshing a list of application

                strcpy ( pResponse, ( current_session->get_list( broker_str, type ) ).c_str() );
                syslog( LOG_NOTICE, "Response: %i bytes", strlen(pResponse)  );
                goto send_response;
            }

            goto send_response;
        }

        // Password management
        if ( new_password != NULL )
        {
            char onebroker[256];
            while ( utils::get_one_broker( broker_line, onebroker ) )
            {
                syslog( LOG_NOTICE, "Connect to broker: %s", onebroker );

                // Change password request
                if ( ( ret = ChangePassword( onebroker, user, domain, password, new_password,
                                             ( char * ) ssCookie, ssRes ) ) != 0 )
                {
                    snprintf( pResponse, BUFFER_131072, "[ERR%05X],[Changing the password],%s",
                        BROKER_ERROR, ( char * ) ssRes );
                    syslog(LOG_NOTICE, "Change password error: %s", pResponse);
                    continue;
                }

                strcpy( pResponse, ( char * ) ssRes );
                break;
            }

            goto send_response;
        }

        // -----------------------------------
        // starting application
        // -----------------------------------

        // NOTE: Do not need to run the auth here as it should already have been done.

/*
        char servername[256];
        char serverip[256];
        char shell[4096] = "\"";
        char workdir[128];
        char ticket[1024];
        char appType[8];
*/

        // If I have a ticket in the pitfile, then I can skip
        // this whole section about looking for a broker
        if ( ssPitFile.FindItem( "ticket:s:", pTicket ) == 0  )

        {
            if ( 0 ) {

            // search for working broker
            char onebroker[256];
            while ( utils::get_one_broker( broker_line, onebroker ) )
            {
                if ( *proxyaddr != 0 )
                {
                    // add a routing entry into the dante.conf
                    if ( strncmp( onebroker, "http://", 7 ) == 0 )
                    {
                        char *p, tmp[256];
                        strncpy( tmp, onebroker+7, strlen( onebroker )-7 );
                        if ( ( p = strchr( tmp, ':' ) ) ) *p = '\0';
                        else tmp[strlen( onebroker )-7] = '\0';
                        current_session->use_proxy( tmp, proxyaddr );
                    }
                }
                syslog( LOG_NOTICE, "Connect to broker: %s", onebroker );
                int num_attempts = retry;
                ret = Authenticate( onebroker, proxyaddr, user, password, domain, tgt, nat, ssCookie );
                while ( ( ret == PN_ERROR_VM_POWERING_UP ||
                          ret == PN_ERROR_VM_OS_STARTING ||
                          ret == PN_ERROR_BROKER_BUSY ) &&
                        --num_attempts > 0 )
                {
                    if ( num_attempts == retry-1 )
                    {
                        snprintf ( pResponse, BUFFER_131072, "PENDING%s", ( char * )ssCookie );
                        syslog(LOG_NOTICE, "%s", pResponse);
                        current_session->send (pResponse);
                    }
                    syslog( LOG_NOTICE, " sleeping.." );
                    sleep ( wait );
                    ret = Authenticate( onebroker, proxyaddr, user, password, domain, tgt, nat, ssCookie );
                }
                if ( ret != 0 )
                {
                    snprintf ( pResponse, BUFFER_131072, "[ERR%05X],[Starting the session],%s",
                        ret, ( char * )ssCookie );
                    syslog(LOG_NOTICE, "%s", (char *)pResponse);
                    if ( ret == 1326 /* Logon failure: unknown user name or password incorrect */ ||
                            ret == 1327 /* Logon failure: Restriction of user account. Possible reasons are blank passwords not allowed, restrictions on hours Logon or restriction st
rategy was applied */ ||
                            ret == 1328 /* Logon failure: violation of restriction of hours of access */ ||
                            ret == 1329 /* Logon failure: user not authorized to log on to this computer */ ||
                            ret == 1330 /* Logon failure: the password for the specified account has expired */ ||
                            ret == 1793 /* The user account has expired */ )
                    {
                        goto send_response;
                    }
                    continue;

                }
                break;
            }
            if ( ret != 0 )
            {
                snprintf ( pResponse, BUFFER_131072, "[ERR%05X],[Starting the session],%s",
                    ret, (char *) ssCookie );
                goto send_response;
            }

            syslog( LOG_NOTICE, "Authentication passed" );
            // working with UI, the application name is always exactly the same from server
            strcpy( app_full, app );

            // here try to access to the broker
            ssRes = "";
            syslog( LOG_NOTICE, "Getting launch info, broker: %s", onebroker );
            int num_attempts = retry;
            ret = GetLaunchInfo( onebroker, proxyaddr, ssCookie, app_full, ssRes );
            while ( ( ret == PN_ERROR_VM_POWERING_UP ||
                      ret == PN_ERROR_VM_OS_STARTING ||
                      ret == PN_ERROR_BROKER_BUSY ) && --num_attempts > 0 )
            {
                snprintf ( pResponse, BUFFER_131072, "PENDING\nPOWERING_UP\n%d\n%d\n%s", wait, num_attempts, ( char * )ssRes );
                syslog( LOG_NOTICE, "%s", pResponse );
                current_session->send ( pResponse );

                syslog( LOG_NOTICE, " sleeping.." );
                sleep ( wait );
                ret = GetLaunchInfo( onebroker, proxyaddr, ssCookie, app_full, ssRes );
            }
            if ( ret != 0 )
            {
                snprintf ( pResponse, BUFFER_131072, "[ERR%05X],[Getting launch information from broker],%s",
                    ret, ( char * )ssRes );
                goto send_response;
            }
            // creating start application info string
            ssRes.FindItem( "servername:s:", servername );

#ifdef DEBUG
            syslog( LOG_NOTICE, "Internal routine found servername:%s", servername );
#endif

            ssRes.FindItem( "serveraddr:s:", serverip );
            ssRes.FindItem( "alternate shell:s:", shell+1 );
            ssRes.FindItem( "shell working directory:s:", workdir );
            ssRes.FindItem( "ticket:s:", pTicket );
            if ( ssRes.FindItem( "apptype:i:", option ) )
            {
                appType = strtol( option, NULL, 10 );
            }
        }
        }
        else
        {
            // creating start application info string
            ssPitFile.FindItem( "servername:s:", servername );
            ssPitFile.FindItem( "serveraddr:s:", serverip );
            ssPitFile.FindItem( "shell working directory:s:", workdir );
            ssPitFile.FindItem( "ticket:s:", pTicket );

            // Already decrypted, see above
            strncpy( shell + 1, app, 4095 );
        }

        // no shell, it's a desktop, using explorer.exe as shell
        if ( shell[1] == 0 )
        {
            strcpy ( shell+1, "explorer.exe" );
        }

        // If the apptype is client content.
        if (appType == 4)
        {
            const char openFileCommand[] = "xdg-open";

            struct stat st;

            if (stat(shell +1 , &st) != 0)
            {
                snprintf(pResponse, BUFFER_131072, "[ERR%05X],[Starting the session],Cannot find a local path: %s. Error code: %d", SYSTEM_ERROR, shell + 1, errno);
                goto send_response;
            }

            snprintf(pResponse, BUFFER_131072, "Success");
            current_session->send ( pResponse );

            syslog ( LOG_NOTICE, "Shell to start: %s ", shell + 1);
            execlp(openFileCommand, openFileCommand, shell + 1, NULL);

            snprintf(pResponse, BUFFER_131072, "[ERR%05X],[Starting the session],Unknown error launching: %s. Error code: %d", SYSTEM_ERROR, shell + 1, errno);
            goto send_response;
        }

        strcat( shell, "\"" );

        syslog ( LOG_NOTICE, "Launch info: %s %s %s %s", servername, serverip, shell, workdir );
        char *p1, *q1;
        char* endOfResponseUnit;
        pResponseData = servername;
        p1 = serverip;
        std::string session_name;
        pnsession * session = NULL;
        while (endOfResponseUnit = strchr(pResponseData, ','))
        {
            // try to find existing session connected to given remote host
            // if succeed share the session, otherwise create a new one
            *endOfResponseUnit = '\0';
            *( q1 = strchr( p1, ',' ) ) = '\0';
            session_name = pResponseData;
            session = current_session->find_session ( session_name );
            if ( session != NULL ) break;
            pResponseData = endOfResponseUnit + 1;
            p1 = q1 + 1;
        }
        if ( session == NULL )
        {
            session_name = pResponseData;
            session = current_session->find_session ( session_name );
        }

        if ( session == NULL )
        {
            unsigned int ip;

            session_name = servername;

            if ( utils::get_free_local_ip( &ip ) != 0 )
            {
                snprintf ( pResponse, BUFFER_131072, "[ERR%05X],[Starting the session],Cannot get a local IP: %s",
                          SYSTEM_ERROR, strerror( i ) );
                syslog ( LOG_NOTICE, "%s", pResponse );
                goto send_response;
            }

            char local_addr[64];
            snprintf ( local_addr, 64, "127.0.0.%d", ip );
            syslog ( LOG_NOTICE, "Creating a new session: %s", session_name.c_str() );

            if ( tgt )  // if TGT file is present it's preferable to use TGT
            {
                snprintf ( pResponse, BUFFER_131072, "STARTqrdesktop\n-G\n%s\n-s\n%s\n", tgtfilename, shell );
            }
            else
            {
                // if TGT file is absent then credentials are necessarily present
                // (it's been checked up before)
                //
                // DCT : If the ticket is in the launch file, then we can
                //       use it to login.  So we do not specify credentials.
                if ( strlen( pTicket ) == 0 )
                {
                    snprintf ( pResponse, BUFFER_131072, "STARTqrdesktop\n-u\n%s\n-p\n%s\n-d\n%s\n-s\n%s\n",
                              user, password, domain, shell );
                }
                else
                {
                    snprintf ( pResponse, BUFFER_131072, "STARTqrdesktop\n-s\n%s\n", shell );
                }
            }

            // add rdesktop options
            if ( *options != 0 )
            {
                int num_quotes = 0;

                // first check if the number of double-quotes in the string is even
                while (pResponseData && *pResponseData)
                {
                    if (*pResponseData++ == '"') ++num_quotes;
                }
                if ( num_quotes & 1 )
                {
                    snprintf ( pResponse, BUFFER_131072, "[ERR%05X],[Starting the session],Wrong rdesktop parameters string; the pit-file may be broken",
                        INPUT_PARAMETERS_ERROR );
                    syslog (LOG_NOTICE, "%s", pResponse);
                    goto send_response;
                }
                // now parse the options string
                pResponseData = options;
                while (pResponseData && isspace(*pResponseData)) ++pResponseData;
                // till the end of the string:
                while (pResponseData && *pResponseData)
                {
                    // go ahead while all whitespaces are between quotes
                    endOfResponseUnit = pResponseData;
                    bool qq = false;

                    while (endOfResponseUnit && *endOfResponseUnit)
                    {
                        if ('"' == *endOfResponseUnit)
                        {
                            qq = !qq;
                        }
                        if (isspace(*endOfResponseUnit) && !qq)
                        {
                            break;
                        }
                        ++endOfResponseUnit;
                    }
                    // add all the passed symbols as a separate parameter
                    strncat(pResponse, pResponseData, endOfResponseUnit-pResponseData);
                    strcat(pResponse, "\n");
                    pResponseData = endOfResponseUnit;
                    while (pResponseData && isspace(*pResponseData)) ++pResponseData;
                }
            }

            if ( *proxyaddr != 0 )
            {
                current_session->use_proxy( serverip, proxyaddr );
                strcat ( pResponse, "--use-proxy\n" );
            }

            // add last five parameters: <localhost> <remotehost> <session name> <sslgateway> <pndnaproxy>

            // PNDNA Support:  In the case that the pndna options are specified
            //                 this section needs to replace the remote desktop's
            //                 IP and port with the pndnaproxy's.
            std::string remotePndnaPortStr;
            memset( pndnaOptions, 0x0, BUFFER_1024 );
            ssPitFile.FindItem( "dnaproxy:s:", pndnaOptions );
            if ( strlen(pndnaOptions) != 0 )
            {
                //  Current session is launched from Web-IT.
                //  If this parameter is empty then user doesn't want to use EOP Xtream.
                //  If it isn't empty - it contains port number for remote host to use by "pndnaproxy".
                //  Format of data is next:
                //      TCP:<port>
                //  We need only port.
                remotePndnaPortStr = ::strrchr(pndnaOptions, ':') + 1; 
                syslog( LOG_INFO, "Turning on EOP Xtream (Web-IT) on %s port.", remotePndnaPortStr.c_str());
            }
            else if ( ssPitFile.FindItem( "dnaproxy:i:", pndnaOptions ) && *pndnaOptions != 0 && *pndnaOptions != '0' )
            {
                //  Current session is launched from vWorkspace GUI.
                syslog( LOG_INFO, "User would like to use EOP Xtream (vWorkspace)" );

                memset( pndnaOptions, 0x0, BUFFER_1024 );
                ssRes.FindItem( "dnaproxy:s:", pndnaOptions);
                if ( strlen(pndnaOptions) != 0 )
                {
                    //  If this parameter is empty then Broker was configured with "Disabled" option of EOP Xtream.
                    //  If it isn't empty - it contains port number for remote host to use by "pndnaproxy".
                    //  Format of data is next:
                    //      TCP:<port>
                    //  We need only port.
                    remotePndnaPortStr = ::strrchr(pndnaOptions, ':') + 1; 
                    syslog( LOG_INFO, "Turning on EOP Xtream (vWorkspace) on %s port.", remotePndnaPortStr.c_str());
                }
                else
                {
                    syslog( LOG_INFO, "Option 'EOP Xtream' on Broker is 'Disabled'." );
                }
            }

            std::string serverAddr = serverip;
            std::string pndnaProxy = serverip;
            bool usePndna = false;
            if ( remotePndnaPortStr.size() > 0 )
            {
                syslog( LOG_NOTICE, "EOP Xtream connection allowed by connector for %s", serverip );

                // Find the next available port number (look up the session?)
                freeLocalPort = getFreeLocalPort( 28000, 28100 );

                std::stringstream localEndpointStr;
                localEndpointStr << "localhost:" << freeLocalPort;
   
                std::string remoteEndpoint;
                if (*sslgateway)
                {
                    remoteEndpoint = local_addr;
                    remoteEndpoint += ":3389";
                }
                else
                {
                    remoteEndpoint = serverip;
                    remoteEndpoint += ':'; 
                    remoteEndpoint += remotePndnaPortStr;
                }

                pid_t pndnaPid = startPNDNAProxy(localEndpointStr.str(), remoteEndpoint);
                
                if (pndnaPid != -1)
                {
                    // This clause is for the parent process
                    // pndnaPid is the PID of the child
                    syslog( LOG_INFO, "pnsession has launched child process PID = %i for pndnaproxy", pndnaPid );
                    syslog( LOG_INFO, "The session handle is %p", current_session );
                    // Associate this PID with the session so that it can be killed when the session terminates
                    current_session->setPndnaProxyPid( pndnaPid );
                    current_session->server()->register_pndnaProxyPid( session_name, pndnaPid );

                    usePndna = true;
                    pndnaProxy = localEndpointStr.str();
                    if (!*sslgateway) 
                    {
                        serverAddr = localEndpointStr.str();
                    }
                }
                else
                {
                    syslog (LOG_NOTICE, "pnsession::process can't start pndnaproxy");
                }
            }
            else
            {
                syslog( LOG_NOTICE, "Regular connection requested for %s", serverip );
            }

            std::string tail = local_addr;
            tail += "\n";
            tail += serverAddr + "\n";
            tail += session_name + "\n";
        
            if (*sslgateway) 
            {
        	    tail += std::string(sslgateway) + "\n";
                if (usePndna)
                {
                    tail += pndnaProxy + "\n";
                }
            }
        
            if ((response.size() - ::strlen(pResponse)) < tail.size())
            {
                response.resize(response.size() + tail.size());
            }
        
            strcat(pResponse, tail.c_str());
        }
        else
        {
            snprintf ( pResponse, BUFFER_131072, "SHARED session: %s", session_name.c_str() );
            syslog ( LOG_NOTICE, "%s", pResponse );

            const std::string shelStr = Utilities::Lexical::ReplaceString<std::string>(std::string(&shell[0]), "\"\"", "\"");

            std::stringstream commandStream;
            commandStream << "SA" << shelStr << "\n" << workdir;

            session->send(commandStream.str().c_str());
        }

    }   // to avoid compiler error "crosses initialization of std::..."

send_response:

    if ( tgt )
    {
        free( tgt );
    }
    poptFreeContext( optCon );
    if ( strlen( pResponse ) == 0 )
    {
        snprintf( pResponse, BUFFER_131072, "[ERR%05X],[],",   /* !!! What to tell user in this case? !!! */
        UNKNOWN_ERROR);
    }
#ifdef DEBUG
    syslog ( LOG_INFO, "At send response: %s", pResponse );
#endif
    current_session->send ( pResponse );

    if(user)         free(user);
    if(password)     free(password);
    if(domain)       free(domain);
    if(app)          free(app);
    if(new_password) free(new_password);
    return NULL;
}
